(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

 
Require Import Basic.
Require Import ExtensionalArrays.
Require Import Model.
Require Import Coq.Program.Wf.

(** Good *)

Lemma axiom_good: forall v : Value,
  Good v = true -> (is_C v = true -> Good_C v = true) /\ (is_E v = true -> Good_E v = true).
Proof. intros; split; intros; unfold Good_C, Good_E; assumption. Qed.

(** ** Operators *)

Lemma axiom_sum : forall i1 i2 : Z,
  (O_Sum (v_int i1) (v_int i2)) = (v_int (Zplus i1 i2)).
Proof. intros; reflexivity. Qed.

Lemma axiom_minus : forall i1 i2 : Z,
  (O_Minus (v_int i1) (v_int i2)) = (v_int (Zminus i1 i2)).
Proof. intros; reflexivity. Qed.

Lemma axiom_mult : forall i1 i2 : Z,
  (O_Mult (v_int i1) (v_int i2)) = (v_int (Zmult i1 i2)).
Proof. intros; reflexivity. Qed.

Lemma axiom_eq : forall v1 v2 : Value,
  (O_EQ v1 v2) = (if (syn_beq_val v1 v2) then v_tt else v_ff).
Proof. intros v1 v2. unfold O_EQ. destruct (syn_beq_val v1 v2); reflexivity. Qed.

Lemma axiom_not : forall v : Value,
  O_Not v = if negb (syn_beq_val v v_tt) then v_tt else v_ff.
Proof. intros. unfold O_Not. destruct v as [[[i | s | b |] | svm | vb] pv];
    simpl; try reflexivity. destruct b; reflexivity. Qed.

Lemma axiom_and : forall v1 v2 : Value,
  (O_And v1 v2) = if andb (syn_beq_val v1 v_tt) (syn_beq_val v2 v_tt) then v_tt else v_ff.
Proof. intros. unfold O_And. 
  destruct v1 as [[[i1 | s1 | b1 |] | svm1 | vb1] pv1];
  destruct v2 as [[[i2 | s2 | b2 |] | svm2 | vb2] pv2]; simpl; try reflexivity;
  try (destruct b1; try reflexivity); try (destruct b2; try reflexivity).
Qed.

Lemma axiom_or : forall v1 v2 : Value,
  (O_Or v1 v2) = if orb (syn_beq_val v1 v_tt) (syn_beq_val v2 v_tt) then v_tt else v_ff.
Proof. intros. unfold O_Or. 
  destruct v1 as [[[i1 | s1 | b1 |] | svm1 | vb1] pv1];
  destruct v2 as [[[i2 | s2 | b2 |] | svm2 | vb2] pv2]; simpl; try reflexivity;
  try (destruct b1; try reflexivity); try (destruct b2; try reflexivity).
Qed.

Lemma axiom_gt : forall i1 i2 : Z,
  (O_GT (v_int i1) (v_int i2)) = match Zcompare i1 i2 with Gt => v_tt | _ => v_ff end.
Proof. intros. unfold O_GT. simpl.
  remember (Zcompare i1 i2) as c. destruct c; simpl; reflexivity. Qed.

Lemma axiom_lt : forall i1 i2 : Z,
  (O_LT (v_int i1) (v_int i2)) = match Zcompare i1 i2 with Lt => v_tt | _ => v_ff end.
Proof. intros. unfold O_LT. simpl.
  remember (Zcompare i1 i2) as c. destruct c; simpl; reflexivity. Qed.

(** Arrays *)

(* XXX proving this from so complicated definitions seems hard 
-- as a first simplification could define alpham' and betam' that are variants
of alpham and betam without refinements refinements (they work RawValues)
and prove that they form an isomorphism
-- the extensionality axiom for arrays might be useful as well
-- the induction principle for arrays too
*)

Lemma axiom_svmap_iso_1 :
  forall am : SVMapArray, FiniteE am = true -> (alpham (betam am)) = am.
Admitted.


Lemma default_alpham : forall svm, proj1_sig (default (alpham svm)) = None.
Proof.
  intros. destruct svm as [rsvm [H1 [H2 H3]]]. unfold alpham.
  induction rsvm as [| [s rv] rsvm'].
  Case "svm=nil". reflexivity.
  Case "svm=(s,rv)::rsvm'".
      (* hard to manipulate H1-H3 since they are used in the conclusion
        If instead we try to erewrite <- IHrsvm' (or rewrite -> IHrsvm' only at the end)
        it works for a while,
        but at the very end we get error: Uninstantiated existential variable *)
      remember H1 as H1'. clear HeqH1'. rewrite fst_split_cons in H1. inversion H1. subst.
      inversion H2. subst.
      remember H3 as H3'. clear HeqH3'. rewrite snd_split_cons in H3. inversion H3. subst.
      specialize IHrsvm' with (H1 := H5) (H2 := H7) (H3 := H9).
    rewrite <- IHrsvm'.
    rewrite WfExtensionality.fix_sub_measure_eq_ext. simpl proj1_sig.
       (* only this didn't completely blow up the proof space *)
    (* rewrite default_store. -- for some reason default gets unfolded *)
    unfold default.
    erewrite eq_exist. (* this really needed for the last next step *)
    reflexivity.
Qed.

(* this doesn't use proof irrelevance *)
Lemma proj1_sig_exist : forall X (P : X -> Prop) (x : X) (y : {x : X | P x}),
  proj1_sig y = x -> exists pf, y = exist _ x pf.
Proof. intros. destruct y as [y' Hy']. simpl in H. subst. exists Hy'. reflexivity. Qed.

(* this uses proof irrelevance; seems more easy to use than proj1_sig_exist *)
Lemma proj1_sig_forall : forall X (P : X -> Prop) (x : X) (y : {x : X | P x}),
  proj1_sig y = x -> forall pf, y = exist _ x pf.
Proof. intros. destruct y as [y' Hy']. simpl in H. subst. apply eq_exist. Qed.

Lemma default_alpham2 : forall svm, exists pf, default (alpham svm) = exist _ None pf.
Proof. intros. apply proj1_sig_exist. apply default_alpham. Qed.

Lemma default_alpham3 : forall svm pf, default (alpham svm) = exist _ None pf.
Proof. intros. apply proj1_sig_forall. apply default_alpham. Qed.

(* can't be used for rewriting (Error: The term provided is not an applied relation.)
Lemma let_proj1_sig : forall X Y (x : X) (P : X -> Prop) (y : sig P) (f : X -> Y),
  let (x, _) := y in f x = let x := proj1_sig y in f x.
Admitted.
*)

Lemma Normal_None : (fun vo : option RawValue => lift_option Normal vo) None.
Proof. simpl. apply I. Qed.

(* XXX this should actually be broken in the axiomatization into 2 separate axioms
-- will have better patterns? *)
Lemma axiom_svmap_iso_2' : forall svm : SVMap,
  (FiniteE (alpham svm)) = true.
Proof. intro svm. unfold FiniteE. erewrite default_alpham3. reflexivity.
  Existential 1 := Normal_None. Qed.

Lemma axiom_svmap_iso_2'' : forall svm : SVMap,
  (betam (alpham svm)) = svm.
Proof.
    destruct svm as [rsvm [H1 [H2 H3]]]. induction rsvm as [| [s rv] rsvm'].
    SCase "svm = nil".
      unfold betam, alpham. simpl.
      (* reflexivity. -- the terms are the same, proofs different *)
      eapply eq_exist.
    SCase "svm = ".
Admitted.

(*    program_simpl. simpl_eq. erewrite proof_irrelevance. simplify_eqs. *)

(** ** Entities *)

Lemma axiom_FiniteE : forall svm : SVMapArray,
  FiniteE svm = true <-> (default svm) = NoValue.
Proof. intro svm; split; intro H; unfold FiniteE, NoValue in *; remember (default svm) as vo;
  destruct vo as [rvo pf]; destruct rvo as [v | ]; try inversion H; try reflexivity;
  try eapply eq_exist.
Qed.

Lemma and_true : forall b1 b2, b1 && b2 = true -> b1 = true /\ b2 = true.
Proof. intros. split; destruct b1; destruct b2; simpl in *; try reflexivity; try inversion H. Qed.

Lemma axiom_Good_E : forall v : Value,
  Good_E v = true <-> andb (is_E v) (FiniteE (alpham (out_E v))) = true.
Proof. intro v; split; intro H; unfold Good_E, FiniteE in *; try reflexivity.
  rewrite H. simpl. erewrite default_alpham3. reflexivity.
(* -- more complicated(?) proof using default_alpham2
  assert(exists pf, default (alpham (out_E v)) = exist _ None pf) as H0. apply default_alpham2.
  destruct H0 as [pf H0']. rewrite H0'. reflexivity.
*)
  apply and_true in H. destruct H. assumption.
  Existential 1 := Normal_None.
Qed.


Lemma axiom_v_eempty : v_eempty = E' (betam (const string NoValue)).
Proof. unfold betam, v_eempty. simpl. apply eq_exist. Qed.

Lemma cmp_str_equivalence :
equivalence string (fun d1 d2 : string => beq_D string cmp_str d1 d2 = true).
Admitted.

(** XXX Trying to prove axiom_v_eupdate' *)

Lemma remove_sort : forall s svm,
  (remove string RawValue eq_str_dec s (sort_svm svm) =
  sort_svm (remove string RawValue eq_str_dec s svm)).
Admitted.

Lemma remove_remove_dup_helper : forall s svm avoid,
     (remove string RawValue eq_str_dec s
        (remove_dup_helper string RawValue eq_str_dec avoid svm)) =
     (remove_dup_helper string RawValue eq_str_dec (s::avoid) svm).
Admitted.

(* This used to be defined in Model.v, it got removed from there, but it's still used below *)
Definition insert_in_sorted_svm := insert_in_sorted le_sv le_sv_dec.


(* 
XXX Cases: d = s and d <> s -- in the first we can remove the inner store,
in the second we can swap the stores -- both by extensionality? *)
(* Note: if we are not careful when we call IH we have that
 contained string eq_str_dec s (s::avoid) = true -- we assume the opposite *)
Lemma betam_store_helper :
forall (a : Array string ValueOption) (s : string) (v : Value) (avoid : list string) pf,
contained eq_str_dec s avoid = right _ pf ->
sort_svm
  (remove_dup_helper string RawValue eq_str_dec avoid
     (map
        (fun s0 : string =>
         let (x, _) := select cmp_str (store a s (SomeValue v)) s0 in
         match x with
         | Some v0 => (s0, v0)
         | None => (s0, G G_Null)
         end) (support (store a s (SomeValue v))))) =
insert_in_sorted_svm (s, proj1_sig v)
  (remove string RawValue eq_str_dec s
     (sort_svm
        (remove_dup_helper string RawValue eq_str_dec avoid
           (map
              (fun s0 : string =>
               let (x, _) := select cmp_str a s0 in
               match x with
               | Some v0 => (s0, v0)
               | None => (s0, G G_Null)
               end) (support a))))).
Proof.
  induction a using arr_ind; intros.
  Case "a = const d".
    unfold sort_svm, remove_dup. simpl.
    rewrite select_store_eq. simpl. rewrite H. reflexivity. apply cmp_str_equivalence.
  Case "a = store a d r".
    unfold remove_dup, insert_in_sorted_svm in *.
    rewrite support_store. simpl map. rewrite select_store_eq. unfold SomeValue in *.
    simpl remove_dup_helper. rewrite H. simpl sort_svm.
    rewrite <- remove_remove_dup_helper. rewrite <- remove_sort.
    (* rewrite IHa. -- can't be applied since (support (store a d r) <> support a *)
    rewrite support_store. simpl map. rewrite select_store_not_eq.
Admitted.

(* XXX Potential lemma that might help (together with the isomorphism) to prove axiom_v_eupdate'
-- might also be useful for proving the isomorphism ... or not *)
Lemma betam_store : forall a s v,
  proj1_sig (betam (store a s (SomeValue v))) =
  insert_in_sorted_svm (s, (proj1_sig v))
    (remove string RawValue eq_str_dec s (proj1_sig (betam a))).
(* Old RHS :
  insert_in_sorted_svm (s, (proj1_sig v))
    (remove_dup_helper string RawValue eq_str_dec (s :: nil) (proj1_sig (betam a))). *)
(* Old RHS: sort_svm (remove_dup eq_str_dec ((s, proj1_sig v) :: proj1_sig (betam a))). *)
Proof.
(* Old proof attempt:
  induction a using arr_ind; intros.
  Case "a = const d".
    unfold sort_svm, remove_dup. simpl.
    rewrite select_store_eq. reflexivity. apply cmp_str_equivalence.
  Case "a = store a d r".
    unfold betam in *. simpl proj1_sig in *. unfold remove_dup, insert_in_sorted_svm in *.
    rewrite support_store. simpl map. rewrite select_store_eq. unfold SomeValue in *.
    simpl remove_dup_helper. simpl sort_svm.
    rewrite remove_sort. rewrite remove_remove_dup_helper.
    (* rewrite IHa. -- can't be applied since  nil <> s :: nil (avoid)
     fine: the latest store overrides another, for remove_dup (s, .. v) :: .. keeps s *)
    (* note: on the RHS: we know that (betam a) is sorted and contains no duplicates
    -- if we can show that sorting is stable then both operations leave it unchanged
    -- so the RHS only adds (s,v) on the right position, if it doesn't already exist *)
*)
  unfold betam. simpl proj1_sig. unfold remove_dup. intros. eapply betam_store_helper.
  remember (contained_nil) as H. edestruct H as [pf H']. erewrite H'. eauto.
  Existential 1 := eq_str_dec.
Qed.


Lemma axiom_v_eupdate' : forall (l : string) (v : Value) (svm : SVMap),
  v_eupdate l v (E' svm) =
     (E' (betam (store (alpham svm) l (SomeValue v)))).
Proof. intros. unfold v_eupdate, E'. destruct svm as [rsvm [H1 [H2 H3]]]. simpl.
  apply eq_exist2.
  induction rsvm as [| [s' v'] rsvm']; unfold remove_dup, alpham, NoValue.
  Case "rsvm=nil".
     simpl. rewrite select_store_eq. reflexivity. unfold EqD. apply cmp_str_equivalence.
  Case "rsvm=(s',v')::rsvm'". admit. (* XXX complicated *)
Qed.
 
Lemma axiom_v_has_field : forall (l : string) (svm : SVMap),
  v_has_field l (E' svm) = true <-> ~ ((select cmp_str (alpham svm) l) = NoValue).
Admitted.

Lemma axiom_v_dot : forall (l : string) (svm : SVMap),
  v_dot l (E' svm) = (of_SomeValue (select cmp_str (alpham svm) l)).
Admitted.

(** ** Collections *)

Lemma default_alphab : forall a, beq_int (default (alphab a)) 0 = true.
Admitted.

Lemma is_C_exists : forall v, is_C v = true -> exists vb, v = C' vb.
Proof. intros v H. destruct v as [[g | svm | vb] Hv]; try inversion H.
  inversion Hv. subst. exists (exist _ vb (conj H1 H2)). unfold C'. simpl. apply eq_exist. Qed.
  
Lemma axiom_Good_C : forall v : Value,
  Good_C v = true <-> is_C v = true /\ Finite (alphab (out_C v)) = true
                      /\ Positive (alphab (out_C v)) = true.
Proof. intro v. unfold Good_C, Finite, Positive. split; intro H. split. assumption.
  split. rewrite default_alphab. reflexivity. apply is_C_exists in H.
  destruct H as [[vb Hvb] H]. unfold C' in H. subst. unfold out_C. simpl.
  admit. (* XXX TODO -- need to define alphab first *)
  destruct H as [H1 H2]. assumption.
Qed.

Lemma axiom_Finite : forall a : VBagArray, Finite a = true <-> beq_int (default a) 0 = true.
Proof. intro a; unfold Finite; split; intro H; assumption. Qed.

Definition ge_int (i1 i2 : Z) : bool :=
  match Zcompare i1 i2 with
  | Gt => true
  | Eq => true
  | Lt => false
  end.

Lemma axiom_Positive : forall a : VBagArray,
  Positive a = true <-> forall v : Value, ge_int (select cmp_val a v) 0 = true.
Proof. intro a; unfold Positive, ge_int; split; intros H. intro v.
  unfold allb_R in *. (* breaking abstraction? *)
  apply and_true in H. destruct H as [H1 H2].
(*  remember (contained Value eq_val_dec v (support a)). *)
(*  specialize contained with (X := Value) (eqX_dec := eq_val_dec) (x := v) (xs := (support a)). *)
  destruct (contained eq_val_dec v (support a)).
  admit. admit. (* XXX TODO *)
  admit. (* XXX TODO *)
Qed.

(* assumes that default is 0 (Finite) + uses the fact that arrays have finite support *)
Definition a_count (a : VBagArray) : Z :=
  List.fold_left (fun s v => Zplus s (select cmp_val a v)) (support a) 0%Z.

Lemma axiom_a_count_const : (a_count (const Value 0%Z)) = 0%Z.
Proof. unfold a_count. simpl. reflexivity. Qed.

Lemma axiom_a_count_select : forall (a : VBagArray) (v : Value),
  Finite a = true -> (a_count a) = Zplus (select cmp_val a v) (a_count (store a v 0%Z)).
Proof. intros a v H. unfold a_count. admit. (* XXX TODO *)
Qed.

Lemma axiom_v_count : forall v : Value,
  is_C v = true -> (v_count v) = (a_count (alphab (out_C v))).
Proof. intros v H. admit. (* XXX TODO, seems complicated *)
Qed.

Lemma axiom_v_zero : v_zero = (C' (betab (const Value 0%Z))).
Admitted.

Lemma axiom_v_add_many : forall (v : Value) (i : Z) (vb : VBag),
  (v_add_many v i (C' vb)) =
     (C' (betab (store (alphab vb) v (Zplus i (select cmp_val (alphab vb) v))))).
Admitted.

Lemma axiom_v_add : forall (v : Value) (vs : Value),
  (v_add v vs) = (v_add_many v 1 vs).
Proof. intros v vs. unfold v_add_many, v_add_many'. reflexivity. Qed.

Lemma axiom_v_accumulate_v_zero : forall (clos : Closure2) (initial : Value),
  (v_accumulate clos v_zero initial) = initial.
Proof. intros clos initial. unfold v_accumulate, v_acc_fold, out_C, v_zero. simpl.
  rewrite WfExtensionality.fix_sub_measure_eq_ext. reflexivity.
Qed.

Lemma axiom_v_accumulate_v_add : forall (clos : Closure2) (initial : Value) (v : Value) (vs : Value),
  (v_accumulate clos (v_add v vs) initial) =
     (v_accumulate clos vs (v_apply2 clos v initial)).
Proof. intros clos initial v vs. unfold v_apply2, v_add. destruct v as [rv Hv]. simpl.
  destruct rv as [| svm | vb]. unfold v_accumulate, v_acc_fold.
  rewrite WfExtensionality.fix_sub_measure_eq_ext. simpl.
  Admitted.

(** XXX in the implementation we have a lot more primitives on collections *)


(* Sigma types suck, so here is an attempt to prove the isomorphism without sigma types *)

Fixpoint alpham' (svm : list (string * RawValue)) : Array string (option RawValue) :=
  match svm with
  | nil => const string None
  | (s,v) :: svm' => store (alpham' svm') s (Some v)
  end.

Definition betam' (a : Array string (option RawValue)) :  list (string * RawValue) :=
  (List.map (fun s =>
    match select cmp_str a s with
    | Some v => (s, v)
    | None => (s, (G G_Null)) (* XXX unreachable? *)
    end) (support a)).

Lemma axiom_svmap_iso_2_no_sigma : forall svm : list (string * RawValue),
  (betam' (alpham' svm)) = svm.
Proof.
  intros. induction svm as [| [l v] svm']. reflexivity.
  unfold betam'. simpl.
  rewrite support_store.
  simpl map. unfold betam' in IHsvm'.
  rewrite select_store_eq.
  assert(forall s, select cmp_str (store (alpham' svm') l (Some v)) s = select cmp_str (alpham' svm') s).
    intros. apply select_store_not_eq. unfold beq_D. simpl. admit. (* XXX need some duplicate removing in supoport? *)
(*  setoid_rewrite H. -- XXX *)
Admitted.

Definition betam'' (a : Array string (option RawValue)) :  list (string * RawValue) :=
  (List.map (fun sov =>
    match sov with
    | (s, Some v) => (s, v)
    | (s, None) => (s, (G G_Null)) (* XXX unreachable? *)
    end) (fst a)).

Lemma axiom_svmap_iso_2_no_sigma_2nd_try_with_a_little_cheating : forall svm : list (string * RawValue),
  (betam'' (alpham' svm)) = svm.
Proof.
  intros. induction svm as [| [l v] svm']; simpl. reflexivity.
  unfold betam''. simpl. 
  unfold betam'' in IHsvm'.
  rewrite IHsvm'.
  reflexivity.
Qed.

(* Now back to the original problem *)

Definition betam11 (a : SVMapArray) : SVMap.
  refine (fun a =>
  exist (fun svm => NoDup (fst (split svm)) /\ Sorted_svm svm /\ IndAll Normal (snd (split svm)))
    (sort_svm (remove_dup eq_str_dec
      (List.map (fun sov =>
                  match sov with
                  | (s, exist (Some v) pf) => (s, v)
                  | (s, exist None pf) => (s, (G G_Null)) (* XXX unreachable? *)
                  end) (fst a)) )) _ ).
(*--- BetamEnd *)
Proof.
  split; try split.
  Case "NoDup". apply nodup_sort_remove_dup.
  Case "Sorted". apply (insertion_sort_sorted (string * RawValue) le_sv le_sv_trans le_sv_total).
  Case "IndAll". simpl. eapply IndAll_Permutation with (xs := 
       (snd
         (split
           (remove_dup eq_str_dec
              (map
                 (fun
                    sov : string *
                          {vo : option RawValue | lift_option Normal vo} =>
                  let (s, y) := sov in
                  let (x, _) := y in
                  match x with
                  | Some v => (s, v)
                  | None => (s, G G_Null)
                  end) (fst a)))))).
    SCase "IndAll".
      apply IndAll_snd_split.
      apply IndAll_remove_dup.
      apply IndAll_map. intros.
      destruct x as [s0 y0]. destruct y0 as [rvo Hrvo]. destruct rvo as [rv |]. simpl in *. assumption.
      apply Normal_null.
    SCase "Permutation". apply Permutation_snd_split.
      apply Permutation_sym. apply insertion_sort_permutation.
Defined.

Lemma axiom_svmap_iso_11 : forall svm : SVMap,
  (betam11 (alpham svm)) = svm.
Proof.
  intros. destruct svm as [svm H]. induction svm as [| [l v] svm']; simpl.
    unfold betam11. erewrite eq_exist. reflexivity.
  erewrite eq_exist. 
  unfold betam11 in *. simpl betam11. 
  erewrite eq_exist. 
(*  erewrite <- IHsvm'. XXX nope
  reflexivity. *)
Admitted.
