(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

Require Import Model.
Require Import Calculus.
Require Import LogicalSemantics.
Require Import Coq.Program.Wf.
Require Import Purity.



Inductive un_op_eval : UnOp -> Value -> Value -> Prop :=
  | uoe_neg_tt : un_op_eval ONot v_tt v_ff
  | uoe_neg_ff : un_op_eval ONot v_ff v_tt.

Inductive bin_op_eval : BinOp -> Value -> Value -> Value -> Prop :=
(* XXX to be defined *)
.


Program Fixpoint vb_to_vs (vb : VBag) {measure List.length vb} : list Value :=
  match vb with
  | nil => nil
  | rv :: vb' => rv :: vb_to_vs vb'
  end.
Next Obligation.
  destruct vb as [rvb Hvb]. simpl in *.
  destruct Hvb as [Hvb1 Hvb2]. subst.
  inversion Hvb2. subst. assumption.
Qed.
Next Obligation.
  destruct vb as [rvb Hvb]. simpl in *.
  destruct Hvb as [Hvb1 Hvb2]. subst.
  inversion Hvb2. subst.
  inversion Hvb1. subst.
  split; assumption.
Qed.

Definition let_seq (y : string) (es : list Exp) (e' : Exp) :=
  List.fold_right (fun e e' => E_Let y e e') e' es.

Definition and_seq (es : list Exp) :=
  List.fold_right (fun e' e => E_And e e') E_True es.

Definition function := (string * Exp)%type.
Program Definition null_function : function := ("", E_Value (v_null)).

Definition functions : string -> function.
  apply (fun s => null_function).
Qed.

Inductive Eval : Exp -> Result -> Prop :=
  | eval_value : forall v,
      Eval (E_Value v) (Return v)
  | eval_un_op_1 : forall e op,
      Eval e Error ->
      Eval (E_UnOp op e) Error
  | eval_un_op_2 : forall e op v,
      Eval e (Return v) ->
      (~exists v', un_op_eval op v v') ->
      Eval (E_UnOp op e) Error
  | eval_un_op_3 : forall e op v v',
      Eval e (Return v) ->
      un_op_eval op v v' ->
      Eval (E_UnOp op e) (Return v')
  | eval_bin_op_1_1 : forall e1 e2 op,
      Eval e1 Error ->
      Eval (E_BinOp op e1 e2) Error
  | eval_bin_op_1_2 : forall e1 e2 op v,
      Eval e1 (Return v) ->
      Eval e2 Error ->
      Eval (E_BinOp op e1 e2) Error
  | eval_bin_op_2 : forall e1 e2 op v1 v2,
      Eval e1 (Return v1) ->
      Eval e2 (Return v2) ->
      (~exists v', bin_op_eval op v1 v2 v') ->
      Eval (E_BinOp op e1 e2) Error
  | eval_bin_op_3 : forall e1 e2 op v1 v2 v',
      Eval e1 (Return v1) ->
      Eval e2 (Return v2) ->
      bin_op_eval op v1 v2 v' ->
      Eval (E_BinOp op e1 e2) (Return v')
  | eval_cond_1 : forall e1 e2 e3 r,
      Eval e1 r ->
      r <> Return(v_tt) ->
      r <> Return(v_ff) ->
      Eval (E_If e1 e2 e3) Error
  | eval_cond_2_1 : forall e1 e2 e3 r,
      Eval e1 (Return v_tt) ->
      Eval e2 r ->
      Eval (E_If e1 e2 e3) r
  | eval_cond_2_2 : forall e1 e2 e3 r,
      Eval e1 (Return v_ff) ->
      Eval e3 r ->
      Eval (E_If e1 e2 e3) r
  | eval_let_1 : forall x e1 e2,
      Eval e1 Error ->
      Eval (E_Let x e1 e2) Error
  | eval_let_2 : forall x e1 e2 v r,
      Eval e1 (Return v) ->
      Eval (subst_exp e2 v x) r ->
      Eval (E_Let x e1 e2) r
(* XXX this produces weak induction principle, but we fix it in the proofs that use it *)
  | eval_entity_1 : forall les1 les2 lj ej,
      (forall l e, In (l,e) les1 -> exists v, Eval e (Return v)) ->
      Eval ej Error ->
      Eval (E_Entity (les1 ++ (lj,ej) :: les2)) Error
(* XXX this produces weak induction principle, but we fix it in the proofs that use it *)
  | eval_entity_2 : forall les v,
      is_E v = true ->
      (forall li ei, In (li,ei) les -> exists vi,
        Eval ei (Return vi) /\ v_has_field li v = true /\ v_dot li v = vi) ->
      (forall li, v_has_field li v = true -> exists ei, In (li, ei) les) ->
      Eval (E_Entity les) (Return v)
  | eval_dot_1_1 : forall e l,
      Eval e Error ->
      Eval (E_Dot e l) Error
  | eval_dot_1_2 : forall e l v,
      Eval e (Return v) ->
      is_E v && v_has_field l v = false ->
      Eval (E_Dot e l) Error
  | eval_dot_2 : forall e l v,
      Eval e (Return v) ->
      is_E v && v_has_field l v = true ->
      Eval (E_Dot e l) (Return (v_dot l v))
  | eval_add_1 : forall e1 e2,
      Eval e1 Error ->
      Eval (E_Add e1 e2) Error
  | eval_add_2_1 : forall e1 e2 v1,
      Eval e1 (Return v1) ->
      Eval e2 Error ->
      Eval (E_Add e1 e2) Error
  | eval_add_2_2 : forall e1 e2 v1 v2,
      Eval e1 (Return v1) ->
      Eval e2 (Return v2) ->
      is_C v2 = false ->
      Eval (E_Add e1 e2) Error
  | eval_add_3 : forall e1 e2 v1 v2,
      Eval e1 (Return v1) ->
      Eval e2 (Return v2) ->
      is_C v2 = true ->
      Eval (E_Add e1 e2) (Return (v_add v1 v2))
  | eval_appl_1 : forall f e,
      Eval e Error ->
      Eval (E_App f e) Error
  | eval_appl_2 : forall f arg ex r v e,
      Eval e (Return v) ->
      (arg,ex) = functions f ->
      Eval (subst_exp ex v arg) r ->
      Eval (E_App f e) r
  | eval_accum_1_1 : forall x y e1 e2 e3,
      Eval e1 Error ->
      Eval (E_Acc x e1 y e2 e3) Error
  | eval_accum_1_2 : forall x y e1 e2 e3 v1,
      Eval e1 (Return v1) ->
      is_C v1 = false ->
      Eval (E_Acc x e1 y e2 e3) Error
  | eval_accum_lets : forall x e1 y e2 e3 r v1 vs,
      Eval e1 (Return v1) ->
      is_C v1 = true ->
      (Permutation vs (vb_to_vs (out_C v1))) ->
      Eval (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs) (E_Var y))) r ->
      Eval (E_Acc x e1 y e2 e3) r
  | eval_test_wrong : forall e T,
      Eval e Error ->
      Eval (E_In e T) Error
  | eval_test_any : forall e v,
      Eval e (Return v) ->
      Eval (E_In e T_Any) (Return v_tt)
  | eval_test_integer : forall e v,
      Eval e (Return v) ->
      Eval (E_In e T_Integer) (Return (v_logical (In_Integer v)))
  | eval_test_text : forall e v,
      Eval e (Return v) ->
      Eval (E_In e T_Text) (Return (v_logical (In_Text v)))
  | eval_test_logical : forall e v,
      Eval e (Return v) ->
      Eval (E_In e T_Logical) (Return (v_logical (In_Logical v)))
  | eval_test_entity_1 : forall e l T v r,
      Eval e (Return v) ->
      is_E v && v_has_field l v = true ->
      Eval (E_In (E_Value (v_dot l v)) T) r ->
      Eval (E_In e (T_Entity l T)) r
  | eval_test_entity_2 : forall e l T v,
      Eval e (Return v) ->
      is_E v && v_has_field l v = false ->
      Eval (E_In e (T_Entity l T)) (Return v_ff)
  | eval_test_coll_1 : forall e T v,
      Eval e (Return v) ->
      is_C v = false ->
      Eval (E_In e (T_Coll T)) (Return v_ff)
  | eval_test_coll_2 : forall e v T r,
      Eval e (Return v) ->
      is_C v = true ->
      Eval (and_seq (List.map (fun v' => E_In (E_Value v') T) (vb_to_vs (out_C v)))) r ->
      Eval (E_In e (T_Coll T)) r
  | eval_test_refine : forall x T v e e' r,
      Eval e (Return v) ->
      Eval (E_And (E_In (E_Value v) T) (subst_exp e' v x)) r ->
      Eval (E_In e (T_Refine x T e')) r
.


Tactic Notation "eval_cases" tactic(first) tactic(c) :=
  first;
  [ c "eval_value"
  | c "eval_un_op_1"
  | c "eval_un_op_2"
  | c "eval_un_op_3"
  | c "eval_bin_op_1_1"
  | c "eval_bin_op_1_2"
  | c "eval_bin_op_2"
  | c "eval_bin_op_3"
  | c "eval_cond_1"
  | c "eval_cond_2_1"
  | c "eval_cond_2_2"
  | c "eval_let_1"
  | c "eval_let_2"
  | c "eval_entity_1"
  | c "eval_entity_2"
  | c "eval_dot_1_1"
  | c "eval_dot_1_2"
  | c "eval_dot_2"
  | c "eval_add_1"
  | c "eval_add_2_1"
  | c "eval_add_2_2"
  | c "eval_add_3"
  | c "eval_appl_1"
  | c "eval_appl_2" 
  | c "eval_accum_1_1"
  | c "eval_accum_1_2"
  | c "eval_accum_lets"
  | c "eval_test_wrong"
  | c "eval_test_any"
  | c "eval_test_integer"
  | c "eval_test_text"
  | c "eval_test_logical"
  | c "eval_test_entity_1"
  | c "eval_test_entity_2"
  | c "eval_test_coll_1"
  | c "eval_test_coll_2"
  | c "eval_test_refine"
].

(* We prove that the previously primitive rules for assert are now derivable *)

Lemma eval_assert_1 : forall e T,
  Eval e Error ->
  Eval (E_Assert e T) Error.
Proof.
  intros. unfold E_Assert. destruct (fresh (fv_exp e ++ fv_type T)).
  apply eval_let_1; assumption.
Qed.

Lemma subst_type_fresh : forall x v T,
  ~ In x (fv_type T) ->
  subst_type T v x = T.
Admitted.

Lemma eval_assert_2 : forall e T v r,
  Eval e (Return v) ->
  Eval (E_If (E_In (E_Value v) T) (E_Value v) (E_Dot (E_Value v_zero) "l")) r ->
  Eval (E_Assert e T) r.
Proof.
  intros. unfold E_Assert. destruct (fresh (fv_exp e ++ fv_type T)).
  eapply eval_let_2. eassumption. simpl. rewrite beq_str_refl.
  rewrite not_in_app in n. destruct n.
  rewrite subst_type_fresh; assumption.
Qed.

(* We prove that the previous (step-by-step) accumulate rules are derivable *)

Lemma eval_accum_2 : forall x y e1 e2 e3 v1,
  Eval e1 (Return v1) ->
  is_C v1 = true ->
  Eval e2 Error ->
  Eval (E_Acc x e1 y e2 e3) Error.
Proof.
  intros. eapply eval_accum_lets; try eassumption.
  apply Permutation_refl.
  apply eval_let_1. assumption.
Qed.

Lemma Permutation_cons_v_remove : forall v v1 vs,
  Permutation vs (vb_to_vs (out_C (v_remove v v1))) ->
  Permutation (v :: vs) (vb_to_vs (out_C v1)).
Admitted.

Lemma subst_exp_let_seq_map_cons : forall e3 e' u x y v0 vs',
  (subst_exp (let_seq y (List.map (fun v => subst_exp e3 v x) (v0::vs')) e') u y) =
  (let_seq y ((subst_exp (subst_exp e3 v0 x) u y) :: (List.map (fun v => subst_exp e3 v x) vs')) e').
Admitted.

Lemma eval_accum_3 : forall x y e1 e2 e3 v1 v2 v,
  Eval e1 (Return v1) ->
  is_C v1 = true ->
  v_mem v v1 = true ->
  Eval e2 (Return v2) ->
  Eval (subst_exp (subst_exp e3 v x) v2 y) Error ->
  Eval (E_Acc x e1 y e2 e3) Error.
Proof.
  intros. eapply eval_accum_lets with (vs := v :: vb_to_vs (out_C (v_remove v v1))); try eassumption.
  apply Permutation_cons_v_remove. apply Permutation_refl.
  eapply eval_let_2; try eassumption.
  rewrite subst_exp_let_seq_map_cons. simpl.
  apply eval_let_1. assumption.
Qed.

Lemma is_C_v_remove : forall v v1,
  is_C v1 = true -> is_C (v_remove v v1) = true.
Admitted.

Lemma eval_accum_4 : forall x y e1 e2 e3 v1 v2 v v' r,
  Eval e1 (Return v1) ->
  is_C v1 = true ->
  v_mem v v1 = true ->
  Eval e2 (Return v2) ->
  Eval (subst_exp (subst_exp e3 v x) v2 y) (Return v') ->
  Eval (E_Acc x (E_Value (v_remove v v1)) y (E_Value v') e3) r ->
  Eval (E_Acc x e1 y e2 e3) r.
Proof.
  intros.
  inversion H4.
    inversion H11.
    inversion H11. subst. rewrite is_C_v_remove in H12; try assumption. inversion H12.
    subst. inversion H14. inversion H9.
      subst. inversion H9; subst. clear H9. inversion H11. subst. clear H11.
      eapply eval_accum_lets with (vs := v :: vs); try eassumption.
      apply Permutation_cons_v_remove. assumption.
      eapply eval_let_2; try eassumption.
      rewrite subst_exp_let_seq_map_cons. simpl.
      eapply eval_let_2. eassumption.
      assumption.
Qed.

Lemma vb_to_vs_out_C_v_empty : forall v1,
  v_empty v1 = v_tt -> (vb_to_vs (out_C v1)) = nil.
Admitted.

Lemma eval_accum_5 : forall x y e1 e2 e3 v1 v2,
  Eval e1 (Return v1) ->
  is_C v1 = true ->
  v_empty v1 = v_tt ->
  Eval e2 (Return v2) ->
  Eval (E_Acc x e1 y e2 e3) (Return v2).
Proof.
  intros. eapply eval_accum_lets; try eassumption.
  apply Permutation_refl.
  eapply eval_let_2; try eassumption.
  rewrite vb_to_vs_out_C_v_empty; try assumption.
  simpl. rewrite beq_str_refl. apply eval_value.
Qed.

(* We prove the old inversion principle for accumulate (trying to make Thorsten's branch obsolete) *)

Lemma vb_to_vs_out_C_empty : forall v1,
  vb_to_vs (out_C v1) = nil -> v_empty v1 = v_tt.
Admitted.

Lemma Permutation_cons_vb_to_vs : forall v' vs' v1,
  Permutation (v' :: vs') (vb_to_vs (out_C v1)) ->
  (v_mem v' v1 = true /\ Permutation vs' (vb_to_vs (out_C (v_remove v' v1)))).
Admitted.

Lemma old_accum_inversion : forall x e1 y e2 e3 r v1 vs,
  Eval e1 (Return v1) ->
  is_C v1 = true ->
  (Permutation vs (vb_to_vs (out_C v1))) ->
  Eval (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs) (E_Var y))) r ->
  (
    (* eval_accum_2 *)
    (r = Error /\ Eval e2 Error) \/

    (* eval_accum_5 *)
    (exists v2,
      r = Return v2 /\
      v_empty v1 = v_tt /\
      Eval e2 (Return v2)
    ) \/

    (* eval_accum_3 *)
    (r = Error /\
      exists v, exists v2, 
        v_mem v v1 = true /\
        Eval e2 (Return v2) /\
        Eval (subst_exp (subst_exp e3 v x) v2 y) Error
    ) \/

    (* eval_accum_4 *)
    (exists v, exists v2, exists v', 
        v_mem v v1 = true /\
        Eval e2 (Return v2) /\
        Eval (subst_exp (subst_exp e3 v x) v2 y) (Return v') /\
        Eval (E_Acc x (E_Value (v_remove v v1)) y (E_Value v') e3) r
    )
  ).
Proof.
  intros.
  inversion H2.
  Case "eval_let_1". left. split; trivial.
  Case "eval_let_2". right. subst.
    generalize dependent v1.
    destruct vs as [| v0 vs']; intros.
    SCase "vs = nil". left.
      simpl in H8. rewrite beq_str_refl in H8. inversion H8. subst.
      exists v. split. reflexivity. split; try assumption.
        apply Permutation_nil in H1. apply vb_to_vs_out_C_empty; assumption.
    SCase "vs = v0 :: vs'". right.
      apply Permutation_cons_vb_to_vs in H1. destruct H1.
      simpl in H8. rewrite beq_str_refl in H8. inversion H8; subst.
      SSCase "eval_let_1". left. split. reflexivity.
        exists v0. exists v. split. assumption. split; assumption.
      SSCase "eval_let_2". right.
        exists v0. exists v. exists v2. repeat split; try assumption.
        eapply eval_accum_lets in H3. apply H3.
        apply eval_value.
        apply is_C_v_remove. assumption.
        eapply eval_let_2. apply eval_value. assumption.
Qed.

(* XXX I don't think this holds (was trying to relate the two structural induction principles)
Lemma old_accum_ind_entailed : forall P : Exp -> Result -> Prop,
  (forall (x y : string) (e1 e2 e3 : Exp) (v1 : Value),
    Eval e1 (Return v1) ->
    P e1 (Return v1) ->
    is_C v1 = true ->
    Eval e2 Error -> P e2 Error -> P (E_Acc x e1 y e2 e3) Error) ->
  (forall (x y : string) (e1 e2 e3 : Exp) (v1 v2 v : Value),
    Eval e1 (Return v1) ->
    P e1 (Return v1) ->
    is_C v1 = true ->
    v_mem v v1 = true ->
    Eval e2 (Return v2) ->
    P e2 (Return v2) ->
    Eval (subst_exp (subst_exp e3 v x) v2 y) Error ->
    P (subst_exp (subst_exp e3 v x) v2 y) Error ->
    P (E_Acc x e1 y e2 e3) Error) ->
  (forall (x y : string) (e1 e2 e3 : Exp) (v1 v2 v v' : Value) (r : Result),
    Eval e1 (Return v1) ->
    P e1 (Return v1) ->
    is_C v1 = true ->
    v_mem v v1 = true ->
    Eval e2 (Return v2) ->
    P e2 (Return v2) ->
    Eval (subst_exp (subst_exp e3 v x) v2 y) (Return v') ->
    P (subst_exp (subst_exp e3 v x) v2 y) (Return v') ->
    Eval (E_Acc x (E_Value (v_remove v v1)) y (E_Value v') e3) r ->
    P (E_Acc x (E_Value (v_remove v v1)) y (E_Value v') e3) r ->
    P (E_Acc x e1 y e2 e3) r) ->
  (forall (x y : string) (e1 e2 e3 : Exp) (v1 v2 : Value),
    Eval e1 (Return v1) ->
    P e1 (Return v1) ->
    is_C v1 = true ->
    v_empty v1 = v_tt ->
    Eval e2 (Return v2) ->
    P e2 (Return v2) -> P (E_Acc x e1 y e2 e3) (Return v2)) ->
  (forall (x : string) (e1 : Exp) (y : string) (e2 e3 : Exp)
          (r : Result) (v1 : Value) (vs : list Value),
    Eval e1 (Return v1) ->
    P e1 (Return v1) ->
    is_C v1 = true ->
    Permutation vs (vb_to_vs (out_C v1)) ->
    Eval (E_Let y e2
             (let_seq y (map (fun v : Value => subst_exp e3 v x) vs)
                (E_Var y))) r ->
    P (E_Let y e2
             (let_seq y (map (fun v : Value => subst_exp e3 v x) vs)
                (E_Var y))) r -> P (E_Acc x e1 y e2 e3) r).
Proof.
  intros P H2 H3 H4 H5. intros.

(* Attempt 2
  remember (E_Let y e2 (let_seq y (map (fun v : Value => subst_exp e3 v x) vs) (E_Var y))) as e.
  induction H7; try inversion Heqe; clear Heqe.
  Case "eval_let_1". subst. eapply H2; try eassumption. (* we don't have P e2 Error *)
*)

(* Attempt 1 
  remember H as H'. clear HeqH'.
  eapply old_accum_inversion in H; try eassumption.
  destruct H as [[G1 G2] |].
  Case "eval_accum_2". subst. eapply H2; try eassumption. admit. (* inversion doesn't seem enough *)  
*)
*)

(* We could try to prove that the old eval_test_coll_2-3_X rules are derivable *)

Lemma eval_test_coll_2_derived : forall e T v v',
      Eval e (Return v) ->
      is_C v = true ->
      v_mem v' v = true ->
      Eval (E_In (E_Value v') T) Error ->
      Eval (E_In e (T_Coll T)) Error.
Proof.
  intros. eapply eval_test_coll_2; try eassumption.
Admitted. (* XXX some induction needed *)

Lemma eval_test_coll_3_1_derived : forall e T v,
      Eval e (Return v) ->
      is_C v = true ->
      (forall v', v_mem v' v = true -> Eval (E_In (E_Value v') T) (Return v_tt)) ->
      Eval (E_In e (T_Coll T)) (Return v_tt).
Proof.
  intros. eapply eval_test_coll_2; try eassumption.
Admitted. (* XXX some induction needed *)

Lemma eval_test_coll_3_2_derived : forall e T v v''',
      Eval e (Return v) ->
      is_C v = true ->
      (* XXX this gives weak induction principle ... because of the "\/" or "exists" ?
      Warning: Ignoring recursive call
      *)
      (forall v', v_mem v' v = true -> exists v'',
          (Eval (E_In (E_Value v') T) (Return v'') /\ In_Logical v'' = true)
      ) ->
      v_mem v''' v = true ->
      Eval (E_In (E_Value v''') T) (Return v_ff) ->
      Eval (E_In e (T_Coll T)) (Return v_ff).
Proof.
  intros. eapply eval_test_coll_2; try eassumption.
Admitted. (* XXX some induction needed *)

(* We prove that the four old eval_test_refine_X rules are now derivable *)

Lemma eval_test_refine_1_derived : forall e1 x T e2 v,
      Eval e1 (Return v) ->
      Eval (E_In (E_Value v) T) Error ->
      Eval (E_In e1 (T_Refine x T e2)) Error.
Proof.
  intros. eapply eval_test_refine; try eassumption.
  eapply eval_bin_op_1_1. assumption.
Qed.

Lemma eval_test_refine_2_1_derived : forall e1 x T e2 v1 v',
      Eval e1 (Return v1) ->
      Eval (E_In (E_Value v1) T) (Return v') ->
      Eval (subst_exp e2 v1 x) Error ->
      Eval (E_In e1 (T_Refine x T e2)) Error.
Proof.
  intros. eapply eval_test_refine; try eassumption.
  eapply eval_bin_op_1_2; eassumption.
Qed.

Lemma eval_test_refine_2_2_derived : forall e1 x T e2 v1 v' v2,
      Eval e1 (Return v1) ->
      Eval (E_In (E_Value v1) T) (Return v') ->
      Eval (subst_exp e2 v1 x) (Return v2) ->
      In_Logical v' && In_Logical v2 = false ->
      Eval (E_In e1 (T_Refine x T e2)) Error.
Proof.
  intros. eapply eval_test_refine; try eassumption.
  eapply eval_bin_op_2; try eassumption.
  admit. (* XXX need definition of bin_op_eval to prove this *)
Admitted.

Lemma eval_test_refine_3_derived : forall e1 x T e2 v1 v' v2,
      Eval e1 (Return v1) ->
      Eval (E_In (E_Value v1) T) (Return v') ->
      Eval (subst_exp e2 v1 x) (Return v2) ->
      In_Logical v' && In_Logical v2 = true ->
      Eval (E_In e1 (T_Refine x T e2)) (Return (O_And v' v2)).
Proof.
  intros. eapply eval_test_refine; try eassumption.
  eapply eval_bin_op_3; try eassumption.
  admit. (* XXX need definition of bin_op_eval to prove this *)
Admitted.


(* XXX no longer used *)
Lemma v_plus_empty : forall c1 c2,
  v_empty c1 = v_tt ->
  v_plus c1 c2 =  c2.
Admitted.

(* XXX no longer used *)
Lemma v_plus_empty_2 : forall c1 c2,
  v_empty c2 = v_tt ->
  v_plus c1 c2 =  c1.
Admitted.

(* XXX no longer used *)
Lemma v_empty_plus_1 : forall c1 c2,
  v_empty (v_plus c1 c2) = v_tt ->
  v_empty c1 = v_tt.
Admitted.

(* XXX no longer used *)
Lemma v_empty_plus_2 : forall c1 c2,
  v_empty (v_plus c1 c2) = v_tt ->
  v_empty c2 = v_tt.
Admitted.

(* XXX no longer used *)
Lemma v_mem_plus_true : forall v c1 c2,
  v_mem v (v_plus c1 c2) = true ->
  v_mem v c1 = true \/ v_mem v c2 = true.
Admitted.

(* XXX no longer used *)
Lemma v_mem_plus : forall v c1 c2,
  v_mem v (v_plus c1 c2) = v_mem v c1 || v_mem v c2.
Admitted.

(* XXX no longer used *)
Definition count c := List.length (proj1_sig (out_C c)).

(* XXX no longer used *)
Lemma count_zero : forall c,
  is_C c = true ->
  count c = 0 ->
  v_empty c = v_tt.
Admitted.

(* XXX no longer used *)
Lemma count_empty : forall c,
  is_C c = true ->
  v_empty c = v_tt ->
  count c = 0.
Admitted.

(* XXX no longer used *)
Lemma v_empty_no_mem : forall c v,
  is_C c = true ->
  v_empty c = v_tt ->
  v_mem v c = true ->
  False.
Admitted.

(* XXX no longer used *)
Lemma proj1_sig_empty : forall c,
  is_C c = true ->
  v_empty c = v_tt ->
  (proj1_sig (out_C c)) = nil.
Admitted.

(* XXX no longer used *)
(* wishfull thinking ;) *)
Lemma permutation_cons_v_remove : forall c v vs,
  is_C c = true ->
  v_mem v c = true ->
  Permutation vs (proj1_sig (out_C (v_remove v c))) ->
  Permutation (proj1_sig v :: vs) (proj1_sig (out_C c)).
Admitted.



Lemma is_pure_subst : forall e v x, is_pure e = true -> is_pure (subst_exp e v x) = true.
Admitted.


(* Proving this requires functional extensionality
-- or, we need to restrict this to R (env_update ...)
*)
Lemma env_update_swap : forall env x1 x2 v1 v2,
  x1 <> x2 ->
  (env_update (env_update env x1 v1) x2 v2) =
  (env_update (env_update env x2 v2) x1 v1).
Admitted.

Lemma env_update_shadow_extensional : forall y v1 v2 env,
  (env_update (env_update env y v1) y v2) = env_update env y v2.
Admitted.


(* Note: alg_pure_result_uniqueness only says that every pure expression has at most(!) one result (confluence)
  - this doesn't give us the existance of a result
  - for Purity we need both existance and uniquenss
- The existance of a result (also) depends on termination(!)
- This proof might be easier if one involves the small-step semantics
  + use the theorems relating the small-step with the big step semantics
  + of the lemma saying that algorithmically pure expressions are terminating
*)
Lemma alg_purity_result_existance : forall e,
  is_pure e = true ->
  fv_exp e = nil -> (* we also need that e is closed since we have no evaluation rule for variables *)
  exists r, Eval e r.
Proof.
  (Exp_cases (induction e) Case); simpl; intros.
  Case "E_Var". inversion H0. (* contradiction *)
  Case "E_Value". eexists. apply eval_value.
  admit. admit.
  Case "E_If".
    apply app_eq_nil in H0. destruct H0. apply app_eq_nil in H1. destruct H1.
    rewrite andb_true_iff in H. destruct H. rewrite andb_true_iff in H. destruct H.
    destruct IHe1 as [r1 HH1]; try assumption.
    destruct IHe2 as [r2 HH2]; try assumption.
    destruct IHe3 as [r3 HH3]; try assumption.
    destruct r1 as [| v1].
    SCase "r1 = Error". exists Error. eapply eval_cond_1; try eassumption; try (intro Hc; inversion Hc).
    SCase "r1 = Return v1". admit. (* ... lost of sub-cases left *)
  admit. admit. admit. admit. admit. admit.
  Case "E_App". inversion H. (* contradiction -- only easy because in the current
                                                 formalization applications are not pure *)
Admitted.


Lemma fv_exp_let_seq : forall x y e3 vs,
  fv_exp (let_seq y (map (fun v : Value => subst_exp e3 v x) vs) (E_Var y))
  = List.remove eq_str_dec x (fv_exp e3).
Admitted.

Lemma vb_to_vs_out_c_nil_v_zero : forall v1,
  is_C v1 = true ->
  vb_to_vs (out_C v1) = nil ->
  v1 = v_zero.
Admitted.

Lemma res_acc_fold_cons_eq : forall f a rv rvb' Hvb Hv Hvb',
  res_acc_fold f (exist _ (rv :: rvb') Hvb) a = 
    match a with
    | Return va => res_acc_fold f (exist _ rvb' Hvb') (f va (exist _ rv Hv))
    | Error => Error
    end.
Admitted.

Lemma res_acc_fold_error : forall f vb,
  res_acc_fold f vb Error = Error.
Admitted.

(* also needed for the accumulate case of L30 *)
Lemma R_let_seq_res_acc : forall y x e3 v1 v2 env,
  y <> x ->
  is_C v1 = true ->
  R (let_seq y (map (fun v : Value => subst_exp e3 v x)
                   (vb_to_vs (out_C v1))) (E_Var y))
             (env_update env y v2) =
  res_accumulate (fun v2 v3 : Value => R e3 (env_update (env_update env x v3) y v2)) v1 v2.
Proof.
  intros. unfold res_accumulate.
  remember (vb_to_vs (out_C v1)) as vs. generalize dependent v1. generalize dependent v2.
  induction vs as [| v vs']; intros.
  Case "vs = nil". symmetry in Heqvs.
    assert (v1 = v_zero). eapply vb_to_vs_out_c_nil_v_zero; try eassumption.
    subst. unfold out_C, v_zero, res_acc_fold.
    simpl. rewrite WfExtensionality.fix_sub_measure_eq_ext. simpl.
    rewrite env_update_eq. reflexivity.
  Case "vs = v :: vs'". rewrite H0.
    remember (out_C v1) as vb. destruct vb as [rvb Hvb].
    destruct v as [rv Hv].
    assert(rvb = rv :: proj1_sig (out_C (v_remove (exist _ rv Hv) v1))). admit. (* XXX *)
    assert(vs' = vb_to_vs (out_C (v_remove (exist (fun x : RawValue => Normal x) rv Hv) v1))). admit. (* XXX *)
    subst.
    assert(Sorted_vb (remove1 eq_rval_dec rv (proj1_sig (out_C v1))) /\
      IndAll Normal (remove1 eq_rval_dec rv (proj1_sig (out_C v1)))) as Hvb'. admit. (* XXX *)
    rewrite res_acc_fold_cons_eq with (Hv := Hv) (Hvb' := Hvb'). simpl.
    rewrite R_subst. rewrite env_update_swap; try assumption.
    remember (R e3 (env_update (env_update env x
           (exist (fun x0 : RawValue => Normal x0) rv Hv)) y v2)) as r.
    destruct r as [| v0 ]; unfold LBind; simpl.
      SCase "r = Error". rewrite res_acc_fold_error. reflexivity.
      SCase "r = Return v0".
        rewrite env_update_shadow_extensional.
        assert(is_C (v_remove (exist _ rv Hv) v1) = true) as H1. apply is_C_v_remove. assumption.
        erewrite IHvs' with (v3 := v_remove (exist _ rv Hv) v1); try trivial.
        rewrite H1.
      assert(forall v1 rv Hv H,
        (out_C (v_remove (exist (fun x0 : RawValue => Normal x0) rv Hv) v1)) =
        (exist (fun vb : list RawValue => Sorted_vb vb /\ IndAll Normal vb))
          (remove1 eq_rval_dec rv (proj1_sig (out_C v1))) H) as Hrem.
        intros. unfold v_remove. unfold out_C at 1. simpl. apply eq_exist.
      erewrite Hrem. reflexivity.
Qed.

Lemma L30_alg_purity : forall e r,
  fv_exp e = nil ->
  is_pure e = true ->
  Eval e r ->
  R e env_empty = r.
Proof.
  intros e r H1 H2 H3.
  (eval_cases (induction H3) Case);
    simpl in *; try reflexivity.
  admit. admit. admit. admit. admit. admit. admit.
  admit. admit. admit. admit. admit. admit. admit.
  admit. admit. admit. admit. admit. admit. admit. admit.
  Case "eval_accum_1_1".
    rewrite IHEval. unfold LBind. reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    repeat rewrite andb_true_iff in H2. destruct H2 as [[[H2' H2''] H2'''] H2'''']. assumption.
  Case "eval_accum_1_2".
    rewrite IHEval. unfold LBind. simpl.
    destruct (Proper (R e2 env_empty)); try trivial.
    simpl. unfold res_accumulate. rewrite H. reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    repeat rewrite andb_true_iff in H2. destruct H2 as [[[H2' H2''] H2'''] H2'''']. assumption.
  Case "eval_accum_lets".
    apply app_eq_nil in H1. destruct H1 as [H1' H1]. apply app_eq_nil in H1. destruct H1 as [H1'' H1'''].
    repeat rewrite andb_true_iff in H2. destruct H2 as [[[H2' H2''] H2'''] HH].
    rewrite IHEval1; try assumption.
    rewrite <- IHEval2. unfold LBind. simpl.
      destruct (Proper (R e2 env_empty)); simpl; try trivial.
      clear IHEval1 IHEval2 H3_0 H3_ H1' H2' e1.
      repeat setoid_rewrite <- forall_bool_true_iff in HH.
      setoid_rewrite <- syn_beq_res_true_iff in HH.
      rewrite <- R_let_seq_res_acc.
      apply split_perm_adj_trans in H0.
      destruct H0 as [H0 | [H0 | H0]].
      SCase "vs = vb_to_vs (out_C v1)". subst. reflexivity.
      SCase "AdjTrans Value vs (vb_to_vs (out_C v1))".
        assert(forall v0, forall env,
          R (let_seq y (map (fun v : Value => subst_exp e3 v x) (vb_to_vs (out_C v1)))
            (E_Var y)) (env_update env y v0) =
          R (let_seq y (map (fun v : Value => subst_exp e3 v x) vs)
            (E_Var y)) (env_update env y v0)).
        induction H0; intros v0 env.
        SSCase "adj_trans_skip".
          simpl. unfold LBind.
          destruct (R (subst_exp e3 x0 x) (env_update env y v0)); simpl; try reflexivity.
          apply IHAdjTrans.
        SSCase "adj_trans_swap". rename y0 into x1.
          simpl. unfold LBind. simpl.
          repeat rewrite R_subst.
          repeat rewrite env_update_shadow_extensional.
          specialize (HH env x0 x1 v0).
          unfold LBind in HH.
          assert(x <> y). admit.
          pattern(env_update (env_update env x x0) y v0) in HH.
          rewrite env_update_swap in HH; try assumption.
          pattern(env_update (env_update env x x1) y v0) in HH.
          rewrite env_update_swap in HH; try assumption.
          destruct (Proper (R e3 (env_update (env_update env y v0) x x0))).
          SSSCase "Proper".
            destruct (Proper (R e3 (env_update (env_update env y v0) x x1))).
            SSSSCase "Proper".
              simpl in *.
              pattern(env_update (env_update env x x0) y
                (out_V (R e3 (env_update (env_update env y v0) x x1)))) in HH.
              rewrite env_update_swap in HH; try assumption.
              pattern(env_update (env_update env x x1) y
                (out_V (R e3 (env_update (env_update env y v0) x x0)))) in HH.
              rewrite env_update_swap in HH; try assumption.
              repeat rewrite HH. reflexivity.
            SSSSCase "not Proper".
              simpl in *.
              pattern(env_update (env_update env x x1) y
                (out_V (R e3 (env_update (env_update env y v0) x x0)))) in HH.
              rewrite env_update_swap in HH; try assumption.
              rewrite <- HH. simpl. reflexivity.
          SSSCase "not Proper".
            simpl in *.
            destruct (Proper (R e3 (env_update (env_update env y v0) x x1)));
              simpl in *; try reflexivity.
            pattern(env_update (env_update env x x0) y
              (out_V (R e3 (env_update (env_update env y v0) x x1)))) in HH.
            rewrite env_update_swap in HH; try assumption.
            rewrite HH. reflexivity.
        apply H1.
      SCase "sequence of more than 1 adj transpositions".
        (* simple generalization of the case above
          (induction on the size of the transposition chain) *)
        admit.
    admit. (* x <> y *) assumption. admit. admit.
  admit. admit. admit. admit. admit. admit. admit.
  admit. admit. admit.
Admitted.

Lemma alg_pure_result_uniqueness : forall e r1 r2,
  fv_exp e = nil ->
  is_pure e = true ->
  Eval e r1 ->
  Eval e r2 ->
  r1 = r2.
Proof.
  intros.
  erewrite <- L30_alg_purity with (r := r1); try eassumption.
  erewrite <- L30_alg_purity with (r := r2); try eassumption.
  reflexivity.
Qed.



(* Proving Lemma 30 from the paper *)

Variable Pure : Exp -> Prop.

Axiom pure_unique_result : forall e, Pure e -> exists! r, Eval e r.

Lemma pure_at_most_one_result : forall e r1 r2, Pure e -> Eval e r1 -> Eval e r2 -> r1 = r2.
Proof.
  intros. apply pure_unique_result in H. unfold unique in H.
  destruct H as [r [H' H'']]. apply H'' in H0. apply H'' in H1. subst. assumption.
Qed.

Lemma pure_exists_result : forall e, Pure e -> exists r, Eval e r.
Proof. intros. apply pure_unique_result in H. destruct H as [r [H' H'']]. exists r. apply H'. Qed. 

Axiom pure_subst : forall e v x,
  Pure e ->
  Pure (subst_exp e v x).

(* Currently using this for subterms
Wrong claim (FIXED!): If e is pure then all sub-expressions of e are pure
Sub-expressions of a Pure expressions don't have to be Pure
- it could be that the sub-expression is unreachable
(e.g. conditionals only execute one of the branches;
the body of accumulates is only executed if the collection is non-empty)
- it could be that the value is later thrown away
(calling a constant function with an impure argument is ok)
- it could be that it simply doesn't matter for the final result
(true || pick {true, false} always produces true)
Conclusion: Purity is (curently) *not* a subterm property!
(algorithmic purity is!)
*)
Lemma pure_magic : forall e, Pure e.
Admitted.

(* more interesting facts about purity *)
Lemma pure_interesting : forall e, Pure e.
Admitted.


Lemma fv_exp_subst : forall e v x,
  fv_exp (subst_exp e v x) = List.remove eq_str_dec x (fv_exp e).
Admitted.

Lemma swap_remove : forall x y A,
  List.remove eq_str_dec y (List.remove eq_str_dec x A) =  
  List.remove eq_str_dec x (List.remove eq_str_dec y A).
Admitted.

Lemma v_empty_true_eq : forall v,
  v_empty v = v_tt ->
  v = exist _ (C nil) C_nil_normal.
Admitted.

Lemma andb_false_iff : forall b1 b2,
  b1 && b2 = false <-> b1 = false \/ b2 = false.
Proof. intros. destruct b1; destruct b2; split; intros; tauto. Qed.

Lemma orb_false_iff : forall b1 b2,
  b1 || b2 = false <-> b1 = false /\ b2 = false.
Proof. intros. destruct b1; destruct b2; split; intros; tauto. Qed.

Lemma negb_false_iff : forall b,
  negb b = false <-> b = true.
Proof. intros. destruct b; split; intros; simpl; auto. Qed.

Lemma negb_true_iff : forall b,
  negb b = true <-> b = false.
Proof. intros. destruct b; split; intros; simpl; auto. Qed.

Lemma orb_true_iff : forall b1 b2,
  b1 || b2 = true <-> b1 = true \/ b2 = true.
Proof. intros. destruct b1; destruct b2; split; intros; tauto. Qed.

Lemma implb_true_iff : forall b1 b2,
  implb b1 b2 = true <-> (b1 = true -> b2 = true).
Proof. intros. destruct b1; destruct b2; split; intros; tauto. Qed.

Lemma syn_beq_val_true : forall v1 v2,
  syn_beq_val v1 v2 = true -> v1 = v2.
Admitted.

Lemma syn_beq_val_false : forall v1 v2,
  syn_beq_val v1 v2 = false -> v1 <> v2.
Admitted.

Lemma logical_only_two : forall v,
  In_Logical v = true ->
  v = v_tt \/ v = v_ff.
Admitted.

Lemma not_logical : forall v,
  In_Logical v = false ->
  v <> v_tt /\ v <> v_ff.
Admitted.

Lemma logical_not_true : forall v,
  In_Logical v = true ->
  v <> v_tt ->
  v = v_ff.
Admitted.

Lemma exists_false_iff : forall X,
  ex (fun _ : X => False) <-> False.
Proof. intro X. split. intro H. destruct H. apply H. intro H. inversion H. Qed.

Lemma no_fields_eentity : forall v,
  is_E v = true ->
  (forall l : string, ~v_has_field l v = true) -> v = v_eempty.
Proof.
  intros v H1 H2.
  destruct v as [[[i | s | b |] | svm | vb] pv]; simpl in *; try (inversion H1; fail).
  unfold v_eempty.
  destruct svm as [| [l0 v0]]. apply eq_exist.
  apply False_ind. specialize (H2 l0). unfold v_has_field in H2. simpl in H2.
  destruct (eq_str_dec l0 l0). simpl in H2. apply H2. reflexivity. apply n. reflexivity.
Qed.

Variable v_eremove : string -> Value -> Value.

Lemma eupdate_eremove : forall l v,
  v_eupdate l (v_dot l v) (v_eremove l v) = v.
Admitted.

Lemma is_E_eremove : forall l v,
  is_E v = true -> is_E (v_eremove l v) = true.
Admitted.

Lemma has_field_eremove_neq : forall l li v,
  li <> l -> v_has_field li (v_eremove l v) = v_has_field li v.
Admitted.

Lemma has_field_eremove_true : forall l li v,
  v_has_field li (v_eremove l v) = true ->
  v_has_field li v = true /\ li <> l.
Admitted.

Lemma dot_eremove_neq : forall l li v,
  v_dot li (v_eremove l v) = v_dot li v.
Admitted.


Lemma Pure_Eval_les_seq_unique : forall x e1 y e2 e3 r1 r2 vs' vs'' v1,
  Pure (E_Acc x e1 y e2 e3) ->
  Eval e1 (Return v1) ->
  is_C v1 = true ->
  Permutation vs' (vb_to_vs (out_C v1)) ->
  Permutation vs'' (vb_to_vs (out_C v1)) ->
(*  (Permutation vs' vs'') -> *)
  Eval (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs') (E_Var y))) r1 ->
  Eval (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs'') (E_Var y))) r2 ->
  r1 = r2.
Proof.
  intros. 
  assert (Eval (E_Acc x e1 y e2 e3) r1) as H'.
    eapply eval_accum_lets with (vs := vs'); eassumption.
  assert (Eval (E_Acc x e1 y e2 e3) r2) as H''.
    eapply eval_accum_lets with (vs := vs''); eassumption.
  eapply pure_at_most_one_result in H'; try eassumption. subst. reflexivity.
Qed.

Lemma Pure_Eval_les_seq_change_order : forall x e1 y e2 e3 r vs' vs'' v1,
  Pure (E_Acc x e1 y e2 e3) ->
  Eval e1 (Return v1) ->
  is_C v1 = true ->
  Permutation vs' (vb_to_vs (out_C v1)) ->
  Permutation vs'' (vb_to_vs (out_C v1)) ->
(*  (Permutation vs' vs'') -> *)
  Eval (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs') (E_Var y))) r ->
  Eval (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs'') (E_Var y))) r.
Proof.
  intros.
(*
  assert (Eval (E_Acc x e1 y e2 e3) r) as H'.
    eapply eval_accum_lets with (vs := vs'); eassumption.
*)
  assert (Pure (E_Let y e2 (let_seq y (map (fun v : Value => subst_exp e3 v x) vs'') (E_Var y)))).
    admit. (* since the small-step operational semantics preserves purity *)
  apply pure_exists_result in H5. destruct H5 as [r' H5].
  assert (r = r').
    eapply Pure_Eval_les_seq_unique with (vs' := vs') (vs'' := vs''); try eassumption.
  subst. assumption.
Qed.

(* Andy's idea, doesn't seem to save us from well-founded induction
(just moves it to a different place) *)
Inductive Eval_NF : Exp -> Result -> Prop :=
  (* ... *)
  | eval_nf_accum_1_1 : forall x y e1 e2 e3,
      Eval_NF e1 Error ->
      Eval_NF (E_Acc x e1 y e2 e3) Error
  | eval_nf_accum_1_2 : forall x y e1 e2 e3 v1,
      Eval_NF e1 (Return v1) ->
      is_C v1 = false ->
      Eval_NF (E_Acc x e1 y e2 e3) Error
  (* the only rule that changes -- we use a fixed order in accumulates, that matches the one in the model *)
  | eval_nf_accum_lets : forall x e1 y e2 e3 r v1,
      Eval_NF e1 (Return v1) ->
      is_C v1 = true ->
      Eval_NF (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) (vb_to_vs (out_C v1))) (E_Var y))) r ->
      Eval_NF (E_Acc x e1 y e2 e3) r
  (* ... *)
.

Lemma Pure_Eval_entails_Eval_NF : forall e r,
  Pure e ->
  Eval e r ->
  Eval_NF e r.
Proof.
  intros.
  (eval_cases (induction H0) Case).
  admit. admit. admit. admit. admit. admit. admit. admit. admit. admit.
  admit. admit. admit. admit. admit. admit. admit. admit. admit. admit.
  admit. admit. admit. admit.
  Case "eval_accum_1_1".
    eapply eval_nf_accum_1_1. apply IHEval. apply pure_magic.
  Case "eval_accum_1_2".
    eapply eval_nf_accum_1_2; try eassumption. apply IHEval. apply pure_magic.
  Case "eval_accum_lets".
    eapply eval_nf_accum_lets. apply IHEval1. apply pure_magic. assumption.
    assert (Permutation (vb_to_vs (out_C v1)) (vb_to_vs (out_C v1))). apply Permutation_refl.
    assert(Eval
      (E_Let y e2
        (let_seq y
          (map (fun v : Value => subst_exp e3 v x) (vb_to_vs (out_C v1))) (E_Var y))) r).
      eapply Pure_Eval_les_seq_change_order with (vs' := vs); eassumption.
    (* almost there, still IHEval2 is not what I need
       Q1: How about induction on the maximal depth of the derivation
            (bring back the additional integer index)?
       A1: Doesn't work since Pure_Eval_les_seq_change_order does NOT preserve the depth (or the size) of the derivation.
           For instance if we reorder (let x1 = stuck in let x2 = v in z) to
           (let x1 = v in let x2 = stuck in z) the depth of the derivation increases.

       Q2: How about well-founded induction using some complex ordering on expressions (#accum, size)
       A2: This Won't work. eval_accum_lets multiplicates e3 n times
              if e3 contains accumulates the expression can get much "bigger"

       Q3: How about well-founded induction using (maximum number of nested accumulates, size) ?
       A3: This seems to decrease even for eval_accum_lets.

       Q4: Wild idea: How about doing well-founded induction on the ->^* relation
       (it is well-founded on Pure expressions -- since they are terminating)
    *)
    admit.
Admitted.





(* Dominique's idea *)
Inductive Eval_D : Exp -> Result -> Prop :=
  (* ... *)
  | eval_d_accum_1_1 : forall x y e1 e2 e3,
      Eval_D e1 Error ->
      Eval_D (E_Acc x e1 y e2 e3) Error
  | eval_d_accum_1_2 : forall x y e1 e2 e3 v1,
      Eval_D e1 (Return v1) ->
      is_C v1 = false ->
      Eval_D (E_Acc x e1 y e2 e3) Error
  (* the only rule that changes -- we use any possible order in the accumulates and 
                                   then we non-deterministically pick a result 
     -- this makes the induction hypothesis for this judgement much stronger                            
  *)
  | eval_d_accum_lets : forall x e1 y e2 e3 rs v1 ri,
      Eval_D e1 (Return v1) ->
      is_C v1 = true ->
      (forall r, In r rs -> exists vs, (Permutation vs (vb_to_vs (out_C v1)))
          /\ Eval_D (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs) (E_Var y))) r) ->
      (forall vs, (Permutation vs (vb_to_vs (out_C v1))) -> exists r, In r rs 
          /\ Eval_D (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs) (E_Var y))) r) ->
      In ri rs -> (* non-deterministic pick *)
      Eval_D (E_Acc x e1 y e2 e3) ri
  (* ... *)
.

Axiom Eval_D_ind_fixed
     : forall P : Exp -> Result -> Prop,
       (forall (x y : string) (e1 e2 e3 : Exp),
        Eval_D e1 Error -> P e1 Error -> P (E_Acc x e1 y e2 e3) Error) ->
       (forall (x y : string) (e1 e2 e3 : Exp) (v1 : Value),
        Eval_D e1 (Return v1) ->
        P e1 (Return v1) -> is_C v1 = false -> P (E_Acc x e1 y e2 e3) Error) ->
       (forall (x : string) (e1 : Exp) (y : string) (e2 e3 : Exp)
          (rs : list Result) (v1 : Value) (ri : Result),
        Eval_D e1 (Return v1) ->
        P e1 (Return v1) ->
        is_C v1 = true ->
        (forall r : Result,
         In r rs ->
         exists vs, (Permutation vs (vb_to_vs (out_C v1))) /\
           Eval_D (E_Let y e2 (let_seq y
                 (map (fun v : Value => subst_exp e3 v x) vs) (E_Var y))) r) ->
        (forall r : Result,
         In r rs ->
         exists vs, (Permutation vs (vb_to_vs (out_C v1))) /\
           P (E_Let y e2 (let_seq y
                 (map (fun v : Value => subst_exp e3 v x) vs) (E_Var y))) r) ->
        (forall vs : list Value,
         Permutation vs (vb_to_vs (out_C v1)) ->
         exists r, In r rs /\
           Eval_D (E_Let y e2 (let_seq y (map (fun v : Value => subst_exp e3 v x) vs) (E_Var y))) r) ->
        (forall vs : list Value,
         Permutation vs (vb_to_vs (out_C v1)) ->
         exists r, In r rs /\
           P (E_Let y e2 (let_seq y (map (fun v : Value => subst_exp e3 v x) vs) (E_Var y))) r) ->
         In ri rs -> P (E_Acc x e1 y e2 e3) ri) ->
       forall (e : Exp) (r : Result), Eval_D e r -> P e r.

(* The lemma in the paper also has fv_exp e = nil as a premise, but it doesn't seem necessary *)
Lemma Eval_D_entails_Eval : forall e r,
  Eval_D e r ->
  Eval e r.
Proof.
  intros.
  induction H using Eval_D_ind_fixed.
  (* XXX missing cases *)
  Case "eval_d_accum_1_1". apply eval_accum_1_1; eassumption.
  Case "eval_d_accum_1_2". eapply eval_accum_1_2; eassumption.
  Case "eval_d_accum_lets".
    specialize (H2 ri H5). destruct H2 as [vs [H21 H22]].
    eapply eval_accum_lets with (vs := vs); try eassumption.
  (* XXX missing cases *)
Qed.

Lemma Pure_Eval_D_les_seq_unique : forall x e1 y e2 e3 r1 r2 vs' vs'' v1,
  Pure (E_Acc x e1 y e2 e3) ->
  Eval_D e1 (Return v1) ->
  is_C v1 = true ->
  Permutation vs' (vb_to_vs (out_C v1)) ->
  Permutation vs'' (vb_to_vs (out_C v1)) ->
  Eval_D (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs') (E_Var y))) r1 ->
  Eval_D (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs'') (E_Var y))) r2 ->
  r1 = r2.
Proof. intros. apply Eval_D_entails_Eval in H4. apply Eval_D_entails_Eval in H5. apply Eval_D_entails_Eval in H0.
  eapply Pure_Eval_les_seq_unique with (vs' := vs') (vs'' := vs''); eassumption.
Qed.

Lemma L30_D : forall e r,
  fv_exp e = nil ->
  Pure e ->
  Eval_D e r ->
  R e env_empty = r.
Proof.
  intros.
  induction H1 using Eval_D_ind_fixed.
  Case "eval_d_accum_1_1". admit. (* as before *)
  Case "eval_d_accum_1_2". admit. (* as before *)
  Case "eval_d_accum_lets".
    simpl in *.
    apply app_eq_nil in H. destruct H. (* apply app_eq_nil in H3. destruct H3. *)
    assert(Pure e1). apply pure_magic.
    rewrite IHEval_D; try assumption. unfold LBind at 1. simpl.

    specialize (H6 (vb_to_vs (out_C v1)) (Permutation_refl _)). destruct H6 as [rj [H61 H62]].
    assert(ri = rj).
      remember H3 as H3'. clear HeqH3'.
      specialize (H3 ri H7). destruct H3 as [vs [H31 H32]].
      specialize (H3' rj H61). destruct H3' as [vs' [H31' H32']].
      eapply Pure_Eval_D_les_seq_unique with (vs' := vs) (vs'' := vs'); try eassumption.
    subst.
    rewrite fv_exp_let_seq in H62. rewrite swap_remove in H62.
    specialize (H62 H8).
    assert (Pure (E_Let y e2 (let_seq y (map (fun v : Value => subst_exp e3 v x)
                      (vb_to_vs (out_C v1))) (E_Var y)))) as H6.
      admit. (* since the small-step operational semantics preserves purity *)
    specialize (H62 H6).
    unfold LBind in *.
    destruct (Proper (R e2 env_empty)); simpl in *; try (subst; trivial; fail).
    rewrite R_let_seq_res_acc in H62; try assumption. admit. (* y <> x  -- syntactic accumulate invariant *)
Admitted. (* all other cases as below *)

(* this uses the fact that *all* executions starting with Pure expressions are terminating  *)
Lemma Pure_exists_Eval_D_result : forall e,
  Pure e -> exists r, Eval_D e r.
Admitted.

Lemma Pure_Eval_D_les_seq_change_order : forall x e1 y e2 e3 r vs' vs'' v1,
  Pure (E_Acc x e1 y e2 e3) ->
  Eval_D e1 (Return v1) ->
  is_C v1 = true ->
  Permutation vs' (vb_to_vs (out_C v1)) ->
  Permutation vs'' (vb_to_vs (out_C v1)) ->
(*  (Permutation vs' vs'') -> *)
  Eval_D (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs') (E_Var y))) r ->
  Eval_D (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs'') (E_Var y))) r.
Proof.
  intros.
  assert(Pure (E_Let y e2
     (let_seq y (map (fun v : Value => subst_exp e3 v x) vs'') (E_Var y)))) as H5.
    admit. (* small-step preserves purity *)
  apply Pure_exists_Eval_D_result in H5.
  destruct H5 as [r' H5].
  apply Eval_D_entails_Eval in H4.
  remember H5 as H5'. clear HeqH5'.
  apply Eval_D_entails_Eval in H5.
  apply Eval_D_entails_Eval in H0.
  assert(r = r').
    eapply Pure_Eval_les_seq_unique with (vs' := vs') (vs'' := vs''); try eassumption.
  subst. assumption.
Qed.

Lemma Pure_Eval_entails_Eval_D : forall e r,
  Pure e ->
  Eval e r ->
  Eval_D e r.
Proof.
  intros.
  (eval_cases (induction H0) Case).
  admit. admit. admit. admit. admit. admit. admit. admit. admit. admit.
  admit. admit. admit. admit. admit. admit. admit. admit. admit. admit.
  admit. admit. admit. admit.
  Case "eval_accum_1_1".
    eapply eval_d_accum_1_1. apply IHEval. apply pure_magic.
  Case "eval_accum_1_2".
    eapply eval_d_accum_1_2; try eassumption. apply IHEval. apply pure_magic.
  Case "eval_accum_lets".
    eapply eval_d_accum_lets with (rs := r :: nil).
    apply IHEval1. apply pure_magic. assumption.
    SCase "TS1".
      intros. exists vs. split. assumption. 
      simpl in H2. destruct H2; try (inversion H2; fail). subst. apply IHEval2.
      admit. (* small-step preserves purity *)
    SCase "TS2".
      intros. exists r. split. simpl. auto.
      eapply Pure_Eval_D_les_seq_change_order with (vs' := vs); try eassumption.
      apply IHEval1. apply pure_magic.
      apply IHEval2.
      admit. (* small-step preserves purity *)
    simpl. auto.
Admitted.

(* The old proof of L30 -- everything works other than the eval_accum_lets case (proved above in L30_D) *)
(* Note: this proof relies on Pure being a subterm property, which was not assumed in the paper -- this is now fixed *)
Lemma L30 : forall e r,
  fv_exp e = nil ->
  Pure e ->
  Eval e r ->
  R e env_empty = r.
Proof.
  intros e r H1 H2 H3.
  (eval_cases (induction H3) Case);
    simpl in *; try reflexivity.
  Case "eval_un_op_1". simpl in *. rewrite IHEval; try trivial. apply pure_magic.
  Case "eval_un_op_2".
    destruct op. simpl in *. rewrite IHEval; try trivial. unfold LBind; simpl.
    remember (In_Logical v) as ilv. symmetry in Heqilv. destruct ilv.
    SCase "In_Logical v = true". contradict H.
      unfold In_Logical in Heqilv. rewrite andb_true_iff in Heqilv. destruct Heqilv as [H' H''].
      destruct v as [[g | svm | vb] pv]; try (inversion H'; fail).
      destruct g as [i | s | b |]; try (inversion H''; fail). destruct b.
      SSCase "v = v_tt". exists v_ff. erewrite eq_exist. apply uoe_neg_tt.
      SSCase "v = v_ff". exists v_tt. erewrite eq_exist. apply uoe_neg_ff.
    SCase "In_Logical v = false". trivial.
    apply pure_magic.
  (* these all proved on paper *)
  admit. admit. admit. admit. admit. admit. admit. admit. admit. admit.
  Case "eval_entity_1".
     (* XXX there is a missing induction hypothesis ... induction principle is weak ! *)
    assert(forall (l1 : string) (e1 : Exp),
      In (l1, e1) les1 -> fv_exp e1 = nil -> Pure e1 -> exists vi, R e1 env_empty = Return vi) as IHEval'.
      admit. (* this is much easier than proving a better induction principle (48 cases) *)
    induction les1 as [| [l1 e1] les1'].
    SCase "les1 = nil". simpl. rewrite IHEval. unfold LBind. reflexivity.
      simpl in H1. apply app_eq_nil in H1. destruct H1. assumption. apply pure_magic.
    SCase "les1 = (l1,e1) :: les1'".
      simpl in *. apply app_eq_nil in H1. destruct H1.
      edestruct IHEval' as [v1 IHEval1]. left. reflexivity. assumption. apply pure_magic.
      rewrite IHEval1. unfold LBind at 1. simpl. rewrite IHles1'.
      reflexivity. assumption. apply pure_magic.
      intros. eapply H. right. eassumption.
      intros. eapply IHEval'. right. eassumption.
      assumption. assumption.
  Case "eval_entity_2".
     (* XXX there is a missing induction hypothesis ... induction principle is weak ! *)
    assert(forall (li : string) (ei : Exp),
      In (li, ei) les -> fv_exp ei = nil -> Pure ei -> exists vi,
        R ei env_empty = Return vi /\ v_has_field li v = true /\ v_dot li v = vi).
      admit. (* this is much easier than proving a better induction principle (48 cases) *)
    generalize dependent v.
    induction les as [| [l e] les']; intros; simpl in *.
    SCase "les=nil". setoid_rewrite exists_false_iff in H3.
      apply no_fields_eentity in H3; try trivial. rewrite H3. reflexivity.
    SCase "les=(l,e)::les'".
      (* XXX there is an implicit assumption that l doesn't appear in les' -- make explicit!? *)
      assert(forall li ei, In (li,ei) les' -> li <> l). admit.
      apply app_eq_nil in H1. destruct H1.
      specialize (IHles' H6 (pure_magic (E_Entity les'))). remember H4 as H4'. clear HeqH4'.
      specialize (H4 l e). destruct H4 as [v4 [H41 [H42 H43]]]. left. reflexivity. assumption. apply pure_magic.
      rewrite H41.
      erewrite IHles' with (v_eremove l v). unfold LBind; simpl.
      rewrite <- H43. rewrite eupdate_eremove. reflexivity.
      apply is_E_eremove. assumption.
      intros li ei Hi. specialize (H0 li ei). destruct H0 as [vi [H01 [H02 H03]]]; try auto.
        exists vi. split. auto. split.
        rewrite has_field_eremove_neq. assumption. eapply H5. eassumption.
        rewrite dot_eremove_neq. assumption.
      intros li Hi. apply has_field_eremove_true in Hi. destruct Hi as [Hi1 Hi2].
        specialize (H3 li Hi1). destruct H3 as [ei H3].
        destruct H3. inversion H3. subst. apply False_ind. apply Hi2. reflexivity.
        exists ei. apply H3.
      intros li ei Hin Hfv Hp. remember Hin as Hin'. clear HeqHin'. apply H5 in Hin.
        specialize (H4' li ei). destruct H4' as [vi [H4'1 [H4'2 H4'3]]]. right. assumption. assumption. assumption. 
        exists vi. split. assumption. split. rewrite has_field_eremove_neq; assumption.
        rewrite dot_eremove_neq; assumption.
  (* these all proved on paper *)
  admit. admit. admit. admit. admit. admit. admit. admit.
  Case "eval_accum_1_1".
    rewrite IHEval. reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    apply pure_magic.
  Case "eval_accum_1_2".
    rewrite IHEval. unfold LBind. destruct (Proper (R e2 env_empty)); simpl; try trivial.
    unfold res_accumulate. rewrite H. reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    apply pure_magic.
  Case "eval_accum_lets". (* XXX TODO: The interesting case *)
    (* Attempt: *)
    apply app_eq_nil in H1. destruct H1. (* apply app_eq_nil in H3. destruct H3. *)
    assert(Pure e1). apply pure_magic.
    rewrite IHEval1; try assumption. unfold LBind at 1. simpl.
    assert(Eval (E_Let y e2 (let_seq y (map (fun v : Value => subst_exp e3 v x) (vb_to_vs (out_C v1))) (E_Var y))) r).
      eapply Pure_Eval_les_seq_change_order with (vs' := vs); try eassumption. apply Permutation_refl.
    (* XXX the problem is that we can't use the induction hypothesis for this changed-order eval *)
    assert(fv_exp e2 ++
          List.remove eq_str_dec y
            (fv_exp
               (let_seq y (map (fun v : Value => subst_exp e3 v x) (vb_to_vs (out_C v1)))
                  (E_Var y))) = nil ->
          Pure
            (E_Let y e2
               (let_seq y (map (fun v : Value => subst_exp e3 v x) (vb_to_vs (out_C v1)))
                  (E_Var y))) ->
          LBind (R e2 env_empty)
            (fun y0 : Value =>
             R
               (let_seq y (map (fun v : Value => subst_exp e3 v x) (vb_to_vs (out_C v1)))
                  (E_Var y)) (env_update env_empty y y0)) = r
    ) as IHEval2'. admit. clear IHEval2. (* XXX let's *MAGICALLY* assume we can get a better induction for now *)
    rewrite fv_exp_let_seq in IHEval2'. rewrite swap_remove in IHEval2'.
    specialize (IHEval2' H3).
    assert (Pure (E_Let y e2 (let_seq y (map (fun v : Value => subst_exp e3 v x)
                      (vb_to_vs (out_C v1))) (E_Var y)))) as H6.
      admit. (* since the small-step operational semantics preserves purity *)
    specialize (IHEval2' H6).
    unfold LBind in *. destruct (Proper (R e2 env_empty)); simpl in *; try trivial.
    rewrite R_let_seq_res_acc in IHEval2'; try assumption. admit. (* y <> x  -- syntactic accumulate invariant *)
  Case "eval_test_wrong".
    rewrite IHEval. unfold LBind. simpl. reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    apply pure_magic.
  Case "eval_test_any".
    rewrite IHEval. unfold LBind. simpl. reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    apply pure_magic.
  Case "eval_test_integer".
    rewrite IHEval. unfold LBind. simpl. destruct (In_Integer v); reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    apply pure_magic.
  Case "eval_test_text".
    rewrite IHEval. unfold LBind. simpl. destruct (In_Text v); reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    apply pure_magic.
  Case "eval_test_logical".
    rewrite IHEval. unfold LBind. simpl. destruct (In_Logical v); reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    apply pure_magic.
  Case "eval_test_entity_1". apply app_eq_nil in H1. destruct H1.
    rewrite IHEval1; try assumption. unfold LBind in *. simpl. rewrite H. simpl.
    rewrite <- IHEval2. simpl. reflexivity.
    simpl. assumption.
    apply pure_interesting.
    apply pure_magic.
  Case "eval_test_entity_2".
    rewrite IHEval. unfold LBind. simpl. rewrite H. reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    apply pure_magic.
  Case "eval_test_coll_1".
    rewrite IHEval. unfold LBind. simpl. rewrite H. reflexivity.
    apply app_eq_nil in H1. destruct H1. assumption.
    apply pure_magic.
  Case "eval_test_coll_2". admit.
(* old rules
  Case "eval_test_coll_2".
    apply app_eq_nil in H1. destruct H1.
    rewrite IHEval1. unfold LBind. simpl. rewrite H.
    remember (exists_bool (fun x : Value => v_mem x v && W T x env_empty)) as b;
      symmetry in Heqb; destruct b; try trivial; simpl.
    apply False_ind. rewrite exists_bool_false in Heqb. specialize (Heqb v').
    rewrite H0 in Heqb. simpl in Heqb.
    unfold LBind in IHEval2. simpl in IHEval2.
    rewrite Heqb in IHEval2.
    assert ((if F T v' env_empty then Return v_tt else Return v_ff) = Error).
      apply IHEval2. simpl. assumption.
      apply pure_interesting.
    destruct (F T v' env_empty); inversion H4.
    assumption. apply pure_magic.
  Case "eval_test_coll_3_1".
    apply app_eq_nil in H1. destruct H1.
    rewrite IHEval. unfold LBind. simpl. rewrite H.
    remember (exists_bool (fun x : Value => v_mem x v && W T x env_empty)) as b;
      symmetry in Heqb; destruct b; simpl.
    SCase "something goes wrong".
      apply exists_bool_true in Heqb. destruct Heqb as [v0 HH].
      rewrite andb_true_iff in HH. destruct HH.
      specialize (H4 v0 H6).
      unfold LBind in H4. simpl in H4.
      rewrite H7 in H4. apply H4. (* strange way to get a contradiction *)
      simpl. assumption.
      apply pure_interesting.
    SCase "alles ok". 
      remember (forall_bool (fun v0 : Value => implb (v_mem v0 v) (F T v0 env_empty))) as c;
        symmetry in Heqc; destruct c; simpl.
      SSCase "c = true". reflexivity.
      SSCase "c = false (contradiction)".
        apply forall_bool_false in Heqc.
        apply False_ind. apply Heqc. intro v0.
        rewrite implb_true_iff. intro Hv0.
        specialize (H4 v0 Hv0).
        unfold LBind in H4. simpl in H4.
        rewrite exists_bool_false in Heqb. specialize (Heqb v0).
        rewrite andb_false_iff in Heqb. destruct Heqb. rewrite Hv0 in H6. inversion H6.
        rewrite H6 in H4. destruct (F T v0 env_empty); try trivial.
        assert(Return v_ff = Return v_tt). apply H4. simpl. assumption.
          apply pure_interesting.
        inversion H7.
    assumption. apply pure_magic.
  Case "eval_test_coll_3_2".
     (* XXX there is a missing induction hypothesis ... induction principle is weak ! *)
    assert (forall v' : Value,
     v_mem v' v = true ->
     fv_type T = nil ->
     Pure (E_In (E_Value v') T) ->
     exists v'' : Value,
       (if W T v' env_empty
        then Error
        else if F T v' env_empty then Return v_tt else Return v_ff) =
       Return v'' /\ In_Logical v' = true) as IHEval3. admit. (* this is much easier than proving a better induction principle (48 cases) *)
    apply app_eq_nil in H1. destruct H1.
    rewrite IHEval1. unfold LBind. simpl. rewrite H.
    remember (exists_bool (fun x : Value => v_mem x v && W T x env_empty)) as b;
      symmetry in Heqb; destruct b; simpl.
    SCase "something goes wrong".
      apply exists_bool_true in Heqb. destruct Heqb as [v0 HH].
      rewrite andb_true_iff in HH. destruct HH.
      specialize (IHEval3 v0 H5).
      unfold LBind in H4. simpl in H4.
      assert(Pure (E_In (E_Value v0) T)). apply pure_interesting.
      specialize (IHEval3 H4 H7).
      destruct IHEval3 as [v'' IHEval3].
      rewrite H6 in IHEval3. destruct IHEval3. inversion H8.
    SCase "alles ok". 
      remember (forall_bool (fun v0 : Value => implb (v_mem v0 v) (F T v0 env_empty))) as c;
        symmetry in Heqc; destruct c; simpl.
      SSCase "c = true (contradiction)".
        rewrite <- forall_bool_true_iff in Heqc.
        specialize (Heqc v'''). rewrite implb_true_iff in Heqc. specialize (Heqc H3).
        unfold LBind in IHEval2. simpl in IHEval2.
        assert(Pure (E_In (E_Value v''') T)). apply pure_interesting.
        specialize (IHEval2 H4 H5).
        destruct (W T v''' env_empty).
        SSSCase "W = true (direct contradiction)". inversion IHEval2.
        SSSCase "W = false (contradicts Heqc)". rewrite Heqc in IHEval2. inversion IHEval2.
      SSCase "c = false". reflexivity.
    assumption. apply pure_magic.
*)
  Case "eval_test_refine". admit.
(*
  Case "eval_test_refine_1".
    apply app_eq_nil in H1. destruct H1. apply app_eq_nil in H0. destruct H0.
    remember (R e1 env_empty) as b; symmetry in Heqb; destruct b; try trivial.
    unfold LBind in *. simpl in *.
    remember (W T v0 env_empty) as c; symmetry in Heqc; destruct c; try trivial. simpl. 
    remember (negb (syn_beq_res (R e2 (env_update env_empty x v0)) (Return v_tt)
       || syn_beq_res (R e2 (env_update env_empty x v0)) (Return v_ff))) as f; symmetry in Heqf; destruct f; try trivial.
    specialize (IHEval1 H (pure_magic e1)). inversion IHEval1. subst.
    remember (F T v env_empty) as d; symmetry in Heqd; destruct d; simpl.
    SCase "F T v env_empty = true".
      assert (Return v_tt = Error). rewrite Heqc in IHEval2. apply IHEval2. rewrite H0. reflexivity.
        apply pure_interesting.
      inversion H3.
    SCase "F T v env_empty = false".
      assert (Return v_ff = Error). rewrite Heqc in IHEval2. apply IHEval2. rewrite H0. reflexivity.
        apply pure_interesting.
      inversion H3.
  Case "eval_test_refine_2_1".
    apply app_eq_nil in H1. destruct H1. apply app_eq_nil in H0. destruct H0.
    rewrite IHEval1. unfold LBind. simpl.
    unfold LBind in IHEval2. simpl in IHEval2.
    remember (W T v1 env_empty) as c; symmetry in Heqc; destruct c; try trivial. simpl. 
    remember (negb (syn_beq_res (R e2 (env_update env_empty x v1)) (Return v_tt)
       || syn_beq_res (R e2 (env_update env_empty x v1)) (Return v_ff))) as f; symmetry in Heqf; destruct f; try trivial.
    rewrite R_subst in IHEval3. rewrite IHEval3 in Heqf. simpl in Heqf. inversion Heqf.
    rewrite fv_exp_subst. assumption.
    apply pure_subst. apply pure_magic.
    assumption. apply pure_magic.
  Case "eval_test_refine_2_2".
    apply app_eq_nil in H1. destruct H1. apply app_eq_nil in H1. destruct H1.
    rewrite IHEval1. unfold LBind. simpl.
    unfold LBind in IHEval2. simpl in IHEval2.
    remember (W T v1 env_empty) as c; symmetry in Heqc; destruct c; try trivial. simpl. 
    remember (negb (syn_beq_res (R e2 (env_update env_empty x v1)) (Return v_tt)
       || syn_beq_res (R e2 (env_update env_empty x v1)) (Return v_ff))) as f; symmetry in Heqf; destruct f; try trivial.
    apply False_ind. (* everything proved by contradiction *)
    rewrite R_subst in IHEval3. rewrite IHEval3 in Heqf.
    rewrite andb_false_iff in H. destruct H; apply not_logical in H; destruct H.
    SCase "In_Logical v' = false".
      destruct (F T v1 env_empty).
      SSCase "F T v1 env_empty = true".
        assert(Return v_tt = Return v'). apply IHEval2. simpl. assumption.
          apply pure_interesting.
        inversion H5. apply H. auto.
      SSCase "F T v1 env_empty = false".
        assert(Return v_ff = Return v'). apply IHEval2. simpl. assumption.
          apply pure_interesting.
        inversion H5. apply H4. auto.
    SCase "In_Logical v2 = false".
      simpl in Heqf. rewrite negb_false_iff in Heqf. rewrite orb_true_iff in Heqf. 
      destruct Heqf; apply syn_beq_val_true in H5. apply H. auto. apply H4. auto.
    rewrite fv_exp_subst. assumption.
    apply pure_subst. apply pure_magic.
    assumption. apply pure_magic.
  Case "eval_test_refine_3".
    apply app_eq_nil in H1. destruct H1. apply app_eq_nil in H1. destruct H1.
    unfold LBind in IHEval2. simpl in IHEval2.
    rewrite IHEval1. unfold LBind. simpl.
    remember (W T v1 env_empty || negb (syn_beq_res (R e2 (env_update env_empty x v1)) (Return v_tt)
          || syn_beq_res (R e2 (env_update env_empty x v1)) (Return v_ff))) as b; symmetry in Heqb; destruct b; try trivial.
    SCase "1". 
      apply False_ind. (* on this case we obtain contradiction(s) *)
      destruct (W T v1 env_empty).
      SSCase "W = true".
        assert (Error = Return v'). apply IHEval2. rewrite H1. reflexivity.
          apply pure_interesting.
        inversion H4.
      SSCase "W = false". simpl in Heqb.
        rewrite negb_true_iff in Heqb. rewrite orb_false_iff in Heqb.
        destruct Heqb. rewrite andb_true_iff in H. destruct H.
        rewrite R_subst in IHEval3.
        assert (R e2 (env_update env_empty x v1) = Return v2).
          apply IHEval3. rewrite fv_exp_subst. assumption.
          apply pure_subst. apply pure_magic.
        rewrite H7 in H4. rewrite H7 in H5.
        apply logical_only_two in H6. destruct H6; subst; simpl in *. inversion H4. inversion H5.
    SCase "2". rewrite orb_false_iff in Heqb. destruct Heqb. rewrite H4 in IHEval2.
      rewrite negb_false_iff in H5. rewrite orb_true_iff in H5.
      rewrite R_subst in IHEval3. rewrite IHEval3. simpl in *.
      destruct (F T v1 env_empty); simpl.
      SSCase "F T v1 env_empty = true".
        assert(Return v_tt = Return v').
          apply IHEval2. simpl. assumption.
          apply pure_interesting.
        inversion H6. subst. unfold O_And. simpl.
        remember (syn_beq_val v2 v_tt) as d; symmetry in Heqd; destruct d.
        SSSCase "d=true".
          apply syn_beq_val_true in Heqd. subst. reflexivity.
        SSSCase "d=false". simpl.
          apply syn_beq_val_false in Heqd.
            rewrite andb_true_iff in H. destruct H.
            apply logical_not_true in H7; try assumption. subst. reflexivity.
      SSCase "F T v1 env_empty = false".
        assert(Return v_ff = Return v').
          apply IHEval2. simpl. assumption.
          apply pure_interesting.
        inversion H6. subst. reflexivity.
    rewrite fv_exp_subst. assumption.
    apply pure_subst. apply pure_magic.
    assumption. apply pure_magic.
*)
Qed.

