(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

Require Import Model.

(** ** Operators *)

Inductive BinOp : Type := 
  | OPlus : BinOp
  | OMinus : BinOp
  | OTimes : BinOp
  | OGt : BinOp
  | OLt : BinOp
  | OEq : BinOp
  | OAnd : BinOp
  | OOr : BinOp.

Inductive UnOp : Type := 
  | ONot : UnOp.

Inductive MType : Type :=
  | T_Any : MType
  | T_Integer : MType
  | T_Text : MType
  | T_Logical : MType
  | T_Coll : MType -> MType
  | T_Entity : string -> MType -> MType
  | T_Refine : string -> MType -> Exp -> MType
with Exp : Type :=
  | E_Var : string -> Exp
  | E_Value : Value -> Exp
  | E_UnOp : UnOp -> Exp -> Exp
  | E_BinOp : BinOp -> Exp -> Exp -> Exp
  | E_If : Exp -> Exp -> Exp -> Exp
  | E_Let : string -> Exp -> Exp -> Exp
  | E_In : Exp -> MType -> Exp
  | E_Entity : list (string * Exp) -> Exp (* XXX need to prevent duplicate labels *)
  | E_Dot : Exp -> string -> Exp
  | E_Add : Exp -> Exp -> Exp
  | E_Acc : string -> Exp -> string -> Exp -> Exp -> Exp (* XXX need to prevent that the 2 variables are equal *)
  | E_App : string -> Exp -> Exp. (* multiple arguments are encoded as entities *)

Inductive Sexp_exp : Exp -> Exp -> Prop :=
  | SE_UnOp : forall op e, Sexp_exp e (E_UnOp op e)
  | SE_BinOp_1 : forall op e1 e2, Sexp_exp e1 (E_BinOp op e1 e2)
  | SE_BinOp_2 : forall op e1 e2, Sexp_exp e2 (E_BinOp op e1 e2)
  | SE_If_1 : forall e1 e2 e3, Sexp_exp e1 (E_If e1 e2 e3)
  | SE_If_2 : forall e1 e2 e3, Sexp_exp e2 (E_If e1 e2 e3)
  | SE_If_3 : forall e1 e2 e3, Sexp_exp e3 (E_If e1 e2 e3)
  | SE_Let_1 : forall x e1 e2, Sexp_exp e1 (E_Let x e1 e2)
  | SE_Let_2 : forall x e1 e2, Sexp_exp e2 (E_Let x e1 e2)
  | SE_In_1 : forall e T, Sexp_exp e (E_In e T)
  | SE_In_2 : forall e e' T, Sexp_type e' T -> Sexp_exp e' (E_In e T)
  | SE_Entity : forall l e les, In (l,e) les -> Sexp_exp e (E_Entity les)
  | SE_Dot : forall e l, Sexp_exp e (E_Dot e l)
  | SE_Add_1 : forall e1 e2, Sexp_exp e1 (E_Add e1 e2)
  | SE_Add_2 : forall e1 e2, Sexp_exp e2 (E_Add e1 e2)
  | SE_Acc_1 : forall x e1 y e2 e3, Sexp_exp e1 (E_Acc x e1 y e2 e3)
  | SE_Acc_2 : forall x e1 y e2 e3, Sexp_exp e2 (E_Acc x e1 y e2 e3)
  | SE_Acc_3 : forall x e1 y e2 e3, Sexp_exp e3 (E_Acc x e1 y e2 e3)
  | SE_App : forall f e, Sexp_exp e (E_App f e)
with Sexp_type : Exp -> MType -> Prop :=
  | ST_Coll : forall e T, Sexp_type e T -> Sexp_type e (T_Coll T)
  | ST_Entity : forall e T l, Sexp_type e T -> Sexp_type e (T_Entity l T)
  | ST_Refine_1 : forall e' x T e, Sexp_type e' T -> Sexp_type e' (T_Refine x T e)
  | ST_Refine_2 : forall x T e, Sexp_type e (T_Refine x T e)
.

(* We could have used the logical semantics too to convert from expressions to values *)

Tactic Notation "Exp_cases" tactic(first) tactic(c) := first; [
    c "E_Var" | c "E_Value" | c "E_UnOp" | c "E_BinOp" | c "E_If" |
    c "E_Let" | c "E_In" | c "E_Entity" |
    c "E_Dot" | c "E_Add" | c "E_Acc" |
    c "E_App" ].

Tactic Notation "MType_cases" tactic(first) tactic(c) := first; [
    c "T_Any" | c "T_Integer" | c "T_Text" | c "T_Logical" | c "T_Coll" |
    c "T_Entity" | c "T_Refine" ].

(* XXX my "remove" hides the one in List
Check remove. *)

Fixpoint fv_exp (e : Exp) {struct e} : list string :=
  match e with
  | E_Var x => x :: nil
  | E_Value v => nil
  | E_UnOp o e => fv_exp e 
  | E_BinOp o e1 e2 => fv_exp e1 ++ fv_exp e2
  | E_If e1 e2 e3 => fv_exp e1 ++ fv_exp e2 ++ fv_exp e3
  | E_Let x e1 e2 => fv_exp e1 ++ (List.remove eq_str_dec x (fv_exp e2))
  | E_In e T => fv_exp e ++ fv_type T
  | E_Entity les => 
    (fix fv_entity les :=
      match les with
      | nil => nil
      | (_, e) :: les' => fv_exp e ++ (fv_entity les')
      end) les
  | E_Dot e l => fv_exp e
  | E_Add e1 e2 => fv_exp e1 ++ fv_exp e2
  | E_Acc x e1 y e2 e3 => fv_exp e1 ++ fv_exp e2 ++
    (List.remove eq_str_dec x (List.remove eq_str_dec y (fv_exp e3)))
  | E_App _ e => fv_exp e
  end
with fv_type (U : MType) {struct U} : list string :=
  match U with
  | T_Any => nil
  | T_Integer => nil
  | T_Text => nil
  | T_Logical => nil
  | T_Coll T => fv_type T
  | T_Entity l T => fv_type T
  | T_Refine x T e => fv_type T ++ (List.remove eq_str_dec x (fv_exp e))
  end.


Fixpoint subst_exp (e : Exp) (v : Value) (x : string) : Exp :=
  match e with
  | E_Var x0 => if beq_str x x0 then E_Value v else E_Var x0
  | E_Value v => E_Value v
  | E_UnOp o e1 => E_UnOp o (subst_exp e1 v x)
  | E_BinOp o e1 e2 => E_BinOp o (subst_exp e1 v x) (subst_exp e2 v x)
  | E_If e1 e2 e3 => E_If (subst_exp e1 v x) (subst_exp e2 v x) (subst_exp e3 v x)
  | E_Let y e1 e2 =>
     if beq_str x y then
       E_Let y (subst_exp e1 v x) e2
     else
       E_Let y (subst_exp e1 v x) (subst_exp e2 v x)
  | E_In e1 T => E_In (subst_exp e1 v x) (subst_type T v x)
  | E_Entity les =>
     E_Entity (List.map (fun le =>
                         match le with
                         | (li,ei) => (li, subst_exp ei v x)
                         end) les)
  | E_Dot e1 l => E_Dot (subst_exp e1 v x) l
  | E_Add e1 e2 => E_Add (subst_exp e1 v x) (subst_exp e2 v x)
  | E_Acc y1 e1 y2 e2 e3 => 
     if beq_str x y1 || beq_str x y2 then
       E_Acc y1 (subst_exp e1 v x) y2 (subst_exp e2 v x) e3
     else
       E_Acc y1 (subst_exp e1 v x) y2 (subst_exp e2 v x) (subst_exp e3 v x)
  | E_App f e => 
     E_App f (subst_exp e v x)
  end
with subst_type (U : MType) (v : Value) (x : string) : MType :=
  match U with
  | T_Coll T => T_Coll (subst_type T v x)
  | T_Entity l T => T_Entity l (subst_type T v x)
  | T_Refine y T e =>
     if beq_str x y then
       T_Refine y (subst_type T v x) e
     else
       T_Refine y (subst_type T v x) (subst_exp e v x)
  | _ => U
  end.

Fixpoint In_str (a:string) (l:list string) {struct l} : bool :=
  match l with
  | nil => false
  | b :: m => beq_str b a || In_str a m
  end.

Fixpoint fresh' (avoid : list string) counter str :=
  match counter with
  | 0 => str
  | S(c) => if In_str str avoid then fresh' avoid c (append str "a") else str
  end.

(* Fresh variables *)
Program Definition fresh (avoid : list string) : {x : string | ~In x avoid}. Admitted.
(*  fresh' avoid (List.length avoid) "a".
Next Obligation.
  induction avoid.
  Case "nil".
    simpl. unfold not. intro. assumption.
  Case "a::avoid".
    simpl. unfold not. intro. destruct H.
    admit. admit.
Defined.*)
(* XXX This is proved for atom in the Metatheory library:
http://www.cis.upenn.edu/~plclub/metalib/current/html/MetatheoryAtom.html#atom_fresh
*)




(* Basic Expressions *)
Definition E_True := E_Value (v_tt).
Definition E_False := E_Value (v_ff).
Definition E_Null := E_Value (v_null).
Definition E_Eq e1 e2 := E_BinOp OEq e1 e2.
Definition E_Or e1 e2 := E_BinOp OOr e1 e2.
Definition E_And e1 e2 := E_BinOp OAnd e1 e2.
Definition E_Not e := E_UnOp ONot e.

(* Type Asserts *)
Definition E_Stuck := E_Dot (E_Value v_zero) "l".
Definition E_Assert (e : Exp) (T : MType) : Exp :=
  let (x, _) := fresh ((fv_exp e) ++ (fv_type T)) in
  E_Let x e (E_If (E_In (E_Var x) T) (E_Var x) E_Stuck).

(* Derived Types *)
(* should only use this when "value" does not appear free in e*)
Definition T_Ref T e := T_Refine "value" T e.

Definition T_Empty := T_Ref T_Any E_False.

(* This only makes sense when e is pure *)
Definition T_Sing e :=
  let (x, _) := fresh (fv_exp e) in
  T_Refine x T_Any (E_Eq (E_Var x) e).

Definition T_Union T U :=
  let (x, _) := fresh (fv_type T ++ fv_type U) in
  T_Refine x T_Any (E_Or (E_In (E_Var x) T) (E_In (E_Var x) U)).
Definition T_Inter T U :=
  let (x, _) := fresh (fv_type T ++ fv_type U) in
  T_Refine x T_Any (E_And (E_In (E_Var x) T) (E_In (E_Var x) U)).
Definition T_Inter_Asym T U :=
  let (x, _) := fresh (fv_type T ++ fv_type U) in
  T_Refine x T (E_In (E_Var x) U).
Definition T_Neg T :=
  let (x, _) := fresh (fv_type T) in
  T_Refine x T_Any (E_Not (E_In (E_Var x) T)).

Definition T_General := T_Union (T_Union (T_Union T_Integer T_Text) T_Logical) (T_Sing E_Null).
Definition T_Collection := T_Coll T_Any.
Definition T_Entity_Empty := T_Neg (T_Union T_General T_Collection).

Fixpoint T_Entity_Multi (lts : list (string * MType)) := 
  match lts with
  | nil => T_Entity_Empty
  | (l,t) :: lts' => T_Inter (T_Entity l t) (T_Entity_Multi lts')
  end.

Definition T_Self (s : string) (T U : MType) := T_Refine s T (E_In (E_Var s) U).

