(*
    Copyright (c) 2010  Thorsten Tarrach
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

Require Export Calculus.
Require Export LogicalSemantics.
Require Export Model.
Require Import Purity.
Require Export Derived.

(**A This file contains definitions and lemmas from the %\dminor% paper %\cite{catalin}%.
     Lemmas from the paper are admitted and not proven here. Some corollaries from
     these lemmas are proven here.%\label{app-dminor}% *)

Definition EnvT := list (string * MType).

Definition envT_empty : EnvT := nil.

Definition envT_update (env : EnvT) (x : string) (t : MType) : EnvT :=
  (x,t)::env.

Definition op_type_un o :=
  match o with
  | ONot => (T_Logical,T_Logical)
  end.

Definition op_type_bi o :=
  match o with
  | OPlus | OMinus | OTimes => (T_Integer,T_Integer,T_Integer)
  | OGt | OLt => (T_Integer,T_Integer,T_Logical)
  | OEq => (T_Any,T_Any,T_Logical)
  | OAnd | OOr => (T_Logical,T_Logical,T_Logical)
  end.


Definition dom (X : Type) (lxs : list (string * X)) := fst (split lxs).
Implicit Arguments dom [X].

Reserved Notation "E '|--' ty" (at level 40).

Reserved Notation "E '|-' e ':::' ty" (at level 40).

Definition fst3 A B C (x:A*B*C) : A := fst (fst x).
Implicit Arguments fst3 [A B C].
Definition snd3 A B C (x:A*B*C) : B := snd (fst x).
Implicit Arguments snd3 [A B C].
Definition trd3 A B C (x:A*B*C) : C := snd x.
Implicit Arguments trd3 [A B C].

Definition FE (E:EnvT) sigma :=
  forall xn Tn, In (xn,Tn) E -> (forall vn, F Tn vn sigma = true).


Inductive wf_type : EnvT -> MType -> Prop :=
  | TypeAny : forall env,
    wf_env env ->
    env |-- T_Any
  | TypeScalarI : forall env,
    wf_env env ->
    env |-- T_Integer
  | TypeScalarL : forall env,
    wf_env env ->
    env |-- T_Logical
  | TypeScalarT : forall env,
    wf_env env ->
    env |-- T_Text
  | TypeColl : forall env T,
    env |-- T ->
    env |-- T_Coll T
  | TypeEnt : forall env l T,
    env |-- T ->
    env |-- T_Entity l T
  | TypeRef : forall env x e T,
    is_pure e = true ->
    ((x,T)::env) |- e ::: T_Logical ->
    env |-- T_Refine x T e

  where "env '|--' ty" := (wf_type env ty)

with wf_env : EnvT -> Prop :=
  | EnvEmpty : wf_env envT_empty
  | EnvVar : forall T x env,
    env |-- T ->
    ~ In x (dom env) ->
    wf_env ((x,T)::env)

with has_type : EnvT -> Exp -> MType -> Prop :=
  | ExpSub : forall e T T' env x,
    env |- e ::: T ->
    is_subtype env (T_Sing_x x e T) T' ->
    ~In x (dom env) ->
    env |- e ::: T'
  | ExpConst : forall env (v : Value),
    wf_env env ->
    env |- E_Value(v) ::: T_Any
  | ExpOperatorUn : forall env op e,
    env |- e ::: (fst (op_type_un op)) ->
    env |- E_UnOp op e ::: (snd (op_type_un op))
  | ExpOperatorBi : forall env op e1 e2,
    env |- e1 ::: (fst3 (op_type_bi op)) ->
    env |- e2 ::: (snd3 (op_type_bi op)) ->
    env |- E_BinOp op e1 e2 ::: (trd3 (op_type_bi op))
  | ExpCond : forall (env:EnvT) e1 e2 e3 T fr,
    fr = proj1_sig (fresh (dom env)) ->
    env |- e1 ::: T_Logical ->
    ((fr, T_Ok_x fr e1) :: env) |- e2 ::: T ->
    ((fr, T_Ok_x fr (E_Not e1)) :: env) |- e3 ::: T ->
    env |- E_If e1 e2 e3 ::: T
  | ExpVar : forall env v T,
    wf_env env ->
    In (v,T) env ->
    env |- E_Var v ::: T
  | ExpLet : forall env x e1 e2 U T,
    env |- e1 ::: T ->
    ((x,T)::env) |- e2 ::: U ->
    not (In x (fv_type U)) ->
    env |- E_Let x e1 e2 ::: U
  | ExpTest : forall env e T,
    env |- e ::: T_Any ->
    env |-- T ->
    env |- E_In e T ::: T_Logical
  | ExpEntity : forall env les lts,
    dom les = dom lts ->
    (forall l e T, In (l,e) les -> In (l,T) lts -> env |- e ::: T) ->
    env |- E_Entity les ::: T_Entity_Multi lts
  | ExpDot : forall env e l T,
    env |- e ::: T_Entity l T ->
    env |- E_Dot e l ::: T
  | ExpAdd : forall env e1 e2 T,
    env |- e1 ::: T ->
    env |- e2 ::: T_Coll T ->
    env |- E_Add e1 e2 ::: T_Coll T
  | ExpAcc : forall env x y e1 e2 e3 T U,
    env |- e1 ::: T_Coll T ->
    env |- e2 ::: U ->
    ((y,U)::(x,T)::env) |- e3 ::: U ->
    not (In x (fv_type U)) ->
    not (In y (fv_type U)) ->
    env |- E_Acc x e1 y e2 e3 ::: U
  (* | DTExpApp TODO *)

  where "env '|-' e ':::' ty" := (has_type env e ty)

with is_subtype : EnvT -> MType -> MType -> Prop :=
  | Subtype : forall E T T',
    E |-- T ->
    E |-- T' ->
    (forall sigma, FE E sigma -> forall v, F T v sigma = true -> F T' v sigma = true) ->
    is_subtype E T T'.


Tactic Notation "has_type_cases" tactic(first) tactic(c) :=
  first;
  [ c "ExpSub"
  | c "ExpConst"
  | c "ExpOperatorUn"
  | c "ExpOperatorBi"
  | c "ExpCond"
  | c "ExpVar"
  | c "ExpLet"
  | c "ExpTest"
  | c "ExpEntity"
  | c "ExpDot"
  | c "ExpAdd"
  | c "ExpAcc"
].

(* Lemma 6 in the paper *)
Lemma implj_T : forall env T,
  env |-- T ->
  wf_env env /\ incl (fv_type T) (dom env)
with implj_sub : forall env T T',
  is_subtype env T T' ->
  env |-- T'
with implj_E : forall env e T,
  env |- e ::: T ->
  env |-- T /\ incl (fv_exp e) (dom env)
.
Admitted.

Corollary env_correct_T : forall env T,
  env |-- T ->
  wf_env env.
Proof.
  intros. apply implj_T in H. apply H.
Qed.

Corollary T_correct_E : forall env e T,
  env |- e ::: T ->
  env |-- T.
Proof.
  intros. apply implj_E in H. apply H.
Qed.

Corollary env_correct_E: forall env e T,
  env |- e ::: T ->
  wf_env env.
Proof.
  intros. apply T_correct_E in H.
  apply env_correct_T in H. assumption.
Qed.

(* Lemma 8 in the paper *)
Lemma weakening_env : forall env env' x T',
  wf_env ((x,T')::env) ->
  ~ In x (dom env') ->
  wf_env (app env' env) ->
  wf_env (app env' ((x,T')::env))
with weakening_T : forall env env' x T T',
  wf_env ((x,T')::env) ->
  ~ In x (dom env') ->
  (app env' env) |-- T ->
  (app env' ((x,T')::env)) |-- T
with weakening_sub : forall env env' x S T T',
  wf_env ((x,T')::env) ->
  ~ In x (dom env') ->
  is_subtype (app env' env) S T ->
  is_subtype (app env' ((x,T')::env)) S T
with weakening_E : forall env env' x e T T',
  wf_env ((x,T')::env) ->
  ~ In x (dom env') ->
  (app env' env) |- e ::: T ->
  (app env' ((x,T')::env)) |- e ::: T
.
Admitted.


Corollary weakening_T' : forall (env:EnvT) x T t,
  ~ In x (dom env) ->
  env |-- t ->
  env |-- T ->
  ((x,t) :: env) |-- T.
Proof.
  intros. 
  assert ((x, t) :: env = app nil ((x, t) :: env)). simpl. reflexivity.
  rewrite (H2). apply weakening_T.
  apply EnvVar. assumption. assumption.
  simpl. unfold not. intro. assumption.
  simpl. assumption.
Qed.

Corollary weakening_E' : forall (env:EnvT) x T t e,
  ~ In x (dom env) ->
  env |-- t ->
  env |- e ::: T ->
  ((x,t) :: env) |- e ::: T.
Proof.
  intros. assert ((x, t) :: env = app nil ((x, t) :: env)). simpl. reflexivity.
  rewrite (H2). apply weakening_E.
  apply EnvVar. assumption. assumption.
  simpl. unfold not. intro. assumption.
  simpl. assumption.
Qed.


(* Additional Lemmas *)

Lemma ExpSubsum : forall e T T' env,
  env |- e ::: T ->
  is_subtype env T T' ->
  env |- e ::: T'.
Proof.
  intros e T T' env H H0. destruct (fresh (dom env)) as [x Hx]. eapply ExpSub with (T := T) (x := x).
  apply H. unfold T_Sing_x. remember (is_pure e) as pe. symmetry in Heqpe. destruct pe. 
  Case "is_pure e".
    inversion H0. apply Subtype. admit. (* XXX *) assumption.
    intros. apply H3. assumption. simpl in H8. rewrite andb_true_iff in H8. apply H8.
  Case "not pure".
    assumption.
  assumption.
Qed.

Lemma ExpSingleton : forall env e T x,
  env |- e ::: T ->
  ~In x (dom env) ->
  env |- e ::: T_Sing_x x e T.
Proof.
  intros env e T x H Hx.
  eapply ExpSub with (x:=x). apply H. apply Subtype.
  unfold T_Sing_x. remember (is_pure e). symmetry in Heqb. destruct b.
  apply TypeRef. simpl. assumption. apply ExpOperatorBi. unfold fst3. simpl.
  eapply ExpSubsum. apply ExpVar. constructor. apply T_correct_E in H. apply H.
  apply Hx. simpl. auto. apply Subtype. admit. (* XXX *) apply TypeAny. constructor. apply T_correct_E in H. apply H.
  apply Hx. intros. simpl. reflexivity. unfold snd3. simpl.
  eapply ExpSubsum. apply weakening_E'. apply Hx. apply T_correct_E in H. apply H. apply H.
  apply Subtype. admit. (* XXX *) apply TypeAny. constructor. apply T_correct_E in H. apply H.
  apply Hx. intros. simpl. reflexivity. apply T_correct_E in H. apply H.
  admit. (* XXX *)
  intros. apply H1. apply Hx.
Qed.

Lemma WellformedTypeUseless : forall e T env x,
  env |- e ::: T ->
  is_pure e = true ->
  ~ In x (dom env) ->
  env |-- (T_Sing_x x e T).
Proof.
  intros. unfold T_Sing_x. rewrite H0.
  assert (env |-- T) as Henv. apply T_correct_E in H. assumption.
  apply TypeRef. simpl. assumption.
  unfold E_Eq. apply ExpOperatorBi. unfold fst3. simpl.
  eapply ExpSubsum. apply ExpVar. constructor.
  assumption. assumption.
  simpl. left. reflexivity.
  apply Subtype. admit. (* XXX *) constructor. constructor.
  assumption. assumption.
  intros. simpl. reflexivity.
  unfold snd3. simpl.
  eapply ExpSubsum. apply weakening_E'. assumption.
  assumption. eassumption.
  apply Subtype. admit. (* XXX *) constructor. constructor.
  assumption. assumption.
  intros. simpl. reflexivity.
Qed.

Lemma any_cast : forall env e T,
  env |- e ::: T ->
  env |- e ::: T_Any.
Proof.
  intros env e T H.
  assert (env |-- T) as HT. apply T_correct_E in H. assumption.
  assert (wf_env env) as Henv. apply env_correct_T in HT. assumption.
  apply ExpSubsum with (T:=T). assumption.
  eapply Subtype. admit. (* XXX *)
  apply TypeAny. assumption.
  intros. simpl. reflexivity.
Qed.

