(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

Require Import Basic.
Require Import Scalars. (* Strings *)
Require Import List.
Require Import Bool.
Require Import Relations.

Section ExtensionalArrays.
  Variable D : Type.
  Variable R : Type.
  Variable cmp_D : D -> D -> comparison.
  Variable beq_R : R -> R -> bool.

  Definition beq_D (d1 d2 : D) : bool :=
    match cmp_D d1 d2 with
    | Eq => true
    | _ => false
    end.
  
  Definition EqD (d1 d2 : D) : Prop := beq_D d1 d2 = true.
  Definition EqR (r1 r2 : R) : Prop := beq_R r1 r2 = true.

  Variable beq_D_equiv : equivalence D EqD.
  Variable beq_R_equiv : equivalence R EqR.

  Lemma beq_D_refl : forall d, beq_D d d = true.
  Proof. destruct beq_D_equiv. apply equiv_refl. Qed.

  Lemma beq_R_refl : forall r, beq_R r r = true.
  Proof. destruct beq_R_equiv. apply equiv_refl. Qed.

  Definition Array := prod (list (D * R)) R.

  (** *** Equality of arrays *)
  (** We need to define a strong enough notion of equality to get extensionality *)

  Fixpoint syn_beq_map (m1 m2 : list (D * R)) :=
    match m1, m2 with
    | nil, nil => true
    | (d1, r1) :: m1', (d2, r2) :: m2' => 
      beq_R r1 r2 && beq_D d1 d2 && syn_beq_map m1' m2'
    | _, _ => false
    end.

  Definition syn_beq_arr (a1 a2 : Array) :=
    match a1, a2 with
    | (m1, def1), (m2, def2) => syn_beq_map m1 m2 && beq_R def1 def2
    end.

  (** XXX TODO: implement this *)
  Definition sort_arr (a : Array) := a.

  (** XXX TODO: implement this *)
  Definition remove_dup_keys_arr (a : Array) := a.

  Definition normalize_arr (a : Array) := sort_arr (remove_dup_keys_arr a).

  Definition beq_arr (a1 a2 : Array) :=
    syn_beq_arr (normalize_arr a1) (normalize_arr a2).


  (** *** Standard operations on arrays *)

  Definition const (def : R) : Array := (nil, def).

  Definition default (a : Array) : R := snd a.

  Lemma default_const: forall (def : R), default (const def) = def.
  Proof. intros def. reflexivity. Qed.

  Fixpoint select_helper (m : list (D * R)) (def : R) (d : D) := 
  match m with
  | nil => def
  | (d', r') :: m' => if beq_D d d' then r' else select_helper m' def d
  end.
  Definition select (a : Array) := select_helper (fst a) (snd a).

  Definition store (a : Array) (d : D) (r : R) := ((d, r) :: fst a, snd a).

  Lemma select_store_eq : forall a d r,
    select (store a d r) d = r.
  Proof.
    intros a d r. unfold select, store, select_helper.
    destruct a; simpl. rewrite beq_D_refl. reflexivity.
  Qed.

  Lemma select_store_not_eq : forall a r d1 d2, beq_D d1 d2 = false ->
    select (store a d2 r) d1 = select a d1.
  Proof.
    intros a r d1 d2 H. unfold select, store, select_helper; simpl.
    rewrite H. reflexivity.
  Qed.

  Lemma select_const : forall d r, select (const r) d = r.
  Proof. intros; unfold select, const; reflexivity. Qed.

  Lemma default_store : forall a d r, default (store a d r) = default a.
  Proof. intros; unfold default, store; reflexivity. Qed.

  Lemma extensionality : forall a1 a2,
    beq_arr a1 a2 = true <-> forall d : D, beq_R (select a1 d) (select a2 d) = true.
  Admitted.

  Definition map_arr (a : Array) (f : R -> R) :=
    (map (fun dr => match dr with (d,r) => (d, f r) end) (fst a), f (snd a)).

  Lemma map_arr_select : forall a f d,
    EqR (select (map_arr a f) d) (f (select a d)).
  Proof.
    intros. unfold EqR, select, select_helper, map_arr; simpl.
    induction (fst a) as [| [d' r'] fsta' ]; simpl in *.
    Case "fst a = nil". apply beq_R_refl.
    Case "fst a = (d', r') :: fsta'".
      destruct (beq_D d d').
      SCase "d = d'". apply beq_R_refl.
      SCase "d != d'". rewrite IHfsta'. reflexivity.
  Qed.

  Lemma map_default : forall a f,
    EqR (default (map_arr a f)) (f (default a)).
  Proof. intros. unfold EqR, default, map_arr; simpl. apply beq_R_refl. Qed.

  (* XXX we might want to remove duplicates first? (use remove_dup)
         we might want to abstract away from the order of elements in support (use Permutation)
	-- both would require stating lemma support_store in a more abstract way
   *)
  Definition support (a : Array) := fst (split (fst a)).
  
  Lemma support_const : forall r, support (const r) = nil.
  Proof. intro r. unfold support,const. reflexivity. Qed.

  Lemma support_store : forall a d r, support (store a d r) = d :: support a.
  Proof. intros a d r. unfold support, store.
    rewrite fst_pair. rewrite fst_split_cons. reflexivity. Qed.
(*
  Lemma select_support_default : forall a d, 
    {In d (support a)} + {(select a d) = default a}.
  Proof. intros a d. destruct a. destruct (contained Value eqX_dec d (support (l,r))).
*)
  Definition allb_R (a : Array) (P : R -> bool) :=
    P (default a) && (forallb (fun d => P (select a d)) (support a)).

  Lemma default_allb_R (a : Array) (P : R -> bool) :
    P (default a) = true /\ allb_R a P = true -> forall d, P (select a d) = true.
  Admitted.

  Definition map2_arr (a1 : Array) (a2 : Array) (f : R -> R -> R) :=
    let def := f (default a1) (default a2) in
    let sup := app (support a1) (support a2) in
    fold_left (fun a d => store a d (f (select a1 d) (select a2 d))) sup (const def).


  Lemma app_eq_nil : forall (X : Type) (xs ys : list X),
    xs ++ ys = nil -> xs = nil /\ ys = nil.
  Proof. intros. split. destruct xs; try reflexivity. simpl in H. inversion H.
     destruct xs. simpl in H. assumption. simpl in H. inversion H. Qed.

  Lemma app_eq_cons : forall (X : Type) (xs ys zs: list X) (z : X),
    xs ++ ys = z :: zs -> (exists xs', xs = z :: xs' /\ zs = xs' ++ ys) \/ (xs = nil /\ ys = z :: zs).
  Proof. intros. destruct xs. right. split. reflexivity. simpl in H. assumption.
    left. simpl in H. exists xs. inversion H. subst. split; reflexivity. Qed.

  Lemma support_nil : forall a, support a = nil -> fst a = nil.
  Proof.
    intros. destruct a as [drs def].
      simpl. unfold support in H. simpl in H.
      destruct drs. reflexivity.
        destruct p. simpl in H. destruct (split drs). simpl in H. inversion H.
  Qed.

  Lemma map2_arr_select : forall a1 a2 f d,
    EqR (select (map2_arr a1 a2 f) d) (f (select a1 d) (select a2 d)).
  Proof.
    intros. unfold EqR, select, select_helper, map2_arr, default; simpl.
    remember (support a1 ++ support a2) as sa1sa2.
    induction sa1sa2 as [| d' sa1sa2' ]; simpl in *.
    Case "sa1sa2 = nil".
      symmetry in Heqsa1sa2. apply app_eq_nil in Heqsa1sa2.
      destruct Heqsa1sa2 as [H1 H2].
      apply support_nil in H1. apply support_nil in H2.
      rewrite H1. rewrite H2. apply beq_R_refl.
    Case "fst a = d' :: sa1sa2'".
      symmetry in Heqsa1sa2.
      apply app_eq_cons in Heqsa1sa2. destruct Heqsa1sa2 as [H | H].
        SCase "". destruct H as [sa1' [H1 H2]].
      Admitted.

      (*
      remember (support a1) as sa1. destruct sa1; simpl in Heqsa1sa2.
      SCase "sa1 = nil". unfold support in Heqsa1sa2. 
      *)
      (*
      rewrite <- Heqsa1sa2.
      destruct (beq_D d d').
      SCase "d = d'". apply beq_R_refl.
      SCase "d != d'". rewrite IHfsta'. reflexivity.
      *)

Lemma map2_arr_default_helper : forall a' a'' f a1 a2 sa1sa2,
default a' = default a'' ->
beq_R
  (snd
    (fold_left
        (fun (a : Array) (d : D) =>
        store a d (f (select a1 d) (select a2 d))) sa1sa2
        a')) (f (snd a1) (snd a2)) = true
->
beq_R
  (snd
     (fold_left
        (fun (a0 : Array) (d : D) =>
         store a0 d (f (select a1 d) (select a2 d))) sa1sa2
        a'')) (f (snd a1) (snd a2)) = true.
Admitted.


  Lemma map2_arr_default : forall a1 a2 f,
    EqR (default (map2_arr a1 a2 f)) (f (default a1) (default a2)).
  Proof.
    intros. unfold EqR, default, map2_arr.
    remember (support a1 ++ support a2) as sa1sa2. induction sa1sa2; simpl.
      SCase "sa1sa2 = nil". unfold default. apply beq_R_refl.
      SCase "sa1sa2 = cons". unfold default in IHsa1sa2. unfold default.
      assert (store (const (f (snd a1) (snd a2))) a (f (select a1 a) (select a2 a)) =
        ((a, f (select a1 a) (select a2 a)) :: nil, snd (const (f (snd a1) (snd a2))))) as H'.
        SSCase "assert". unfold store; reflexivity.
      rewrite H'. unfold const. simpl.
Admitted.

  (** Much easier to prove map2_arr_default with this alternative definition *)
  Definition map2_arr' (a1 : Array) (a2 : Array) (f : R -> R -> R) :=
    let def := f (default a1) (default a2) in
    let sup := app (support a1) (support a2) in
    (fold_left (fun a d => (d, f (select a1 d) (select a2 d)) :: a) sup nil, def).

  Lemma map2_arr'_default : forall a1 a2 f,
    EqR (default (map2_arr' a1 a2 f)) (f (default a1) (default a2)).
  Proof. intros. unfold EqR, default, map2_arr. simpl. unfold default. apply beq_R_refl. Qed.

  Lemma arr_ind : forall (P : Array -> Prop),
    (forall d, P (const d)) ->
    (forall (d : D) (r : R) (a : Array), P a -> P (store a d r)) ->
    forall a : Array, P a.
  Proof. intros. destruct a as [drs def]. unfold const, store in *.
    induction drs as [| [d' r'] drs']. apply H.
    specialize H0 with (d := d') (r := r') (a := (drs', def)). simpl in H0.
    apply H0. apply IHdrs'.
  Qed.

End ExtensionalArrays.

Implicit Arguments const [R].
Implicit Arguments select [D R].
Implicit Arguments store [D R].
Implicit Arguments default [D R].
Implicit Arguments support [D R].
Implicit Arguments allb_R [D R].
