(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

(** Inspired by:
CoqArt chapter 1 insertion sort on integers:
http://www.labri.u-bordeaux.fr/perso/casteran/CoqArt/horizon/index.html
and Xavier Leroy's mergesort:
http://moscova.inria.fr/~xleroy/coq/Mergesort.html
*)

Require Import Case.
Require Import List.
Require Import Relation_Definitions.


Section SORT.

(** A type equipped with a decidable total preorder:
- a \u2264 b and b \u2264 c then a \u2264 c (transitivity);
- a \u2264 b or b \u2264 a (totality).
*)

Variable A: Type.
Variable le: relation A.
Hypothesis le_trans: transitive A le.
Hypothesis le_total: forall x y, le x y \/ le y x.
Variable le_dec: forall (x y: A), {le x y} + {~le x y}.

Lemma le_refl: forall x, le x x.
Proof.
  intros. destruct (le_total x x); auto. 
Qed.

Lemma le_not: forall x y, ~le x y -> le y x.
Proof.
  intros. destruct (le_total x y). contradiction. auto.
Qed.

(** What it means for a list to be sorted in increasing order. *)

(*--- SortedBegin *)
Definition leAll (x : A) (ys : list A) := forall y, In y ys -> le x y.
Inductive Sorted: list A -> Prop :=
  | Sorted_nil: Sorted nil
  | Sorted_cons: forall hd tl, leAll hd tl -> Sorted tl -> Sorted (hd :: tl).
(*--- SortedEnd *)

Lemma Sorted_sing : forall (x : A), Sorted (x :: nil).
Proof. intros. apply Sorted_cons. unfold leAll. intros y H0. inversion H0. apply Sorted_nil. Qed.

Fixpoint insert_in_sorted (a : A) (al : list A) : list A :=
  match al with
  | nil => a :: nil
  | cons a' al' =>
      match le_dec a a' with
      | left _ =>  a :: a' :: al'
      | right _ => a' :: (insert_in_sorted a al')
      end
  end.

Lemma leAll_permutation : forall x xs ys,
  leAll x xs -> Permutation xs ys -> leAll x ys.
Proof.
  unfold leAll.
  intros x xs ys H H0 y H1.
  apply H.
  apply Permutation_sym in H0. eapply Permutation_in in H0.
  eassumption. assumption.
Qed.

Lemma leAll_cons : forall x y xs,
  leAll y xs -> le y x -> leAll y (x :: xs).
Proof. unfold leAll. intros. simpl in H1. destruct H1. subst. assumption.
  apply H. assumption. Qed.

Lemma Sorted_leAll_cons : forall x y xs,
  Sorted (x :: xs) -> le y x -> leAll y (x :: xs).
Proof.
  unfold leAll.
  intros. simpl in H1. destruct H1.
  Case "x = z". subst. assumption.
  Case "In x' xs'".
    inversion H. eapply le_trans. eassumption. apply H4. assumption.
Qed.

Theorem insert_in_sorted_permutation : forall (x : A) xs,
  Permutation (x :: xs) (insert_in_sorted x xs).
Proof.
  intros. induction xs as [| x' xs'].
  Case "xs = nil". simpl. apply Permutation_refl.
  Case "xs = x' :: xs'". simpl.
    case (le_dec x x'); intro H.
    SCase "left". apply Permutation_refl.
    SCase "right".
      eapply Permutation_trans. apply perm_swap. apply perm_skip. assumption.
Qed.

Theorem insert_in_sorted_sorted : forall (x : A) xs,
  Sorted xs -> Sorted (insert_in_sorted x xs).
Proof.
  intros. induction xs as [| x' xs']; simpl.
  Case "xs = nil". apply Sorted_cons. intros x0 H0. inversion H0. apply Sorted_nil.
  Case "xs = x' :: xs'".
    case (le_dec x x'); intro H'.
    SCase "le_dec x x' = left _".
      apply Sorted_cons; try assumption. apply Sorted_leAll_cons; assumption.
   SCase "le_dec x x' = right _".
      inversion H. subst. apply IHxs' in H3.
      apply Sorted_cons; try assumption.
      assert (leAll x' (x :: xs')) as H''.
        apply leAll_cons. assumption. apply le_not. assumption.
      inversion H. subst.
      apply leAll_permutation with (xs := x :: xs'). apply H''.
      apply insert_in_sorted_permutation.
Qed.

Fixpoint insertion_sort (xs : list A) :=
  match xs with
  | nil => nil
  | x :: xs' => insert_in_sorted x (insertion_sort xs')
  end.

Theorem insertion_sort_permutation : forall xs,
  Permutation xs (insertion_sort xs).
Proof.
  induction xs as [| x xs']; simpl. apply perm_nil.
  eapply perm_trans. eapply perm_skip. apply IHxs'.
  apply insert_in_sorted_permutation.
Qed.

Theorem insertion_sort_sorted : forall xs, Sorted (insertion_sort xs).
Proof.
  induction xs as [| x xs']; simpl. apply Sorted_nil.
  apply insert_in_sorted_sorted. assumption.
Qed.


Fixpoint update_in_sorted (a : A) (al : list A) : list A :=
  match al with
  | nil => a :: nil
  | cons a' al' =>
    match le_dec a a' with
    | left _ => 
      match le_dec a' a with
      | left _ => a :: al'  (* duplicate, keep newest *)
      | right _ => a :: a' :: al'
      end
    | right _ => a' :: (update_in_sorted a al')
    end
  end.

Fixpoint equiv_In (a : A) (l : list A) {struct l} : Prop :=
  match l with
  | nil => False
  | b :: m => (le b a /\ le a b) \/ equiv_In a m
  end.

Theorem update_in_sorted_fresh_permutation : forall (x : A) xs,
  ~equiv_In x xs -> (* additional precondition *)
  Permutation (x :: xs) (update_in_sorted x xs).
Proof.
  intros x xs H0. induction xs as [| x' xs'].
  Case "xs = nil". simpl. apply Permutation_refl.
  Case "xs = x' :: xs'". simpl.
    case (le_dec x x'); intro H.
    SCase "left".
      case (le_dec x' x); intro H'.
      SSCase "left".
        simpl in H0. apply False_ind. apply H0. left. split; assumption.
      SSCase "right". apply Permutation_refl.
    SCase "right".
      eapply Permutation_trans. apply perm_swap. apply perm_skip.
      apply IHxs'. simpl in H0. intro Hc. apply H0. right. assumption.
Qed.

Theorem update_in_sorted_sub_elements : forall (x : A) xs y,
  In y (update_in_sorted x xs) -> In y (x :: xs) .
Proof.
  intros x xs y H0. induction xs as [| x' xs'].
  Case "xs = nil". simpl in *. trivial.
  Case "xs = x' :: xs'". simpl in *.
    destruct (le_dec x x').
    SCase "left".
      destruct (le_dec x' x); simpl in *; tauto.
    SCase "right".
      simpl in *; tauto.
Qed.

Lemma le_leAll_trans : forall x x' xs,
  le x x' -> leAll x' xs -> leAll x xs.
Proof. intros. unfold leAll in *. intros y Hy. eapply le_trans. eassumption. apply H0. assumption. Qed.

Lemma leAll_sub_elements : forall x xs ys,
  leAll x xs -> (forall y, In y ys -> In y xs) -> leAll x ys.
Proof.
  unfold leAll.
  intros x xs ys H H0 y H1.
  apply H. apply H0. assumption.
Qed.

Theorem update_in_sorted_sorted : forall (x : A) xs,
  Sorted xs -> Sorted (update_in_sorted x xs).
Proof.
  intros. induction xs as [| x' xs']; simpl.
  Case "xs = nil". apply Sorted_cons. intros x0 H0. inversion H0. apply Sorted_nil.
  Case "xs = x' :: xs'".
    case (le_dec x x'); intro H'.
    SCase "le_dec x x' = left _".
      case (le_dec x' x); intro H''.
      SSCase "le_dec x' x = left _".
        inversion H; subst.
        apply Sorted_cons. eapply le_leAll_trans; eassumption. assumption.
      SSCase "le_dec x' x = right _".
        apply Sorted_cons; try assumption. apply Sorted_leAll_cons; assumption.
   SCase "le_dec x x' = right _".
      inversion H. subst. apply IHxs' in H3.
      apply Sorted_cons; try assumption.
      assert (leAll x' (x :: xs')) as H''.
        apply leAll_cons. assumption. apply le_not. assumption.
      inversion H. subst.
      apply leAll_sub_elements with (xs := x :: xs'). apply H''.
      apply update_in_sorted_sub_elements.
Qed.

End SORT.

(* permutation is actually stronger, the number of occurences has to match *)
Definition same_elements (X : Type) (x1s x2s : list X) :=
  (forall x, In x x1s -> In x x2s) /\
  forall x, In x x2s -> In x x1s.
Implicit Arguments same_elements [X].

Theorem permutation_same_elements : forall X (x1s x2s : list X),
  Permutation x1s x2s -> same_elements x1s x2s.
Proof.
  intros. unfold same_elements. induction H; split; try destruct IHPermutation;
    try destruct IHPermutation1; try destruct IHPermutation2; intros.
    inversion H. inversion H.
    simpl in *. destruct H2. left. assumption. right. apply H0. assumption.
    simpl in *. destruct H2. left. assumption. right. apply H1. assumption.
    simpl in *. destruct H. right; left. assumption. destruct H. left. assumption. right; right. assumption.
    simpl in *. destruct H. right; left. assumption. destruct H. left. assumption. right; right. assumption.
    apply H3. apply H1. assumption. apply H2. apply H4. assumption.
Qed.


(** A property of permutations that is missing from the List library:
  a permutation of a list without duplicates is a list without duplicates.
  Taken from: http://moscova.inria.fr/~xleroy/coq/Mergesort.html *)
Lemma Permutation_NoDup:
  forall (A : Type) (l l': list A), Permutation l l' -> NoDup l -> NoDup l'.
Proof.
  induction 1; intros.
  constructor.

  inversion H0; subst. constructor; auto.
  red; intro; elim H3. apply Permutation_in with l'; auto. apply Permutation_sym; auto.

  inversion H; subst. inversion H3; subst. 
  constructor. simpl. simpl in H2. intuition.
  constructor. simpl in H2. intuition. auto.

  auto.
Qed.


Implicit Arguments Sorted [A].
Implicit Arguments insert_in_sorted [A].
Implicit Arguments insertion_sort [A].
Implicit Arguments update_in_sorted [A].
