(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

Require Import Model.
Require Import Calculus.

Require Import ClassicalEpsilon.

(** We give a shallow embedding of classical FOL in Coq.
FOL formulas (not necessarily closed) are represented as functions in 
(string -> Value) -> bool / Prop 
*)

(** FOL bool mapped to Coq bool *)

(* Quantification in bool *)
Definition forall_bool (A : Type) (P : A -> bool) : bool :=
   match ClassicalEpsilon.excluded_middle_informative (forall x, P x = true) with
  | left Ptrue => true
  | right Pfalsee => false
  end.
Implicit Arguments forall_bool [A].

Lemma forall_bool_true_iff : forall A P,
  (forall x : A, P x = true) <-> forall_bool P = true.
Proof.
  intros. unfold forall_bool. split;
    destruct (excluded_middle_informative (forall x : A, P x = true));
    try trivial; intro H.
  apply n in H. inversion H.
  inversion H.
Qed.

Lemma forall_bool_false : forall A P,
  forall_bool P = false <->
  ~(forall x : A, P x = true).
Proof.
  split; intros. unfold forall_bool in *.
  destruct (excluded_middle_informative (forall x : A, P x = true)).
  inversion H. assumption.
  unfold forall_bool in *.
  destruct (excluded_middle_informative (forall x : A, P x = true)).
  apply H in e. inversion e. reflexivity.
Qed.


Definition exists_bool (A : Type) (P : A -> bool) : bool :=
   match ClassicalEpsilon.excluded_middle_informative (exists x, P x = true) with
  | left Ptrue => true
  | right Pfalsee => false
  end.
Implicit Arguments exists_bool [A].

Lemma exists_bool_false : forall A P,
  exists_bool P = false <->
  forall x : A, P x = false.
Proof.
  split; intros. unfold exists_bool in H.
  destruct (excluded_middle_informative (exists x : A, P x = true)).
  inversion H. remember (P x) as b; symmetry in Heqb; destruct b.
    apply False_ind. apply n. exists x. apply Heqb.
    reflexivity.
  unfold exists_bool.
  destruct (excluded_middle_informative (exists x : A, P x = true)).
  destruct e. specialize (H x). rewrite H in H0. inversion H0.
  trivial.
Qed.

Lemma exists_bool_true : forall A P,
  exists_bool P = true <->
  exists x : A, P x = true.
Proof.
  split; intros. unfold exists_bool in H.
  destruct (excluded_middle_informative (exists x : A, P x = true)).
  trivial. inversion H.
  unfold exists_bool.
  destruct (excluded_middle_informative (exists x : A, P x = true)).
  trivial. apply n in H. inversion H.
Qed.

(** Alternative: FOL bool mapped to Coq Prop 

Definition LCond (A : Type) (H : Prop) (a b : A) : A :=
  match ClassicalEpsilon.excluded_middle_informative H with
  | left Ptrue => a
  | right Pfalsee => b
  end.
*)


Definition LBind (t1 : Result) (xt2 : Value -> Result) :=
  if negb (Proper(t1)) then Error
  else xt2 (out_V t1).

(* XXX order of arguments to v_dot is different in the paper *)

Definition syn_beq_res (r1 r2 : Result) : bool :=
  match r1, r2 with
  | Return v1, Return v2 => syn_beq_val v1 v2
  | Error, Error => true
  | _, _ => false
  end.


Definition Env := string -> Value.

(* This doesn't work, variables are not values 
Definition env_empty (x : string) := E_Var x.
*)
Definition env_empty (x : string) := v_null.

Definition env_update (env : string -> Value) (x : string) (v : Value) :=
  fun y => if beq_str x y then v else env y.

Fixpoint F (U : MType) (t : Value) (env : string -> Value) {struct U} : bool :=
  match U with
  | T_Any => true
  | T_Integer => In_Integer t
  | T_Text => In_Text t
  | T_Logical => In_Logical t
  | T_Coll T => is_C t && (forall_bool (fun v => implb (v_mem v t) (F T v env))) (* XXX don't really need such a general forall here *)
  | T_Entity l T => is_E t && v_has_field l t && F T (v_dot l t) env
  | T_Refine x T e => F T t env && (syn_beq_res (R e (env_update env x t)) (Return v_tt))
  end
with W (U : MType) (t : Value) (env : string -> Value) {struct U} : bool :=
  match U with
  | T_Any => false
  | T_Integer => false
  | T_Text => false
  | T_Logical => false
  | T_Coll T => is_C t && (exists_bool (fun x => andb (v_mem x t) (W T x env)))
  | T_Entity l T => is_E t && v_has_field l t && W T (v_dot l t) env
  | T_Refine x T e => W T t env
     || (negb (orb (syn_beq_res (R e (env_update env x t)) (Return v_tt))
                                 (syn_beq_res (R e (env_update env x t)) (Return v_ff))))
  end
with R (e : Exp) (env : string -> Value) {struct e} : Result :=
  match e with
  | E_Var x =>  Return (env x)
  | E_Value v => Return v
  | E_UnOp o e => 
    LBind (R e env) (fun x =>
      match o with
      | ONot => if In_Logical x then Return (O_Not x) else Error
      end)
  | E_BinOp o e1 e2 => 
    LBind (R e1 env) (fun x1 => LBind (R e2 env) (fun x2 =>
      match o with
      (* Need to do these checks here, since the O_* operations are for the optimized semantics *)
      | OPlus => if In_Integer x1 && In_Integer x2 then Return (O_Sum x1 x2) else Error
      | OMinus => if In_Integer x1 && In_Integer x2 then Return (O_Minus x1 x2) else Error
      | OTimes => if In_Integer x1 && In_Integer x2 then Return (O_Mult x1 x2) else Error
      | OGt => if In_Integer x1 && In_Integer x2 then Return (O_GT x1 x2) else Error
      | OLt => if In_Integer x1 && In_Integer x2 then Return (O_LT x1 x2) else Error
      | OEq => Return (O_EQ x1 x2)
      | OAnd => if In_Logical x1 && In_Logical x2 then Return (O_And x1 x2) else Error
      | OOr => if In_Logical x1 && In_Logical x2 then Return (O_Or x1 x2) else Error
      end))
  | E_If e1 e2 e3 =>
    LBind (R e1 env) (fun x =>
      if syn_beq_val x v_tt then R e2 env
      else if syn_beq_val x v_ff then R e3 env
      else Error)
  | E_Let x e1 e2 => LBind (R e1 env) (fun y => R e2 (env_update env x y))
  | E_In e T =>
     LBind (R e env) (fun x =>
       if W T x env then Error
       else if F T x env then Return v_tt
       else Return v_ff)
  | E_Entity les => 
    (* Coq can't prove this terminating les_to_entity les (fun e => R e env) -- inlining *)
    (fix les_to_entity les :=
      match les with
      | nil => Return(v_eempty)
      | (l, e) :: les' =>
        LBind (R e env) (fun x =>
        LBind (les_to_entity les') (fun y =>
          (Return (v_eupdate l x y))))
      end) les
  | E_Dot e l =>
     LBind (R e env) (fun x =>
       if is_E x && v_has_field l x then Return (v_dot l x) else Error)
  | E_Add e1 e2 =>
     LBind (R e1 env) (fun x1 => LBind (R e2 env) (fun x2 =>
       if is_C x2 then Return (v_add x1 x2) else Error))
  | E_Acc x e1 y e2 e3 =>
     LBind (R e1 env) (fun x1 => LBind (R e2 env) (fun x2 =>
     res_accumulate (fun v1 v2 => R e3 (env_update (env_update env x v2) y v1)) x1 x2))
(*  | e_app f es => Error   
XXX parametrize everything by the semantics of functions!
*)
  | _ => Error
  end.


(** ** Substitution *)

Lemma env_update_same : forall x v env, env_update env x v x = v.
Admitted.

Lemma env_update_different : forall x y v env, x <> y -> env_update env x v y = env y.
Admitted.

Lemma env_update_overwrite : forall x y v u env,
 (env_update (env_update env x v) x u) y = (env_update env x u) y.
Admitted.

Lemma env_update_extensional_overwrite : forall x v u env,
 (env_update (env_update env x v) x u) = (env_update env x u).
Admitted.

Lemma env_update_extensional_swap : forall x y v u env,
  x <> y -> (env_update (env_update env x v) y u) = (env_update (env_update env y u) x v).
Admitted.

Lemma beq_str_false : forall s1 s2, beq_str s1 s2 = false -> s1 <> s2.
Admitted.

(* Note: this roughly corresponds to Lemma 5 in the paper *)
(* Most important cases checked below *)
Lemma R_subst : forall e v x env,
  R (subst_exp e v x) env =
  R e (env_update env x v)
with F_subst : forall T v x u env,
  F (subst_type T v x) u env =
  F T u (env_update env x v)
with W_subst : forall T v x u env,
  W (subst_type T v x) u env =
  W T u (env_update env x v).
Proof.
  Case "R_subst". clear R_subst.
    (Exp_cases (induction e) SCase); intros.
    SCase "E_Var". simpl.
      remember (beq_str x s) as b. symmetry in Heqb. destruct b.
      apply beq_str_true in Heqb. subst. rewrite env_update_same. reflexivity.
      apply beq_str_false in Heqb. rewrite env_update_different; try assumption. reflexivity.
    admit. admit. admit. admit. admit.
    SCase "E_In". simpl. rewrite IHe. destruct (R e (env_update env x v)); try trivial.
      unfold LBind; simpl. rewrite W_subst.
      destruct (W m v0 (env_update env x v)); try trivial.
      rewrite F_subst. destruct (F m v0 (env_update env x v)); trivial.
    admit. admit. admit. admit. admit.
  Case "F_subst". clear F_subst.
    (MType_cases (induction T) SCase); intros.
    admit. admit. admit. admit. admit. admit.
    SCase "T_Refine". simpl. rewrite <- IHT.
      remember (beq_str x s) as b. symmetry in Heqb. destruct b.
      SSCase "b=true".
        apply beq_str_true in Heqb. subst. simpl.
        rewrite env_update_extensional_overwrite. reflexivity.
      SSCase "b=false".
        apply beq_str_false in Heqb. simpl.
        rewrite R_subst. rewrite env_update_extensional_swap. reflexivity.
          intro Hc. subst. apply Heqb. reflexivity.
  Case "W_subst". clear W_subst.
    (MType_cases (induction T) SCase); intros.
    admit. admit. admit. admit. admit. admit.
    SCase "T_Refine". simpl. rewrite <- IHT.
      remember (beq_str x s) as b. symmetry in Heqb. destruct b.
      SSCase "b=true".
        apply beq_str_true in Heqb. subst. simpl.
        rewrite env_update_extensional_overwrite. reflexivity.
      SSCase "b=false".
        apply beq_str_false in Heqb. simpl.
        rewrite R_subst. rewrite env_update_extensional_swap. reflexivity.
          intro Hc. subst. apply Heqb. reflexivity.
Admitted.


(** ** Semantics of Derived Froms *)
Lemma R_E_Eq : forall env e1 e2,
  (R (E_Eq e1 e2) env) =
    LBind (R e1 env) (fun x1 => LBind (R e2 env) (fun x2 => 
    Return (v_logical (syn_beq_val x1 x2))
    )).
Proof. intros env e1 e2. simpl. unfold O_EQ. reflexivity. Qed.

Lemma F_T_Empty : forall env t, (F T_Empty t env) = false.
Proof. intros. reflexivity. Qed.


Lemma R_not_in_fv_exp : forall x e v env,
  ~ In x (fv_exp e) -> (R e (env_update env x v)) = R e env.
Admitted.

Lemma syn_beq_val_v_logical_v_tt : forall b, syn_beq_val (v_logical b) v_tt = b.
Proof. intro b. destruct b; reflexivity. Qed.

Lemma O_EQ_equals_tt : forall v1 v2, syn_beq_val (O_EQ v1 v2) v_tt = true ->
  v1 = v2.
Proof.
  intros. unfold O_EQ in H. rewrite syn_beq_val_v_logical_v_tt in H.
  apply syn_beq_val_equals. assumption.
Qed.


Lemma not_in_app_and : forall X (x : X) (xs ys : list X),
  ~In x (xs ++ ys) -> ~In x xs /\ ~In x ys.
Proof.
  intros. induction xs; simpl in *; split; try auto; apply not_or_rev in H; destruct H.
  intro Ho. destruct Ho. apply H. assumption. apply IHxs in H0. destruct H0. apply H0.
  assumption. apply IHxs in H0. destruct H0. assumption.
Qed.

Lemma F_not_in_fv_type : forall x T t v env,
  ~ In x (fv_type T) -> (F T t (env_update env x v)) = F T t env.
Admitted.

Lemma W_not_in_fv_type : forall x T t v env,
  ~ In x (fv_type T) -> (W T t (env_update env x v)) = W T t env.
Admitted.

Lemma F_T_Union : forall env T U t,
  W T t env = false -> W U t env = false ->
  F (T_Union T U) t env = orb (F T t env) (F U t env).
Proof. intros env T U t H1 H2. unfold T_Union. destruct (fresh (fv_type T ++ fv_type U)) as [x H'].
  simpl. rewrite env_update_same. unfold LBind. simpl.
  apply not_in_app_and in H'. destruct H' as [H'1 H'2].
  erewrite <- W_not_in_fv_type in H1; try eassumption.
  erewrite <- W_not_in_fv_type in H2; try eassumption.
  rewrite H1. rewrite H2.
  eapply F_not_in_fv_type in H'1. eapply F_not_in_fv_type in H'2.
  rewrite H'1. rewrite H'2.
  destruct (F T t env); destruct (F U t env); simpl; unfold O_Or, v_logical, v_tt, syn_beq_val;
  reflexivity.
Qed.

(* symmetric version *)
Lemma F_T_Inter : forall env T U t,
  W T t env = false -> W U t env = false ->
  F (T_Inter T U) t env = andb (F T t env) (F U t env).
Proof. intros env T U t H1 H2. unfold T_Inter. destruct (fresh (fv_type T ++ fv_type U)) as [x H'].
  simpl. rewrite env_update_same. unfold LBind. simpl.
  apply not_in_app_and in H'. destruct H' as [H'1 H'2].
  erewrite <- W_not_in_fv_type in H1; try eassumption.
  erewrite <- W_not_in_fv_type in H2; try eassumption.
  rewrite H1. rewrite H2.
  eapply F_not_in_fv_type in H'1. rewrite H'1.
  eapply F_not_in_fv_type in H'2. rewrite H'2.
  destruct (F T t env); destruct (F U t env); simpl; unfold O_And, v_logical, v_tt, syn_beq_val;
  reflexivity.
Qed.

(* asymmetric version *)
Lemma F_T_Inter_Asym : forall env T U t,
  (* W T t env = false -> *) W U t env = false ->
  F (T_Inter_Asym T U) t env = andb (F T t env) (F U t env).
Proof. intros env T U t (* H1 *) H2. unfold T_Inter_Asym. destruct (fresh (fv_type T ++ fv_type U)) as [x H'].
  simpl. rewrite env_update_same. unfold LBind. simpl.
  apply not_in_app_and in H'. destruct H' as [H'1 H'2].
  (* erewrite <- W_not_in_fv_type in H1; try eassumption. *)
  erewrite <- W_not_in_fv_type in H2; try eassumption.
  (* rewrite H1. Useless? *) rewrite H2.
  (* eapply F_not_in_fv_type in H'1. rewrite H'1. useless *)
  eapply F_not_in_fv_type in H'2. rewrite H'2.
  destruct (F U t env); simpl; unfold O_And, v_logical, v_tt, syn_beq_val;
  reflexivity.
Qed.

Lemma F_T_Neg : forall env T t,
  W T t env = false -> F (T_Neg T) t env = negb (F T t env).
Proof.
  intros env T t H. unfold T_Neg. destruct (fresh (fv_type T)) as [x H'].
  simpl. rewrite env_update_same. unfold LBind. simpl.
  erewrite <- W_not_in_fv_type in H; try eassumption.
  rewrite H.
  eapply F_not_in_fv_type in H'. rewrite H'.
  destruct (F T t env); simpl; unfold O_Not, v_logical, syn_beq_val;
  reflexivity.
Qed.

(* XXX TODO !!! uses T_Inter, so more assumptions needed *)
Lemma F_T_Entity_Multi : forall env lts t,
  F (T_Entity_Multi lts) t env = fold_left (fun a lt => match lt with (l,T) =>
    andb a (andb (v_has_field l t) (F T (v_dot l t) env)) end) lts true.
Admitted.

Lemma W42 : forall x T e t env,
  W (T_Refine x T e) t env = false -> W (T_Refine x T (E_Not e)) t env = false.
Admitted.

Lemma O_Not_54 : forall e env,
  syn_beq_res (Return (O_Not (out_V (R e env)))) (Return v_tt) =
    syn_beq_res (Return (out_V (R e env))) (Return v_ff).
Admitted.

Lemma Proper_out_V : forall r, Proper r = true -> Return (out_V r) = r.
Proof. intros r H. destruct r; simpl in *. inversion H. reflexivity. Qed.

Lemma syn_beq_val_syn_beq_res : forall v1 v2,
  syn_beq_val v1 v2 = syn_beq_res (Return v1) (Return v2).
Proof. intros v1 v2. reflexivity. Qed.

Lemma syn_beq_res_if_v_tt : forall b : bool,
  (syn_beq_res (if b then Return v_tt else Return v_ff) (Return v_tt)) = b.
Admitted.



Lemma F_T_Self : forall env s T U t,
  W U t (env_update env s t) = false ->
    F (T_Self s T U) t env = andb (F T t env) (F U t (env_update env s t)).
Proof.
  intros env s T U t H. unfold T_Self. simpl. unfold LBind. simpl.
  rewrite env_update_same. rewrite H. rewrite syn_beq_res_if_v_tt. reflexivity.
Qed.



(* Note: Any closed formula or term can be directly evaluated in the model, using Coq
-- same for open formulas / terms in a fixed(!) environment that binds all its variables 

Eval compute in (F T_Logical (v_int 5) env_empty).
     = false
     : bool
*)

(* We can even extract efficient OCaml code to do this 
  -- we could use this in the type-checker for optimization: constant propagation
Extraction "foo" F W R.
*)

(* For open formulas / terms some proving is needed *)

Lemma test3465 : forall x : Value,
  F (T_Sing E_True) x env_empty = true -> F T_Logical x env_empty = true.
Proof. intros x H. unfold T_Sing in *. simpl in *. destruct (fresh nil).
  simpl in H. 
  rewrite env_update_same in H.
  unfold O_EQ in H. unfold syn_beq_val in H.
  destruct x as [[[i | s | b |] | svm | vb] pv]; simpl in *; try inversion H.
  destruct b; compute in *. reflexivity. inversion H1.
Qed.


Lemma W_unused_var : forall x T v env,
  ~ In x (fv_type T) ->
  W T v (env_update env x v) = W T v env.
Admitted.

Lemma F_unused_var : forall x T v env,
  ~ In x (fv_type T) ->
  F T v (env_update env x v) = F T v env.
Admitted.

Lemma assert_encoding_correct : forall e T env,
  R (E_Assert e T) env =
     LBind (R e env) (fun x =>
       if W T x env then Error
       else if F T x env then Return x
       else Error).
Proof.
  intros. unfold E_Assert. destruct (fresh (fv_exp e ++ fv_type T)).
  simpl. unfold LBind. simpl. destruct (R e env) as [| v]; simpl. trivial.
  rewrite env_update_same.
  rewrite not_in_app in n. destruct n.
  rewrite W_unused_var; try assumption.
  destruct (W T v env); simpl; try trivial.
  rewrite F_unused_var; try assumption.
  destruct (F T v env); simpl; trivial.
Qed.
