(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

Require Export Values.
Require Export ValueEquality.
Require Import Coq.Program.Tactics.
Require Import Program.Wf.

(**
Value      -> Value = {x : RawValue | Normal x}
SVMap      -> {svm : list (string * RawValue) | NoDup (fst (split svm))
                                        /\ Sorted_svm svm
                                        /\ IndAll Normal (snd (split svm))}.
VBag       -> {vb : list RawValue | Sorted_vb vb /\ IndAll Normal vb}. 

- the advantage of doing it this way is that equality can be syntactic
- but we need to prove that the normal form is preserved by all operations
   - turns out that we can do these proofs using tactics and without cluttering the code
- some of the implementations need to be more complicated if normal forms are to be preserved
   - eg. for adding a RawValue to a collection (already ordered) we use insertion sorting
*)


(*--- NormalBegin *)
Inductive Normal : RawValue -> Prop :=
  | normal_G : forall g, Normal (G g)
  | normal_E : forall svm, 
    NoDup (fst (split svm)) ->
    Sorted_svm svm ->
    IndAll Normal (snd (split svm)) ->
      Normal (E svm)
  | normal_C : forall vb,
    Sorted_vb vb -> IndAll Normal vb -> Normal (C vb).
(*--- NormalEnd *)

(** Definining Coq types modelling Z3 sorts *)

(*--- InterSortsBegin*)
Definition Value := {x : RawValue | Normal x}.
Definition SVMap :=
  {svm : list (string * RawValue) | NoDup (fst (split svm))
    /\ Sorted_svm svm /\ IndAll Normal (snd (split svm)) }.
Definition VBag :=
  {vb : list RawValue | Sorted_vb vb /\ IndAll Normal vb}.
(*--- InterSortsEnd*)

(*--- ValueOptionBegin *)
Definition lift_option (X : Type) (P : X -> Prop) : option X -> Prop :=
  fun ox => match ox with None => True | Some x => P x end.
Definition ValueOption :=
  {vo : option RawValue | lift_option _ Normal vo}.
(*--- ValueOptionEnd *)
Implicit Arguments lift_option [X].

Lemma Normal_null : Normal (G G_Null).
Proof. apply normal_G. Qed.


Program Definition NoValue : ValueOption := None.
Program Definition SomeValue (v : Value) : ValueOption := Some v.
Next Obligation. destruct v. simpl. assumption. Qed.
Program Definition of_SomeValue (vo : ValueOption) : Value :=
  match vo with Some v => v | _ => G (G_Null) end.
Next Obligation. destruct vo as [rvo H]. destruct rvo as [rv | ]; simpl in *;
  inversion Heq_vo. subst. assumption.
Qed.
Next Obligation. apply Normal_null. Qed.

(* XXX could use an opaque arbitrary value for branches where we need to return something
-- as long as no proof unfolds arbitrary it should be better than using (G G_Null) directly *)
Proposition arbitrary : Value.
exact (exist (fun x => Normal x) (G G_Null) (normal_G G_Null)).
Qed. (* Qed makes this opaque -- can't be unfolded *)

(** ** Scalars (copy paste from ModelVariant1.v)
[[
:datatypes ((General 
  (G_Integer (of_G_Integer Int))
  (G_Text    (of_G_Text String))
  (G_Logical (of_G_Logical bool))
  G_Null))
]]
*)

Definition of_G_Integer g :=
  match g with
  | G_Integer i => i
  | _ => 0%Z (* arbitrary *)
  end.
Definition is_G_Integer g :=
  match g with
  | G_Integer _ => true
  | _ => false
  end.

Definition of_G_Text g :=
  match g with
  | G_Text s => s
  | _ => "" (* arbitrary *)
  end.
Definition is_G_Text g :=
  match g with
  | G_Text _ => true
  | _ => false
  end.

Definition of_G_Logical g :=
  match g with
  | G_Logical b => b
  | _ => false (* arbitrary *)
  end.
Definition is_G_Logical g :=
  match g with
  | G_Logical _ => true
  | _ => false
  end.

Definition is_G_Null g :=
  match g with
  | G_Null => true
  | _ => false
  end.


(** ** Values
[[
:datatypes (
  (RawValue
    (G (out_G General)) ;; General
    (E (out_E SVMap))   ;; entity: finite map from String to RawValue
    (C (out_C VBag))    ;; collection: finite multiset of RawValue
    (L (out_L VList)))  ;; list of RawValue               
  (VList Nil (Cons (out_Head RawValue) (out_Tail VList))))
]]
Note: We do not model lists (in the theory/paper they are encoded).
*)

Program Definition out_G (v : Value) : General :=
  match v with
  | G g => g
  | _ => G_Integer 0 (* arbitrary *)
  end.
Program Definition is_G (v : Value) : bool :=
  match v with
  | G _ => true
  | _ => false
  end.

(*--- outEBegin *)
Program Definition out_E (v : Value) : SVMap :=
  match v with | E svm => svm | _ => nil end.
(*--- outEEnd *)
Next Obligation.
    unfold proj1_sig in Heq_v. destruct v as [rv pf]. subst. inversion pf. subst.
    repeat try split; assumption.
Qed.
Next Obligation.
    simpl; repeat try split; try apply NoDup_nil; try apply All_nil; try apply Sorted_nil.
Qed.

(*--- isEBegin *)
Program Definition is_E (v : Value) : bool :=
  match v with | E _ => true | _ => false end.
(*--- isEEnd *)

Program Definition out_C (v : Value) : VBag :=
  match v with
  | C vb => vb
  | _ => nil (* arbitrary *)
  end.
Next Obligation.
    unfold proj1_sig in Heq_v. destruct v as [rv pf]. subst. inversion pf. subst.
    split; assumption.
Qed.
Next Obligation.
    simpl; repeat try split; try apply NoDup_nil; try apply All_nil; try apply Sorted_nil.
Qed.
Program Definition is_C (v : Value) : bool :=
  match v with
  | C _ => true
  | _ => false
  end.

(** We also need to (re)define the constructors G, E and C
to match the sorts in the axiomatization *)
Program Definition G' (g : General) : Value := G g.
Next Obligation. apply normal_G. Qed.

Program Definition E' (svm : SVMap) : Value := E svm.
Next Obligation.
  destruct svm as [rsvm [H1 [H2 H3]]]. simpl. apply normal_E; assumption.
Qed.

Program Definition C' (vb : VBag) : Value := C vb.
Next Obligation.
  destruct vb as [rsvm [H1 H2]]. simpl. apply normal_C; assumption.
Qed.


(** ** ValueOption
[[
:datatypes ((ValueOption
  NoValue
  (SomeValue (of_SomeValue RawValue))))
]] *)



(** ** Arrays -- copy paste from ModelVariant1.v *)
(** XXX Again we are nesting proofs inside terms*)
Require Import ExtensionalArrays.

(*--- SVMapArrayBegin *)
Definition SVMapArray := Array string ValueOption.
(*--- SVMapArrayEnd *)

(*--- VBagArrayBegin *)
Definition VBagArray := Array Value Z.
(*--- VBagArrayEnd *)

(** XXX this should be used everywhere -- defined earlier*)
Definition beq_int (i1 i2 : Z) :=
  match Zcompare i1 i2 with
  | Eq => true
  | _ => false
  end.

(*--- FinitePositiveBegin *)
Definition Finite (vb : VBagArray) : bool := beq_int (default vb) 0%Z.
Program Definition cmp_val (v1 v2 : Value) := cmp_rval v1 v2.
Program Definition Positive (vb : VBagArray) : bool :=
  allb_R cmp_val vb (fun (z : Z) =>
    (match Zcompare z 0 with Lt => false | _ => true end)).
(*--- FinitePositiveEnd *)

(*--- FiniteEBegin *)
Definition FiniteE (svm : SVMapArray) : bool := 
  match default svm with exist None _ => true | _ => false end.
(*--- FiniteEEnd *)

(** XXX Coq bug:
Program Fixpoint fold_svm (A : Type) (svm : SVMap) (f : A -> string -> Value -> A) (a : A)
  {measure (List.length svm)} : A :=
  match svm with
  | nil => a
  | (s,v) :: svm' => fold_svm svm' f (f a s v)
  end.
Anomaly: uncaught exception Not_found. Please report.
*)

(** XXX TODO: show (limited) isomorphism -- finite SVMapArray ~ SVMap *)

(*--- AlphamBegin *)
Program Fixpoint alpham (svm : SVMap) {measure List.length svm} : SVMapArray :=
  match svm with
  | nil => const string NoValue
  | (s,v) :: svm' => store (alpham svm') s (SomeValue v)
  end.
(*--- AlphamEnd *)
Next Obligation.
  destruct svm as [svm0 [H1 [H2 H3]]]. simpl in *. subst. split.
  rewrite fst_split_cons in H1. inversion H1. assumption.
  rewrite snd_split_cons in *. split.
  inversion H2. assumption. inversion H3. assumption.
Qed.
Next Obligation.
  destruct svm as [svm0 [H1 [H2 H3]]]. simpl in *. subst.
  rewrite snd_split_cons in H3. inversion H3. assumption.
Qed.
(* this was needed when proving finiteness right away
Next Obligation. remember 
        (alpham'
           (exist
              (fun svm'0 : SVMap =>
               List.length (proj1_sig svm'0) < List.length (proj1_sig svm))
              (exist
                 (fun svm0 : list (string * RawValue) =>
                  NoDup (fst (split svm0)) /\
                  Sorted_svm svm0 /\ IndAll Normal (snd (split svm0))) svm'
                 (alpham'_obligation_2 svm alpham' s v svm' Heq_svm))
              (alpham'_obligation_3 svm alpham' s v svm' Heq_svm))) as s0.
  destruct s0 as [am H]. simpl. unfold FiniteE. rewrite default_store.
  unfold FiniteE in H. apply H.
Qed.
*)


(* Initially wanted to write this but Program didn't like it 
fold_left (fun a sv => match sv with (s,v) => store a s (SomeValue v) end) svm (const string NoValue). *)

Definition sort_svm := insertion_sort le_sv le_sv_dec.

(* not needed
Definition le_str (s1 s2 : string) : Prop := cmp_str s1 s2 = Lt \/ cmp_str s1 s2 = Eq.
Definition Sorted_str (ss : list string) : Prop := Sorted le_str ss.
Definition le_str_dec:
  forall s1 s2 : string, {le_str s1 s2} + {~ le_str s1 s2}.
Admitted.
Definition sort_str := insertion_sort le_str le_str_dec.
*)


Lemma Permutation_fst_split : forall (X Y : Type) (xs ys : list (X * Y)),
  Permutation xs ys -> Permutation (fst (split xs)) (fst (split ys)).
Proof.
  intros X Y xs ys H. induction H.
  Case "perm_nil". simpl. apply perm_nil.
  Case "perm_skip".
    destruct x as [x y]. repeat rewrite fst_split_cons. apply perm_skip. assumption.
  Case "perm_swap". destruct x as [x1 y1]. destruct y as [x2 y2].
    repeat rewrite fst_split_cons. apply perm_swap.
  Case "perm_trans". eapply perm_trans. eassumption. assumption.
Qed.

Lemma Permutation_snd_split : forall (X Y : Type) (xs ys : list (X * Y)),
  Permutation xs ys -> Permutation (snd (split xs)) (snd (split ys)).
Proof.
  intros X Y xs ys H. induction H.
  Case "perm_nil". simpl. apply perm_nil.
  Case "perm_skip".
    destruct x as [x y]. repeat rewrite snd_split_cons. apply perm_skip. assumption.
  Case "perm_swap". destruct x as [x1 y1]. destruct y as [x2 y2].
    repeat rewrite snd_split_cons. apply perm_swap.
  Case "perm_trans". eapply perm_trans. eassumption. assumption.
Qed.

Lemma nodup_sort_remove_dup : forall svm,
  NoDup (fst (split (sort_svm (remove_dup eq_str_dec svm)))).
Proof.
  intros. eapply Permutation_NoDup. eapply Permutation_fst_split.
  eapply insertion_sort_permutation. apply remove_dup_sound.
Qed.

Lemma IndAll_Permutation : forall X (P : X -> Prop) xs ys,
  IndAll P xs -> Permutation ys xs -> IndAll P ys.
Proof.
  intros X P xs ys H1 H2. induction H2.
  Case "perm_nil". apply All_nil.
  Case "perm_skip". inversion H1. subst. apply All_cons. assumption.
    apply IHPermutation. assumption.
  Case "perm_swap". inversion H1. inversion H3. subst. apply All_cons. assumption.
    apply All_cons; assumption.
  Case "perm_trans". auto.
Qed.

Lemma IndAll_map : forall X Y (f : X -> Y) (P : Y -> Prop) xs,
  (forall x, In x xs -> P (f x)) ->  IndAll P (map f xs).
Proof.
  intros. induction xs as [| x xs']; simpl in *. apply All_nil.
  apply All_cons. apply H. left. reflexivity.
  apply IHxs'. intros. apply H. right. assumption.
Qed.

Lemma IndAll_forall : forall (X : Type) (P : X -> Prop) (xs : list X),
  IndAll P xs -> forall x, In x xs -> P x.
Admitted.

Lemma forall_IndAll : forall (X : Type) (P : X -> Prop) (xs : list X),
  (forall x, In x xs -> P x) -> IndAll P xs.
Admitted.

Lemma IndAll_remove_dup : forall (X Y : Type) (P : (X * Y) -> Prop) eqX_dec xys,
  IndAll P xys -> IndAll P (remove_dup eqX_dec xys).
Proof.
  intros X Y P eq_X_dec xys H.
  apply forall_IndAll. intros xy H'. destruct xy as [x y].
  eapply IndAll_forall in H. eassumption.
  eapply remove_dup_complete2. eassumption.
Qed.
    
Lemma IndAll_snd_split : forall (X Y : Type) (P : Y -> Prop) (xys : list (X * Y)),
  IndAll (fun xy => match xy with (x,y) => P y end) xys -> IndAll P (snd (split xys)).
Proof.
  intros X Y P xys H. induction xys as [| [x y] xys']. simpl. apply All_nil.
  rewrite snd_split_cons. inversion H. subst. apply All_cons. assumption.
  apply IHxys'. assumption.
Qed.

(* training for the big proof of betam
Lemma YYY : forall a : SVMapArray,
  IndAll Normal (List.map (fun (s : string) =>
                  let (x, _) := select cmp_str a s in
                  match x with
                  | Some v => v
                  | None => (G G_Null) (* this is unreachable, since x is in support *)
                  end) (support a)).
Proof. intro a. unfold SVMapArray in a.
  apply IndAll_map. intros s H.
    destruct (select cmp_str a s) as [rvo Hrvo]. destruct rvo as [rv |]. simpl in *. assumption.
    apply Normal_null.
Qed.
*)

(* using map instead of fold in the definition of betam helps a lot 
-- otherwise I would have needed to prove this lemma (I have no clue how)
Lemma fold_stuff : forall a : SVMapArray,
  IndAll Normal (fold_left (fun (vs : list RawValue) (s : string) =>
                  let (x, _ (* This is the point where a proof is dropped *)) := select cmp_str a s in
                  match x with
                  | Some v => v :: vs
                  | None => vs
                  end) (support a) nil).
*)

(* XXX This is probably the place where nesting proofs inside terms will harm us
-- the trick was to use map instead of fold *)
(** We assume that default svm = NoValue *)
(*--- BetamBegin *)
Definition betam (a : SVMapArray) : SVMap.
  refine (fun a =>
  exist (fun svm => NoDup (fst (split svm)) /\ Sorted_svm svm /\ IndAll Normal (snd (split svm)))
    (sort_svm (remove_dup eq_str_dec
      (List.map (fun s =>
                  match select cmp_str a s with
                  | exist (Some v) pf => (s, v)
                  | exist None pf => (s, (G G_Null)) (* XXX unreachable? *)
                  end) (support a)) )) _ ).
(*--- BetamEnd *)
Proof.
  split; try split.
  Case "NoDup". apply nodup_sort_remove_dup.
  Case "Sorted". apply (insertion_sort_sorted (string * RawValue) le_sv le_sv_trans le_sv_total).
  Case "IndAll". simpl. eapply IndAll_Permutation with (xs := 
       (snd
         (split
           (remove_dup eq_str_dec
              (map
                 (fun s : string =>
                  let (x, _) := select cmp_str a s in
                  match x with
                  | Some v => (s, v)
                  | None => (s, G G_Null)
                  end) (support a)))))).
    SCase "IndAll".
      apply IndAll_snd_split.
      apply IndAll_remove_dup.
      apply IndAll_map. intros.
      destruct (select cmp_str a x) as [rvo Hrvo]. destruct rvo as [rv |]. simpl in *. assumption.
      apply Normal_null.
    SCase "Permutation". apply Permutation_snd_split.
      apply Permutation_sym. apply insertion_sort_permutation.
Defined. (* Can't end with Qed, otherwise or will get "Error: betam is opaque." *)


(** Collections: VBagArray *)

(** XXX TODO: show (limited) isomorphism -- finite + positive VBagArray ~ VBag *)
(*--- AlphabBegin *)
Definition alphab (vb : VBag) : VBagArray. Admitted.
(*--- AlphabEnd *)

(*--- BetabBegin *)
Definition betab (vb : VBagArray) : VBag. Admitted.
(*--- BetabEnd *)


(** Constants *)

Obligation Tactic := (intros; apply normal_G).
(*--- GeneralConstrBegin *)
Program Definition v_tt : Value := G (G_Logical true).
Program Definition v_ff : Value := G (G_Logical false).
Program Definition v_logical (b : bool) : Value := G (G_Logical b).
Program Definition v_int i : Value := G (G_Integer i).
Program Definition v_text s : Value := G (G_Text s).
Program Definition v_null : Value := G (G_Null).
(*--- GeneralConstrEnd *)
Obligation Tactic := program_simpl. (* reverting to default *)

(*--- GeneralTestersBegin *)
Definition In_Logical v := (is_G v) && is_G_Logical (out_G v).
Definition In_Integer v  := (is_G v) && is_G_Integer (out_G v).
Definition In_Text v := (is_G v) && is_G_Text (out_G v).
(*--- GeneralTestersEnd *)

Definition In_E nv := is_E nv.
Definition In_C nv := is_C nv.


(** XXX For Results we are nesting proofs inside terms
-- fortunately, results are not used in our optimized semantics + axioms
*)
(*--- ResultBegin *)
Inductive Result : Type :=
  | Error : Result
  | Return : Value -> Result.
(*--- ResultEnd *)
(*--- ProperBegin *)
Definition Proper (res : Result) :=
  match res with | Return v => true | Error => false end.
Definition out_V (res : Result) : Value :=
  match res with | Return v => v | Error => v_null  end.
(*--- ProperEnd *)



(** In the model all values are Good *)
(*--- AllGoodBegin *)
Definition Good (v : Value) := true.
(*--- AllGoodEnd *)

(*--- GoodCEBegin *)
Definition Good_C := is_C.
Definition Good_E := is_E.
(*--- GoodCEEnd *)


Program Definition syn_beq_val (v1 v2 : Value) := syn_beq_rval v1 v2.
Lemma syn_beq_val_refl : forall v, syn_beq_val v v = true.
Proof.
 intros. unfold syn_beq_val. apply syn_beq_rval_refl.
Qed.
Lemma syn_beq_val_sym : forall v1 v2, syn_beq_val v1 v2 = syn_beq_val v2 v1.
Admitted.


(*--- OperatorsBegin *)
Definition O_Sum v1 v2 :=
  v_int (Zplus (of_G_Integer(out_G v1)) (of_G_Integer(out_G v2))).
Definition O_Minus v1 v2 :=
  v_int (Zminus (of_G_Integer(out_G v1)) (of_G_Integer(out_G v2))).
Definition O_Mult v1 v2 :=
  v_int (Zmult (of_G_Integer(out_G v1)) (of_G_Integer(out_G v2))).
Definition O_GT v1 v2 :=
  match Zcompare (of_G_Integer(out_G v1)) (of_G_Integer(out_G v2))
  with Gt => v_tt | _ => v_ff end.
Definition O_LT v1 v2 :=
  match Zcompare (of_G_Integer(out_G v1)) (of_G_Integer(out_G v2))
  with Lt => v_tt | _ => v_ff end.
Definition O_EQ v1 v2 := v_logical (syn_beq_val v1 v2).
Definition O_Not v := v_logical (negb (of_G_Logical (out_G v))).
Definition O_And v1 v2 := v_logical (andb (of_G_Logical (out_G v1))
                                        (of_G_Logical (out_G v2))).
Definition O_Or v1 v2 := v_logical (orb (of_G_Logical (out_G v1))
                                      (of_G_Logical (out_G v2))).
(*--- OperatorsEnd *)
Definition O_Append v1 v2 :=
  v_text (append (of_G_Text(out_G v1)) (of_G_Text(out_G v2))).

(** ** Entities *)
Require Import TheoryList.
Require Import Coq.Bool.Sumbool.

Lemma assoc_some : forall (X Y : Type) eq_f (x : X) (xys : list (X*Y)) (y : Y),
  Some y = assoc eq_f x xys ->
    exists x', (exists pf, eq_f x x' = left (x <> x') pf) /\ In (x',y) xys.
Proof.
  intros X Y eq_f x xys y H. induction xys as [| [x' y'] xys'].
  Case "xys=nil". simpl in H. inversion H.
  Case "xys=(x',y') :: xys'". simpl in H. unfold value in H.
    remember (eq_f x x') as eqxx'. destruct eqxx' as [H0 | H0].
    SCase "x = x'". exists x'. split.
      SSCase "TS1". exists H0. rewrite Heqeqxx'. reflexivity.
      SSCase "TS2". simpl in *. inversion H. rewrite <- H2. left. reflexivity.
    SCase "x <> x'". simpl in *. apply IHxys' in H. destruct H as [x0 [H1 H2]].
      exists x0. split. assumption. right. assumption.
Qed.


Lemma pair_in_snd : forall (X Y : Type) x y (xys : list (X * Y)),
  In (x,y) xys -> In y (snd (split xys)).
Proof.
  intros X Y x y xys H. induction xys as [ | [x' y'] xys']. inversion H.
    simpl in H. rewrite snd_split_cons. simpl. destruct H.
      inversion H. subst. left. reflexivity.
      right. apply IHxys'. assumption.
Qed.

(*--- EntitiesBegin *)
Program Definition v_has_field (s : string) (v : Value) : bool :=
  match TheoryList.assoc eq_str_dec s (out_E v) with
  | Some v => true | None => false end.
Program Definition v_dot (s : string) (v : Value) : Value :=
  match TheoryList.assoc eq_str_dec s (out_E v) with
  | Some v => v | None => v_null end.
(*--- EntitiesEnd *)
Next Obligation.
  destruct v as [v1 pf]. destruct v1; simpl in *; try inversion Heq_anonymous.
  apply assoc_some in H0. destruct H0 as [s' [H1 H2]]. destruct H1 as [Hs H1'].
  inversion pf. clear H1'. rewrite <- H in *. clear H. subst.
  apply pair_in_snd in H2. apply All_In with (xs := snd (split svm)); assumption.
Qed.

(*--- EmptyEntityBegin *)
Program Definition v_eempty : Value := E nil.
(*--- EmptyEntityEnd *)
Next Obligation.
  apply normal_E; simpl; try apply NoDup_nil; try apply All_nil; try apply Sorted_nil.
Qed.


Lemma not_or : forall (P Q : Prop), ~(P \/ Q) -> ~P /\ ~Q.
Proof. intros P Q H. split; intro H'; apply H. left. assumption. right. assumption. Qed.

Lemma if_dec_right : forall (A B : Prop) (C : Type) (H : {A} + {B}) (x y : C),
  ~A -> (if H then x else y) = y.
Proof. intros. destruct H. apply H0 in a. inversion a. reflexivity. Qed.

(** The following proofs are related to "completeness" of duplicate elimination
-- only duplicates are removed, nothing else

XXX One reasons the proofs are so long and messy is that at the time we did them
we did't yet fully characterize what remove_dup does
-- now we have a better characterization: remove_dup_complete1 and remove_dup_complete2
   TODO: see if any of the proofs here can be simplified
*)

Lemma avoid_exist_false : forall avoid s, (forall s0 : string, In s0 avoid -> s <> s0) -> 
  existsb (fun xa : string => if eq_str_dec xa s then true else false) avoid = false.
Admitted.

Definition update_in_sorted_svm := update_in_sorted le_sv le_sv_dec.

Lemma le_sv_asymm : forall s v s' v',
  le_sv (s, v) (s', v') ->
  le_sv (s', v') (s, v) ->
  s = s'.
Admitted.

Lemma le_sv_not : forall s v s' v',
  ~ le_sv (s, v) (s', v') ->
  s <> s'.
Admitted.

Lemma in_fst_split : forall X Y (x : X) (y : Y) xys,
  In (x, y) xys ->
  In x (fst (split xys)).
Admitted.

Lemma in_snd_split : forall X Y (x : X) (y : Y) xys,
  In (x, y) xys ->
  In y (snd (split xys)).
Admitted.

Lemma in_fst_split_exists : forall X Y (x : X) (xys : list (X * Y)),
  In x (fst (split xys)) ->
  exists y, In (x,y) xys.
Admitted.

Lemma in_snd_split_exists : forall X Y (y : Y) (xys : list (X * Y)),
  In y (snd (split xys)) ->
  exists x, In (x,y) xys.
Admitted.

Lemma not_le_sv_leAll_not_In_fst_split : forall s v s' v' svm',
  ~le_sv (s',v') (s,v) ->
  leAll _ le_sv (s',v') svm' ->
  ~In s (fst (split svm')).
Proof.
  intros s v s' v' svm' H2 H3 H4. unfold leAll in H3.
  apply in_fst_split_exists in H4. destruct H4 as [v0 H4]. specialize (H3 _ H4).
  unfold le_sv in H2, H3. apply H2. assumption.
Qed.

Lemma cmp_str_refl : forall s,
  cmp_str s s = Eq.
Admitted.

Lemma In_All : forall (X : Type) (P : X -> Prop) xs,
  (forall x, In x xs -> P x) -> IndAll P xs.
Admitted.

Theorem update_in_sorted_NoDup : forall s v svm,
  NoDup (fst (split svm)) ->
  Sorted_svm svm ->
  NoDup (fst (split (update_in_sorted_svm (s,v) svm))).
Proof.
  intros. induction svm as [| [s' v'] svm'].
  Case "nil". simpl. constructor. intro Hc. inversion Hc. apply NoDup_nil. 
  Case "(s',v') :: svm'".
    rewrite fst_split_cons in H. simpl in *.
    destruct (le_sv_dec (s,v) (s',v')) as [H1 | H1].
    SCase "le_sv (s,v) (s',v')".
      destruct (le_sv_dec (s',v') (s,v)) as [H2 | H2].
      SSCase "le_sv (s',v') (s,v)".
        rewrite fst_split_cons. 
        assert(s = s'). apply le_sv_asymm with v v'; assumption. subst.
        apply H.
      SSCase "~le_sv (s',v') (s,v)".
        do 2 rewrite fst_split_cons.
        remember H2 as H2'. clear HeqH2'. apply le_sv_not in H2.
        constructor; [| assumption]. intro Hc. simpl in Hc. destruct Hc.
          subst. apply H2. reflexivity.
          inversion H0; subst.
          eapply not_le_sv_leAll_not_In_fst_split; eassumption.
    SCase "le_sv (s',v') (s,v)".
      inversion H. inversion H0. subst. specialize (IHsvm' H5 H9).
      rewrite fst_split_cons. constructor; [| apply IHsvm'].
      remember H1 as H1'. clear HeqH1'. apply le_sv_not in H1.
      intro Hc. apply in_fst_split_exists in Hc. destruct Hc as [v0 Hc].
      assert(In (s',v0) ((s,v) :: svm')) as Hc'.
        eapply update_in_sorted_sub_elements. apply Hc.
      simpl in Hc'. destruct Hc'.
      SSCase "1".
        inversion H2; subst. simpl in H1'. apply not_or in H1'.
        destruct H1' as [_ H1']. apply H1'. apply cmp_str_refl.
      SSCase "2".
        apply in_fst_split in H2. apply H4. apply H2.
Qed.

Lemma normal_E_sing : forall s v, Normal v -> Normal (E ((s, v) :: nil)).
Proof.
  intros. apply normal_E; simpl. apply NoDup_cons. intro Hc. inversion Hc. apply NoDup_nil.
  apply Sorted_sing.
  apply All_cons. assumption. apply All_nil.
Qed.

(*--- UpdateEntityBegin *)
Program Definition v_eupdate (s : string) (v e : Value) : Value :=
  E (update_in_sorted_svm (s, v) (out_E e)).
(*--- UpdateEntityEnd *)
Next Obligation.
  destruct v as [v Hv].
  destruct e as [e He]. simpl.
  (* The only interesting case is when e is an entitiy *)
  destruct e; try (simpl; apply normal_E_sing; assumption).
  simpl. inversion He. symmetry in H. subst.
  apply normal_E.
  Case "TS1 NoDup".
    apply update_in_sorted_NoDup; assumption.
  Case "TS2 Sorted_svm".
    apply (update_in_sorted_sorted _ le_sv le_sv_trans le_sv_total). assumption.
  Case "TS3 IndAll".
      assert (IndAll Normal (snd (split ((s,v) :: svm)))) as H.
        rewrite snd_split_cons. apply All_cons; assumption.
      rewrite snd_split_cons in H.
      inversion H; subst.
      apply In_All. intros v' H'.
      apply in_snd_split_exists in H'. destruct H' as [s' H'].
      assert(In (s',v') ((s,v) :: svm)) as H''.
        eapply update_in_sorted_sub_elements. apply H'.
      simpl in H''. destruct H''. inversion H3. subst. assumption.
      apply in_snd_split in H3.
      eapply All_In in H. eassumption. simpl. right. assumption.
Qed.

(** Collections *)

Lemma eq_rval_dec : forall v1 v2 : RawValue, {v1 = v2} + {v1 <> v2}.
Admitted.

Require Import Coq.Program.Equality.

(* XXX Doesn't Coq.Init.Specif or Coq.Program already prove any such lemma?
-- proof crucially depends on proof irrelevance *)
Lemma eq_exist : forall X (x : X) (P : X -> Prop) px py,
  exist (fun x => P x) x px = exist (fun x => P x) x py.
Proof. intros. rewrite proof_irrelevance with (p1 := px) (p2 := py). reflexivity. Qed.

Lemma eq_exist2 : forall X (x y : X) (P : X -> Prop) px py, x = y ->
  exist (fun x => P x) x px = exist (fun x => P x) y py.
Proof. intros. subst. apply eq_exist. Qed.

Lemma exist_forall : forall X (P : X -> Prop) (x : X) (px : P x) (Q : {x : X | P x} -> Prop),
  Q (exist (fun x => P x) x px) -> forall pf' : P x, Q (exist (fun x => P x) x pf').
Proof. intros. erewrite eq_exist. eassumption. Qed.

Lemma not_eq_exist : forall X (x y : X) (P : X -> Prop) px py, x <> y ->
  ~(exist (fun x => P x) x px = exist (fun x => P x) y py).
Proof. intros. intro Hc. inversion Hc. apply H. assumption. Qed.

Lemma eq_val_dec : forall v1 v2 : Value, {v1 = v2} + {v1 <> v2}.
Proof. intros. destruct v1 as [rv1 Hv1]. destruct v2 as [rv2 Hv2].
  remember (eq_rval_dec rv1 rv2) as rv1rv2. destruct rv1rv2.
  left. apply eq_exist2. assumption.
  right. apply not_eq_exist. assumption.
Qed.

Definition insert_in_sorted_vb := insert_in_sorted le_rval le_rval_dec.

Section Remove1.

    Variable A : Type.
    Hypothesis eq_dec : forall x y : A, {x = y}+{x <> y}.

    Fixpoint remove1 (x : A) (l : list A){struct l} : list A :=
      match l with
        | nil => nil
        | y::tl => if (eq_dec x y) then tl else y::(remove1 x tl)
      end.

    Lemma remove1_In : forall x xs y,
      In y (remove1 x xs) -> In y xs.
    Proof.
      intros.
      induction xs.
      SCase "nil".
        simpl in *. assumption.
      SCase "a::vb".
        simpl. remember (eq_dec x a) as eq. destruct eq; intros. simpl in H.
        rewrite <- Heqeq in H. right. assumption.
        simpl in H. rewrite <- Heqeq in H. simpl in H. destruct H.
        left. assumption.
        right. apply IHxs. assumption.
    Qed.

End Remove1.
Implicit Arguments remove1 [A].


Theorem remove1_sorted : forall x xs,
  Sorted_vb xs -> Sorted_vb (remove1 eq_rval_dec x xs).
Proof.
  intros. induction xs as [| x' xs']; simpl.
  Case "xs = nil". apply Sorted_nil.
  Case "xs = x' :: xs'".
    case (eq_rval_dec x x'); intro H'.
    SCase "le_dec x x' = left _".
      inversion H. assumption.
    SCase "le_dec x x' = right _".
      inversion H. subst. apply IHxs' in H3.
      apply Sorted_cons; try assumption.
      
      unfold leAll. unfold leAll in H2. intros. specialize (H2 y).
      apply H2.
      apply remove1_In in H0. assumption.
Qed.

(*--- CollectionsBegin *)
Program Definition v_mem (v cv : Value) : bool :=
  TheoryList.mem eq_rval_dec v (out_C cv).
Program Definition v_add (v cv : Value) : Value :=
  (C (insert_in_sorted_vb v (out_C cv))). 
(*--- CollectionsEnd *)
Next Obligation.
  apply normal_C.
  Case "TS1: Sorted". apply (insert_in_sorted_sorted RawValue le_rval le_rval_trans le_rval_total).
    destruct cv as [c pf]. inversion pf; subst; simpl; try apply Sorted_nil. assumption.
  Case "TS2: IndAll".
    destruct cv as [c pc]. destruct v as [v pv].
    inversion pc; subst; simpl; try apply All_cons; try apply All_nil; try assumption.
    assert (IndAll Normal (v::vb)). apply All_cons; assumption. clear H0.
    eapply IndAll_Permutation. eassumption. unfold insert_in_sorted_vb.
    apply Permutation_sym. apply insert_in_sorted_permutation.
Qed.
Program Definition v_empty (cv : Value) : Value :=
  (v_logical (beq_nat (List.length (out_C cv)) 0)).
Program Definition v_remove (v cv : Value) : Value :=
  (C (remove1 eq_rval_dec v (out_C cv))).
Next Obligation.
  apply normal_C.
  Case "TS1: Sorted". apply remove1_sorted.
    destruct cv as [c pf]. inversion pf; subst; simpl; try apply Sorted_nil. assumption.
  Case "TS2: IndAll".
    destruct cv as [c pc]. destruct v as [v pv].
    inversion pc; subst; simpl; try apply All_cons; try apply All_nil; try assumption.
    clear pc. clear H. induction vb.
    SCase "nil".
      simpl. apply All_nil.
    SCase "a::vb".
      simpl. case (eq_rval_dec); intros. inversion H0. assumption.
      apply All_cons. inversion H0. assumption.
      apply IHvb. inversion H0. assumption.
Defined.


(*--- AddManyBegin *)
Fixpoint v_add_many' (v : Value) (n : nat) (cv : Value) : Value :=
  match n with 0 => cv | S n' => v_add_many' v n' (v_add v cv) end.
Definition v_add_many (v : Value) (i : Z) (cv : Value) : Value :=
  v_add_many' v (Zabs_nat i) cv.
(*--- AddManyEnd *)


(*--- ClosureRes2Begin *)
Definition ClosureRes2 := Value -> Value -> Result.
(*--- ClosureRes2End *)

(*--- ResAccFoldBegin *)
Program Fixpoint res_acc_fold (f : ClosureRes2) (vb : VBag) (a : Result) {measure List.length vb} : Result :=
  match vb with
  | nil => a
  | v :: vb' => match a with Return va => res_acc_fold vb' (f va v)
                | Error => Error end
  end.
(*--- ResAccFoldEnd *)
Next Obligation.
  destruct vb as [vb1 Hvb]. simpl in Heq_vb. inversion Hvb. subst.
  split. inversion H.  assumption. inversion H0.  assumption.
Qed.
Next Obligation.
  destruct vb as [vb1 Hvb]. simpl in Heq_vb. inversion Hvb. subst. inversion H0. assumption.
Qed.
(*--- ResAccumulateBegin *)
Definition res_accumulate (f : ClosureRes2) (cv v : Value) : Result :=
  if is_C cv then res_acc_fold f (out_C cv) (Return v) else Error.
(*--- ResAccumulateEnd *)


Lemma C_nil_normal : Normal (C nil).
Proof. apply normal_C. apply Sorted_nil. apply All_nil. Qed.

Lemma normal_C_sort_all_normal : forall vb,
  Normal (C vb) -> IndAll Normal (insertion_sort le_rval le_rval_dec vb).
Proof.
    intros vb H. inversion H.
    eapply IndAll_Permutation. eassumption. unfold insert_in_sorted_vb.
    apply Permutation_sym. apply insertion_sort_permutation.
Qed.

Lemma All_app : forall X (P : X -> Prop) xs ys,
  IndAll P xs -> IndAll P ys -> IndAll P (xs ++ ys).
Proof.
  intros. induction xs as [| x xs']; simpl. assumption.
  inversion H. apply All_cons. assumption.
  apply IHxs'. assumption.
Qed. 

(*--- ZeroBegin *)
Program Definition v_zero : Value := C nil.
(*--- ZeroEnd *)
Next Obligation. apply normal_C. apply Sorted_nil. apply All_nil. Qed.
Program Definition v_unit (v : Value) : Value := C (v :: nil).
Next Obligation. apply normal_C. apply Sorted_sing.
  apply All_cons. destruct v. simpl. assumption. apply All_nil. Qed.
Definition insertion_sort_vb := insertion_sort le_rval le_rval_dec.
Program Definition v_plus (c1 c2 : Value) : Value :=
  C (insertion_sort_vb ((out_C c1) ++ (out_C c2))).
Next Obligation.
  destruct c1 as [vc1 pc1]. destruct c2 as [vc2 pc2].
  destruct vc1 as [g1 | svm1 | vb1]; destruct vc2 as [g2 | svm2 | vb2];
  simpl; try rewrite <- app_nil_end; try apply C_nil_normal;
    apply normal_C; unfold insertion_sort_vb;
    try apply (insertion_sort_sorted RawValue le_rval le_rval_trans le_rval_total);
    try (apply normal_C_sort_all_normal; assumption).
  Case "only interesting case, both v1 and v2 are collections".
    apply IndAll_Permutation with (xs:=vb1++vb2);
      try (apply Permutation_sym; apply insertion_sort_permutation).
    inversion pc1. inversion pc2. eapply All_app; assumption.
Qed.

(** More on closures *)
Definition Closure := Value -> Value.
Definition v_apply (c : Closure) v := c v.

(*--- Closure2Begin *)
Definition Closure2 := Value -> Value -> Value.
Definition v_apply2 (c : Closure2) v1 v2 := c v1 v2.
(*--- Closure2End *)

(** XXX: what's with the closure_tag? *)


(*--- VAccFoldBegin *)
Program Fixpoint v_acc_fold (f : Closure2) (vb : VBag) (a : Value)
  {measure List.length vb} : Value :=
  match vb with nil => a | v :: vb' => v_acc_fold vb' (f a v) end.
(*--- VAccFoldEnd *)
Next Obligation. destruct vb as [vb0 Hvb0]. simpl in *. subst. destruct Hvb0 as [H1 H2].
  split. inversion H1. assumption. inversion H2. assumption. Qed.
Next Obligation. destruct vb as [vb0 Hvb0]. simpl in *. subst. destruct Hvb0 as [H1 H2].
  inversion H2. assumption. Qed.
(*--- VAccumulateBegin *)
Definition v_accumulate (clos:Closure2) c := v_acc_fold clos (out_C c).
(*--- VAccumulateEnd *)

(** XXX See "More stuff on collections" in ModelVariant1.v *)

Program Definition v_count c := Z_of_nat (List.length (out_C c)).

(* Equalty of values, second part *)
Lemma syn_beq_val_equals : forall v1 v2, syn_beq_val v1 v2 = true ->
  v1 = v2.
Proof.
  intros. unfold syn_beq_val in H. destruct v1. destruct v2.
  simpl in H. apply syn_beq_rval_equals in H. apply eq_exist2. assumption.
Qed.
