(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

Require Import List.
Require Import Model.
Require Import Calculus.
Require Import LogicalSemantics.


Lemma syn_beq_res_true_iff : forall r1 r2,
  r1 = r2 <-> syn_beq_res r1 r2 = true.
Admitted.

Lemma env_update_eq : forall x v env,
  env_update env x v x = v.
Admitted.

Lemma env_update_neq : forall x2 x1 v env,
  x1 <> x2 ->
  env_update env x2 v x1 = env x1.
Admitted.


(* XXX
An initial definition of algorithmic purity.
Accumulates are pure when their body is order-preserving:
SWAPPING: forall x y z, f(x,f(y,z)) = f(y,f(x,z)) 
-- this is all we need! we don't need associativity + commutativity
Function applications are not yet allowed in algorithmically pure expressions
*)


Fixpoint is_pure (e : Exp) {struct e} : bool :=
  match e with
  | E_Var x => true (* They can only be substituted by values, which are pure *)
  | E_Value v => true
  | E_UnOp o e => is_pure e
  | E_BinOp o e1 e2 => is_pure e1 && is_pure e2
  | E_If e1 e2 e3 => is_pure e1 && is_pure e2 && is_pure e3
  | E_Let x e1 e2 => is_pure e1 && is_pure e2
  | E_In e T => is_pure e && negb (contains_impure_expressions T)
  | E_Entity les => true
  | E_Dot e l => is_pure e
  | E_Add e1 e2 => is_pure e1 && is_pure e2
  | E_Acc x e1 y e2 e3 => 
     is_pure e1 && is_pure e2 && is_pure e3 &&
     (* Swapping *)
     forall_bool (fun env =>
     forall_bool (fun v1 => forall_bool (fun v2 => forall_bool (fun v3 =>
       syn_beq_res
         (LBind (R e3 (env_update (env_update env x v2) y v3)) (fun z =>
                (R e3 (env_update (env_update env x v1) y z))))
         (LBind (R e3 (env_update (env_update env x v1) y v3)) (fun z =>
                (R e3 (env_update (env_update env x v2) y z))))
     ))))
  | E_App _ es => false
  end
with contains_impure_expressions T :=
  match T with
  | T_Refine x T1 e => negb (is_pure e) || contains_impure_expressions T1
  | T_Any => false
  | T_Integer => false
  | T_Text => false
  | T_Logical => false
  | T_Coll T => contains_impure_expressions T
  | T_Entity l T => contains_impure_expressions T
  end.

Fixpoint contains_impure_refinements e :=
  match e with
  | E_In e1 T => contains_impure_refinements e1 || contains_impure_expressions T
  | E_Var x => false
  | E_Value v => false
  | E_UnOp o e => contains_impure_refinements e 
  | E_BinOp o e1 e2 => contains_impure_refinements e1 || contains_impure_refinements e2
  | E_If e1 e2 e3 => contains_impure_refinements e1 || contains_impure_refinements e2 || contains_impure_refinements e3
  | E_Let x e1 e2 => contains_impure_refinements e1 || contains_impure_refinements e2
  | E_Entity les => 
    (fix cir_entity les :=
      match les with
      | nil => false
      | (_, e) :: les' => contains_impure_refinements e || (cir_entity les')
      end) les
  | E_Dot e l => contains_impure_refinements e
  | E_Add e1 e2 => contains_impure_refinements e1 || contains_impure_refinements e2
  | E_Acc x e1 y e2 e3 => contains_impure_refinements e1 || contains_impure_refinements e2 ||
    contains_impure_refinements e3
  | E_App _ e => contains_impure_refinements e
  end.


Lemma count_pure : forall e x y,
  is_pure e = true ->
  x <> y ->
  is_pure (E_Acc x e y (E_Value (v_int 0)) (E_BinOp OPlus (E_Var y) (E_Value (v_int 1)))) = true.
Proof.
  intros e x y H1 H2. simpl. rewrite H1. simpl.
  repeat setoid_rewrite <- forall_bool_true_iff. intros env v1 v2 v3.
  rewrite <- syn_beq_res_true_iff. unfold LBind. simpl.
  repeat rewrite env_update_eq.
  destruct v1 as [[[i1 | s1 | b1 |] | svm1 | vb1] pv1];
  destruct v2 as [[[i2 | s2 | b2 |] | svm2 | vb2] pv2]; simpl; try reflexivity.
Qed.






Lemma values_pure : forall v,
  is_pure (E_Value v) = true.
Proof. trivial. Qed.


Ltac destruct_if :=
  match goal with
    | [ |- context[if ?X then _ else _] ] => destruct X
  end.

Ltac destruct_if_syn_beq_val :=
  match goal with
    | [ |- context[if syn_beq_val ?X ?Y then _ else _] ] => destruct (syn_beq_val X Y)
  end.

Lemma count_occurs_pure : forall e1 e2 x y,
  is_pure e1 = true ->
(*  is_value_ind e2 -> -- this causes the state space to explode *)
  e2 = E_Value (v_int 0) ->
  x <> y ->
  is_pure (E_Acc x e1 y (E_Value (v_int 0))
    (E_If (E_Eq (E_Var x) e2) (E_BinOp OPlus (E_Var y) (E_Value (v_int 1))) (E_Var y))) = true.
Proof.
  intros e1 e2 x y H1 H2 H3. simpl. rewrite H1.
(*  remember H2 as H2'. clear HeqH2'. apply values_pure' in H2'. rewrite H2'. *)
  rewrite H2.
  simpl.
  repeat setoid_rewrite <- forall_bool_true_iff. intros env v1 v2 v3.
  rewrite <- syn_beq_res_true_iff. unfold LBind. simpl.
  repeat rewrite env_update_eq.
  repeat (rewrite env_update_neq; try assumption; repeat rewrite env_update_eq).
(*  destruct H2; simpl. *)
(* manual
  destruct (syn_beq_val (O_EQ v1 (G' (G_Integer 0))) v_tt);
  destruct (syn_beq_val (O_EQ v1 (G' (G_Integer 0))) v_ff);
  destruct (syn_beq_val (O_EQ v2 (G' (G_Integer 0))) v_tt);
  destruct (syn_beq_val (O_EQ v2 (G' (G_Integer 0))) v_ff);
*)
  repeat destruct_if_syn_beq_val; simpl; try auto;
  destruct v1 as [[[i1 | s1 | b1 |] | svm1 | vb1] pv1];
  destruct v2 as [[[i2 | s2 | b2 |] | svm2 | vb2] pv2];
  destruct v3 as [[[i3 | s3 | b3 |] | svm3 | vb3] pv3]; try reflexivity.
Qed.

Ltac destruct_Proper_R :=
  match goal with
    | [ |- context[if negb (Proper (R ?X ?Y)) then _ else _] ] => destruct (Proper (R X Y))
  end.

Lemma v_add_swap : forall v1 v2 v3,
  (v_add v1 (v_add v2 v3)) =
  (v_add v2 (v_add v1 v3)).
Admitted.

(* Filtering out values that don't fulfill a condition: from-where expression *)
Lemma from_where_pure : forall e1 e2 x y,
  is_pure e1 = true ->
  is_pure e2 = true ->
  x <> y ->
  is_pure (E_Acc x e1 y (E_Value (v_zero)) (E_If (e2) (E_Add (E_Var x) (E_Var y)) (E_Var y))) = true.
Proof.
  intros e1 e2 x y H1 H2 H3. simpl. rewrite H1. rewrite H2. simpl.
  repeat setoid_rewrite <- forall_bool_true_iff. intros env v1 v2 v3.
  rewrite <- syn_beq_res_true_iff. unfold LBind. simpl.
  repeat rewrite env_update_eq.
(* XXX exponential state-space, things blow up, Windows kills CoqIde for being "inactive"
-- still, I think this should work
  repeat (rewrite env_update_neq; try assumption; repeat rewrite env_update_eq).
  repeat destruct_if_syn_beq_val; repeat destruct_Proper_R; simpl; try auto;
  destruct v1 as [[[i1 | s1 | b1 |] | svm1 | vb1] pv1];
  destruct v2 as [[[i2 | s2 | b2 |] | svm2 | vb2] pv2];
  destruct v3 as [[[i3 | s3 | b3 |] | svm3 | vb3] pv3]; simpl; try reflexivity;
  try (rewrite v_add_swap; reflexivity).
*)
Admitted.

Definition E_Plus e1 e2 := 
  let (x, _) := fresh (fv_exp e1 ++ fv_exp e2) in
  let (y, _) := fresh (x :: (app (fv_exp e1) (fv_exp e2))) in
  (E_Acc x e1 y e2 (E_Add (E_Var x) (E_Var y))).


Lemma plus_pure : forall e1 e2,
  is_pure e1 = true ->
  is_pure e2 = true ->
  is_pure (E_Plus e1 e2) = true.
Proof.
  intros e1 e2 H1 H2. unfold E_Plus.
  destruct (fresh (fv_exp e1 ++ fv_exp e2)) as [x Hx].
  destruct (fresh (x :: (app (fv_exp e1) (fv_exp e2)))) as [y Hy].
  assert(x <> y). simpl in Hy. intro Hc. subst. apply Hy. left. reflexivity.
  simpl. rewrite H1. rewrite H2. simpl.
  repeat setoid_rewrite <- forall_bool_true_iff. intros env v1 v2 v3.
  rewrite <- syn_beq_res_true_iff. unfold LBind. simpl.
  repeat rewrite env_update_eq.
  repeat (rewrite env_update_neq; try assumption; repeat rewrite env_update_eq).
  destruct v1 as [[[i1 | s1 | b1 |] | svm1 | vb1] pv1];
  destruct v2 as [[[i2 | s2 | b2 |] | svm2 | vb2] pv2];
  destruct v3 as [[[i3 | s3 | b3 |] | svm3 | vb3] pv3]; simpl; try reflexivity;
  try (rewrite v_add_swap; reflexivity).
Qed.

(* we'll need something like this below
Lemma plus_swap : forall v1 v2 v3,
  (R (plus v1 (plus v2 v3))) =
  (R (plus v2 (plus v1 v3))).
*)

Definition E_Bind x e1 e2 := 
  let (y, _) := fresh (x :: (app (fv_exp e1) (fv_exp e2))) in
  (E_Acc x e1 y (E_Value v_zero) (E_Plus e2 (E_Var y))).


Lemma bind_pure : forall x e1 e2,
  is_pure e1 = true ->
  is_pure e2 = true ->
  is_pure (E_Bind x e1 e2) = true.
Proof.
  intros x e1 e2 H1 H2. unfold E_Bind.
  destruct (fresh (x :: (app (fv_exp e1) (fv_exp e2)))) as [y Hy].
  assert(x <> y). simpl in Hy. intro Hc. subst. apply Hy. left. reflexivity.
  simpl. rewrite H1. rewrite plus_pure; try (simpl; trivial).
  repeat setoid_rewrite <- forall_bool_true_iff. intros env v1 v2 v3.
  rewrite <- syn_beq_res_true_iff. unfold LBind. simpl.
  repeat rewrite env_update_eq.
  repeat (rewrite env_update_neq; try assumption; repeat rewrite env_update_eq).
  repeat destruct_Proper_R;
  destruct v1 as [[[i1 | s1 | b1 |] | svm1 | vb1] pv1];
  destruct v2 as [[[i2 | s2 | b2 |] | svm2 | vb2] pv2];
  destruct v3 as [[[i3 | s3 | b3 |] | svm3 | vb3] pv3]; simpl; try reflexivity.
Admitted. (* XXX TBD *)

