(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)


Require Export Case.

(** ** Scalar Values *)
Require Import String.
Require Import ZArith. (* integers: Z *)


(*
Require Import Int.
Module Raw (I:Int)(X:OrderedType). Import I. Module II:=MoreInt(I). Import II.
Open Scope Int_scope.
*)

(*--- GeneralBegin *)
Inductive General : Type :=
  | G_Integer : Z -> General
  | G_Text : string -> General
  | G_Logical : bool -> General
  | G_Null : General.
(*--- GeneralEnd *)

(** ** Identifiers and Strings *)

(** Identifiers are represented by natural numbers, using the wrapper [Id]. 
 Two identifiers are equal if and only if they are represented by the same natural number. *)

Require Import Arith.

Inductive id : Type := 
  Id : nat -> id.

Definition beq_id id1 id2  := 
  match (id1, id2) with (Id n1, Id n2) => beq_nat n1 n2 end.
Hint Unfold beq_id. 

Theorem beq_id_refl : forall i,
  true = beq_id i i.
Proof. 
intros. destruct i. unfold beq_id. apply beq_nat_refl. 
Qed.

Theorem not_eq_beq_false : forall n n' : nat,
     n <> n'
  -> beq_nat n n' = false.
Proof. 
  intros n n' H.
  unfold not in H.
  remember (beq_nat n n') as nn'.
  destruct nn'.
  Case "nn' = true". apply False_ind. apply H. apply beq_nat_true.
    rewrite <- Heqnn'. reflexivity.
  Case "nn' = false". reflexivity.
Qed.

Theorem not_eq_beq_id_false : forall i1 i2,
  i1 <> i2 -> beq_id i1 i2 = false.
Proof.
  intros i1 i2 H.
  destruct i1 as [n1].
  destruct i2 as [n2].
  unfold beq_id.
  apply not_eq_beq_false.
  intro Hc.
  rewrite Hc in H.
  apply H.
  reflexivity.
Qed.

Theorem beq_id_eq : forall i1 i2,
  true = beq_id i1 i2 -> i1 = i2.
Proof.
  intros i1 i2 H.
  destruct i1 as [n1].
  destruct i2 as [n2].
  unfold beq_id in H. apply beq_nat_eq in H. rewrite H. reflexivity.
Qed.

Theorem beq_id_false_not_eq : forall i1 i2,
  beq_id i1 i2 = false -> i1 <> i2.
Proof.
  intros i1 i2 H Hc.
  destruct i1 as [n1].
  destruct i2 as [n2].
  unfold beq_id in H. apply beq_nat_false in H.
  inversion Hc as [Hc'].
  apply H. apply Hc'.
Qed.

(** String Equality *)

Require Import Ascii.

Fixpoint beq_ascii (c1 c2 : ascii) {struct c1} : bool :=
  match c1, c2 with
  | Ascii b0 b1 b2 b3 b4 b5 b6 b7, Ascii b8 b9 b10 b11 b12 b13 b14 b15
      => (eqb b0 b8) && (eqb b1 b9) && (eqb b2 b10) && (eqb b3 b11)
          && (eqb b4 b12) && (eqb b5 b13) && (eqb b6 b14) && (eqb b7 b15)
  end.

Theorem beq_ascii_refl : forall (c : ascii),
  beq_ascii c c = true.
Proof.
intros.
destruct c. simpl.
repeat rewrite eqb_reflx.
simpl. reflexivity.
Qed.

Hint Resolve beq_ascii_refl.

Ltac chop_apply_andb H :=
  apply andb_true_iff in H;
  inversion_clear H as (Htmp1, Htmp2);
  apply eqb_prop in Htmp2;
  rewrite Htmp2;
  clear Htmp2;
  rename Htmp1 into H.

Theorem beq_ascii_true : forall (c1 c2 : ascii),
  beq_ascii c1 c2 = true -> c1 = c2.
Proof.
destruct c1. destruct c2. intros Ha.
inversion Ha as [H].
do 7 (chop_apply_andb H).
apply eqb_prop in H; rewrite H.
reflexivity.
Qed.

Hint Resolve beq_ascii_true.

Theorem not_eq_beq_ascii_false : forall i1 i2,
  i1 <> i2 -> beq_ascii i1 i2 = false.
Proof.
  intros i1 i2 H.
  remember (beq_ascii i1 i2) as beq.
  destruct (beq). symmetry in Heqbeq.
  apply beq_ascii_true in Heqbeq.
  destruct H. assumption.
  reflexivity.
Qed.

Fixpoint beq_str (s1 s2 : string) {struct s1} : bool :=
  match s1,s2 with
  | EmptyString, EmptyString => true
  | String c1 s1', String c2 s2' =>
      if beq_ascii c1 c2 then beq_str s1' s2' else false
  | _, _ => false
  end.

Theorem beq_str_refl : forall (s : string),
  beq_str s s = true.
Proof with simpl in *; auto.
  intros.
  induction s as [| c s']...
  Case "s = String c s'".
    rewrite beq_ascii_refl.
    apply IHs'.
Qed.

Hint Resolve beq_str_refl.

Ltac introver := intro HHH; inversion HHH; clear HHH.

(* Boolean if-then-else Rewriting Rules *)

Lemma if_to_and : forall (a b : bool),
  (if a then b else false) = true -> (a = true /\ b = true).
Proof.
intros a b H.
induction a.
Case "a = true".
  split. reflexivity. apply H.
Case "a = false".
  inversion H.
Qed.

Theorem beq_str_true : forall (s s' : string),
  beq_str s s' = true -> s = s'.
Proof with simpl in *; auto.
  induction s as [| c sx].
  Case "s = EmptyString".
    destruct s'... introver.
  Case "s = String c sx".
    induction s' as [| c' sx']. introver.
    SCase "s' = String c' sx'".
      introver. apply if_to_and in H0. inversion_clear H0.
      apply beq_ascii_true in H; rewrite H.
      apply IHsx in H1; rewrite H1...
Qed.

Hint Resolve beq_str_true.


Theorem not_eq_beq_str_false : forall i1 i2,
  i1 <> i2 -> beq_str i1 i2 = false.
Proof.
  intros i1 i2 H.
  remember (beq_str i1 i2) as beq.
  destruct beq. symmetry in Heqbeq.
  apply beq_str_true in Heqbeq. destruct H. assumption.
  reflexivity.
Qed.


Theorem beq_str_sym : forall s1 s2,
  beq_str s1 s2 = beq_str s2 s1.
Proof.
  intros.
  remember (beq_str s1 s2) as r. symmetry in Heqr. destruct r.
  apply beq_str_true in Heqr. rewrite Heqr. symmetry.
  apply beq_str_refl.
  symmetry. apply not_eq_beq_str_false. intro.
  rewrite H in Heqr. assert (beq_str s1 s1 = true). apply beq_str_refl.
  rewrite Heqr in H0. inversion H0.
Qed.


(** ** Equality on scalar values (syntactic) *)

Definition beq_gen (g1 : General) (g2 : General) : bool := 
  match g1, g2 with
  | G_Integer i1, G_Integer i2 => 
    match Zcompare i1 i2 with
    | Eq => true
    | _ => false
    end
  | G_Text s1, G_Text s2 => beq_str s1 s2
  | G_Logical b1, G_Logical b2 => andb (implb b1 b2) (implb b2 b1)
  | G_Null, G_Null => true
  | _, _ => false
  end.
