(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)

Require Import Model.
Require Import Calculus.
Require Import LogicalSemantics.
Require Import BigStepSemantics.
Require Import Purity. (* only for syn_beq_res_true_iff *)



Inductive Step : Exp -> Exp -> Prop :=
  | step_un_op : forall op v v',
      un_op_eval op v v' ->
      Step (E_UnOp op (E_Value v)) (E_Value v')
  | step_bin_op : forall op v1 v2 v',
      bin_op_eval op v1 v2 v' ->
      Step (E_BinOp op (E_Value v1) (E_Value v2)) (E_Value v')
  | step_cond_true : forall e2 e3,
      Step (E_If (E_Value v_tt) e2 e3) e2
  | step_cond_false : forall e2 e3,
      Step (E_If (E_Value v_ff) e2 e3) e3
  | step_let_val : forall x v e,
      Step (E_Let x (E_Value v) e) (subst_exp e v x)
  | step_entity : forall les v, (* quite complicated/indirect definition -- like big-step semantics *)
      is_E v = true ->
      (forall li ei, In (li,ei) les -> exists vi,
        ei = (E_Value vi) /\ v_has_field li v = true /\ v_dot li v = vi) ->
      (forall li, v_has_field li v = true -> exists ei, In (li, ei) les) ->
      Step (E_Entity les) (E_Value v)
  | step_dot : forall l v,
      is_E v && v_has_field l v = true ->
      Step (E_Dot (E_Value v) l) (E_Value (v_dot l v))
  | step_add : forall v1 v2,
      is_C v2 = true ->
      Step (E_Add (E_Value v1) (E_Value v2)) (E_Value (v_add v1 v2))
  | step_appl : forall f arg ex v,
      (arg,ex) = functions f ->
      Step (E_App f (E_Value v)) (subst_exp ex v arg)
  | step_accum_lets : forall x y e2 e3 v1 vs,
      is_C v1 = true ->
      (Permutation vs (vb_to_vs (out_C v1))) ->
      Step (E_Acc x (E_Value v1) y e2 e3)
           (E_Let y e2 (let_seq y (List.map (fun v => subst_exp e3 v x) vs) (E_Var y)))
  | step_test_any : forall v,
      Step (E_In (E_Value v) T_Any) (E_Value v_tt)
  | step_test_integer : forall v,
      Step (E_In (E_Value v) T_Integer) (E_Value (v_logical (In_Integer v)))
  | step_test_text : forall v,
      Step (E_In (E_Value v) T_Text) (E_Value (v_logical (In_Text v)))
  | step_test_logical : forall v,
      Step (E_In (E_Value v) T_Logical) (E_Value (v_logical (In_Logical v)))
  | step_test_entity_1 : forall l T v,
      is_E v && v_has_field l v = true ->
      Step (E_In (E_Value v) (T_Entity l T)) (E_In (E_Value (v_dot l v)) T)
  | step_test_entity_2 : forall l T v,
      is_E v && v_has_field l v = false ->
      Step (E_In (E_Value v) (T_Entity l T)) (E_Value v_ff)
  | step_test_coll_1 : forall T v,
      is_C v = false ->
      Step (E_In (E_Value v) (T_Coll T)) (E_Value v_ff)
  | step_test_coll_3 : forall T v,
      (* XXX note that this could also be made non-deterministic
         (now I choose arbitray order, since here it doesn't matter anyway 
         -- it's always confluent since && is A+C) *)
      is_C v = true ->
      Step (E_In (E_Value v) (T_Coll T))
           (and_seq (List.map (fun v' => E_In (E_Value v') T) (vb_to_vs (out_C v))))
  | step_test_refine : forall x T v e,
      Step (E_In (E_Value v) (T_Refine x T e)) (E_And (E_In (E_Value v) T) (subst_exp e v x))
(* context rules *)
  | step_bin_op_ctxt_1 : forall op e1 e2 e1',
      Step e1 e1' ->
      Step (E_BinOp op e1 e2) (E_BinOp op e1' e2)
  | step_bin_op_ctxt_2 : forall op v1 e2 e2',
      Step e2 e2' ->
      Step (E_BinOp op (E_Value v1) e2) (E_BinOp op (E_Value v1) e2')
  | step_cond_ctxt : forall e1 e2 e3 e1',
      Step e1 e1' ->
      Step (E_If e1 e2 e3) (E_If e1' e2 e3)
  | step_test_ctxt : forall e T e',
      Step e e' ->
      Step (E_In e T) (E_In e' T)
  | step_let_ctxt : forall x e1 e2 e1',
      Step e1 e1' ->
      Step (E_Let x e1 e2) (E_Let x e1' e2)
  | step_entity_ctxt : forall les1 les2 l e e',
      (forall li ei, In (li,ei) les1 -> exists vi, ei = (E_Value vi)) ->
      Step e e' ->
      Step (E_Entity (les1 ++ (l, e) :: les2)) (E_Entity (les1 ++ (l, e') :: les2))
  |  step_dot_ctxt : forall l e e',
      Step e e' ->
      Step (E_Dot e l) (E_Dot e' l)
  | step_add_ctxt_1 : forall e1 e2 e1',
      Step e1 e1' ->
      Step (E_Add e1 e2) (E_Add e1' e2)
  | step_add_ctxt_2 : forall v1 e2 e2',
      Step e2 e2' ->
      Step (E_Add (E_Value v1) e2) (E_Add (E_Value v1) e2')
  | step_acc_ctxt : forall x e1 y e2 e3 e1',
      Step e1 e1' ->
      Step (E_Acc x e1 y e2 e3) (E_Acc x e1' y e2 e3)
  | step_app_ctxt : forall f e e',
      Step e e' ->
      Step (E_App f e) (E_App f e')
.


Tactic Notation "step_cases" tactic(first) tactic(c) :=
  first;
  [ c "step_un_op"
  | c "step_bin_op"
  | c "step_cond_true"
  | c "step_cond_false"
  | c "step_let_val"
  | c "step_entity"
  | c "step_dot"
  | c "step_add"
  | c "step_appl"
  | c "step_accum_lets"
  | c "step_test_any"
  | c "step_test_integer"
  | c "step_test_text"
  | c "step_test_logical"
  | c "step_test_entity_1"
  | c "step_test_entity_2"
  | c "step_test_coll_1"
  | c "step_test_coll_3"
  | c "step_test_refine"
  | c "step_bin_op_ctxt_1"
  | c "step_bin_op_ctxt_2"
  | c "step_cond_ctxt"
  | c "step_let_ctxt"
  | c "step_test_ctxt"
  | c "step_entity_ctxt"
  | c "step_dot_ctxt"
  | c "step_add_ctxt_1"
  | c "step_add_ctxt_2"
  | c "step_acc_ctxt"
  | c "step_app_ctxt"
].

Lemma syn_beq_res_false_iff : forall r1 r2,
  syn_beq_res r1 r2 = false <-> r1 <> r2.
Admitted.

Lemma not_logical_rev : forall v,
  v <> v_tt /\ v <> v_ff ->
  In_Logical v = false.
Admitted.

(* Direct attempt: doesn't work for case step_accum_lets!
See proof in Appendix B for a real proof *)
Lemma P6_failed : forall e e',
  Pure e ->
  Step e e' ->
  forall env, R e env = R e' env.
Proof.
  intros e e' HP H.
  (step_cases (induction H) Case); intros; simpl.
  Case "step_un_op". admit.
  Case "step_bin_op". admit.
  Case "step_cond_true". unfold LBind. reflexivity.
  Case "step_cond_false". unfold LBind. reflexivity.
  Case "step_let_val". unfold LBind. simpl. rewrite R_subst. reflexivity.
  Case "step_entity". admit. (* XXX complicated *)
  Case "step_dot". unfold LBind; simpl. rewrite H. reflexivity.
  Case "step_add". unfold LBind; simpl. rewrite H. reflexivity.
  Case "step_appl". admit. (* XXX functions not modeled properly in the logical semantics *)
  Case "step_accum_lets". unfold LBind; simpl.
    destruct (Proper (R e2 env)); simpl; try trivial.
    rewrite <- R_let_seq_res_acc; [|admit|trivial].
    admit. (* XXX broken XXX
              - the order of the lets is different (permutation)
              -- need to use purity assumption, but no direct proof seems possible *)
  Case "step_test_any". unfold LBind. reflexivity.
  Case "step_test_integer". unfold LBind. simpl. unfold v_logical. destruct (In_Integer v); reflexivity.
  Case "step_test_text". unfold LBind. simpl. unfold v_logical. destruct (In_Text v); reflexivity.
  Case "step_test_logical". unfold LBind. simpl. unfold v_logical. destruct (In_Logical v); reflexivity.
  Case "step_test_entity_1". unfold LBind. simpl. rewrite H. simpl. reflexivity.
  Case "step_test_entity_2". unfold LBind. simpl. rewrite H. simpl. reflexivity.
  Case "step_test_coll_1". unfold LBind. simpl. rewrite H. simpl. reflexivity.
  Case "step_test_coll_3". unfold LBind. simpl. rewrite H. simpl. admit. (* XXX complicated *)
  Case "step_test_refine". unfold LBind. simpl.
    destruct (W T v env); simpl. trivial.
    remember (syn_beq_res (R e (env_update env x v)) (Return v_tt)
           || syn_beq_res (R e (env_update env x v)) (Return v_ff)) as f; symmetry in Heqf; destruct f; simpl.
    SCase "1".
      rewrite orb_true_iff in Heqf. do 2 rewrite <- syn_beq_res_true_iff in Heqf. rewrite R_subst.
      destruct (F T v env); simpl; destruct Heqf; rewrite H; compute; reflexivity.
    SCase "2".
      destruct (F T v env); simpl;
        remember (Proper (R (subst_exp e v x) env)) as p; destruct p; symmetry in Heqp;
        simpl; try trivial;
        rewrite R_subst; rewrite not_logical_rev; try trivial; split; intro Hc;
        rewrite orb_false_iff in Heqf; do 2 rewrite syn_beq_res_false_iff in Heqf; destruct Heqf;
        rewrite R_subst in Heqp; destruct (R e (env_update env x v)); simpl in *; try inversion Heqp; subst;
        try (apply H; reflexivity); try (apply H0; reflexivity).
Admitted. (* Case step_accum_lets doesn't work with such a direct approach *)


(* Defining purity *)

Require Import Relations. (* reflexive transitive closure *)

Definition StepMany := clos_refl_trans _ Step.

(* accessibility predicate Acc is defined in Coq.Init.Wf as
Inductive Acc (A : Type) (R : A -> A -> Prop) (x : A) : Prop :=
    Acc_intro : (forall y : A, R y x -> Acc R y) -> Acc R x
*)
Definition terminating e := Acc Step e.

(* XXX Q: given a terminating e, can we produce a normal value to which e reduces?
This would allow us to formalize the logical semantics of functions in Coq
This is basically an interpreter for terminating expressions
*)
Require Import Recdef.
(* More natural definition 
- unfortunately we are required to prove that "Step" is well-founded without being given H
Function normal (e : Exp) (H : terminating e) {wf Step e} : Exp :=
  match e with (* still need proper a definition for the interpreter here *)
  | _ => E_Null
  end.
*)
Program Definition Step_term (e e' : {x : Exp | terminating x}) := Step e e'.
Function normal (e : {x : Exp | terminating x}) {wf Step_term e} : Exp :=
  match e with (* still need proper a definition for the interpreter here *)
  | _ => E_Null
  end.
(* still need to prove that Step_term is well-founded 
Proof.
  intro e. destruct e as [x H]. unfold Step_term, terminating in *.
  constructor. intros. inversion H. destruct y. simpl in *. apply H1 in H0.
*)
Admitted.
(* Last thing to prove would be that normal indeed produces a normal form of e *)


Definition unique_result e := exists! r, StepMany e r.

Axiom labeled_pure : string -> bool.

Definition calls_labeled_pure e :=
  forall f e', e = E_App f e' -> labeled_pure f = true.

Axiom subst_exp_many : Exp -> (string -> Value) -> Exp. 
(* XXX Thorsten had a definition of this, but didn't seem optimal
(was explicitly carrying "avoid" sets around to prevent capturing,
instead of shadowing things in sigma -- for this we would still need to use (string -> option Value) internally though *)

(* Coq doesn't like this recursive definition:
1. The term "forall e' : Exp, Sexp_exp e' e -> Pure e'" has type "Type" while it is expected to have type "Prop".
2. Recursive call to ClosedPure has principal argument equal to
"subst_exp_many e sigma" instead of a subterm of e.
Note: There might be ways to fix it though, e.g. proving that Sexp_exp is well-founded and using well-founded recursion
Fixpoint ClosedPure e := 
  fv_exp e = nil /\
  terminating e (* 1 *) /\
  unique_result e /\ (* 2 *)
  calls_labeled_pure e /\ (* 3 *)
  (forall e' : Exp, Sexp_exp e' e -> Pure e') (* 4 *)
with Pure e :=
  (forall sigma : string -> Value, ClosedPure (subst_exp_many e sigma)).
*)

(* This inductive definition seems to work *)
Inductive ClosedPure : Exp -> Prop :=
 | CP : forall e,
          fv_exp e = nil /\
          terminating e (* 1 *) /\
          unique_result e /\ (* 2 *)
          calls_labeled_pure e /\ (* 3 *)
          (forall e' : Exp, Sexp_exp e' e -> Pure e') (* 4 -- note that here we call Pure, since e' could be open *)
      -> ClosedPure e
with Pure : Exp -> Prop :=
 | P : forall e, (forall sigma : string -> Value, ClosedPure (subst_exp_many e sigma)) -> Pure e.

(* We assume a fixed + sound algorithm for termination checking of
  (possibly mutually recursive) procedures *)
Axiom termination_check : forall f,
  {forall v, terminating (E_App f v)} + {True}.

(* An attempt to redefine alg. purity in the same inductive style (currently recursively defined in Purity.v)
Note: Not sure if this brings us anything over the recursive definition
-- it definitely looks *less* algorithmic
-- this is the current definition from the paper!
-- would it maybe be easier for proofs? Seems it is easy to relate this to purity, see below. *)
Inductive AlgPure : Exp -> Prop :=
 | AP : forall e,
           calls_labeled_pure e -> (* 1 *)
           (forall x e1 y e2 e3, e = E_Acc x e1 y e2 e3 ->
             forall_bool (fun env =>
             forall_bool (fun v1 => forall_bool (fun v2 => forall_bool (fun v3 =>
               syn_beq_res
                 (LBind (R e3 (env_update (env_update env x v2) y v3)) (fun z =>
                        (R e3 (env_update (env_update env x v1) y z))))
                 (LBind (R e3 (env_update (env_update env x v1) y v3)) (fun z =>
                        (R e3 (env_update (env_update env x v2) y z))))
             )))) = true )  -> (* 2 *)
           (forall e' : Exp, Sexp_exp e' e -> AlgPure e') (* 3 *)
           -> AlgPure e.

Axiom labeled_pure_f_has_Alg_Pure_body : forall f arg body,
  labeled_pure f = true ->
  (arg,body) = functions f ->
  AlgPure body.

Axiom labeled_pure_f_termination_check_succeeds : forall f,
  labeled_pure f = true ->
  if (termination_check f) then True else False.

Lemma L9 : forall e sigma, AlgPure e -> AlgPure (subst_exp_many e sigma).
Admitted.

Lemma L10 : forall e, AlgPure e -> terminating e.
Admitted.

Lemma L11_plus_L13 : forall e, AlgPure e -> unique_result e.
Admitted.

Lemma L_calls_labeled_pure : forall e,
  AlgPure e ->
  calls_labeled_pure e.
Proof. intros. inversion H. assumption. Qed.

(* a closing substitution really produces closed expressions *) 
Lemma subst_exp_many_closes : forall e sigma, fv_exp (subst_exp_many e sigma) = nil.
Admitted.

Lemma Sexp_exp_subst_exp : forall e e' sigma,
  Sexp_exp e' (subst_exp_many e sigma) ->
  exists e'', e' = subst_exp_many e'' sigma /\ Sexp_exp e'' e.
Admitted.

Lemma AlgPure_Sexp : forall e,
  AlgPure e ->
  (forall e' : Exp, Sexp_exp e' e -> AlgPure e').
Proof. intros. inversion H. subst. apply H3. assumption. Qed.

Lemma subst_exp_many_closed : forall e sigma,
  fv_exp e = nil ->
  subst_exp_many e sigma = e.
Admitted.

Lemma closed_exp_ClosedPure_equiv_Pure : forall e,
  fv_exp e = nil ->
  (ClosedPure e <-> Pure e).
Proof.
  split; intros.
    constructor. intros sigma. rewrite subst_exp_many_closed; assumption.
    inversion H0; subst. specialize (H1 env_empty). rewrite subst_exp_many_closed in H1. apply H1. assumption.
Qed.

Lemma Pure_subst : forall e sigma,
  Pure e ->
  Pure (subst_exp_many e sigma).
Proof.
  intros. rewrite <- closed_exp_ClosedPure_equiv_Pure.
  inversion H; subst. apply H0. apply subst_exp_many_closes.
Qed.

(* Algorithmic purity implies purity: as simple as this can get *)
Theorem AlgPure_implies_Pure : forall e, AlgPure e -> Pure e
with AlgPure_implies_Pure_for_Types : forall T e, Sexp_type e T -> AlgPure e -> Pure e.
Proof.
Case "AlgPure_implies_Pure". clear AlgPure_implies_Pure.
  (Exp_cases (induction e) SCase); intros; apply P; intros; apply CP;
    (split; [apply subst_exp_many_closes |]);
    (split; [apply L10; apply L9; assumption |]);
    (split; [apply L11_plus_L13; apply L9; assumption |]);
    (split; [apply L_calls_labeled_pure; apply L9; assumption |]);
    intros e' H'; apply Sexp_exp_subst_exp in H'; destruct H' as [e'' [H1 H2]]; subst.
  (* all things above -- up to part (4) of the definition of Pure -- don't require induction *)
  SCase "E_Var". inversion H2. (* variables don't have subexpressions *)
  SCase "E_Value". inversion H2. (* values don't have subexpressions *)
  SCase "E_UnOp".
    inversion H2; subst. eapply AlgPure_Sexp in H; try eassumption. apply Pure_subst. apply IHe. apply H.
  SCase "E_BinOp". (* like E_UnOp *)
    inversion H2; subst.
    SSCase "SE_BinOp1".
      eapply AlgPure_Sexp in H; try eassumption. apply Pure_subst. apply IHe1. apply H.
    SSCase "SE_BinOp2".
      eapply AlgPure_Sexp in H; try eassumption. apply Pure_subst. apply IHe2. apply H.
  SCase "E_If". (* like E_UnOp *)
    inversion H2; subst; eapply AlgPure_Sexp in H; try eassumption; apply Pure_subst; eauto.
  SCase "E_Let". inversion H2; subst.
    SSCase "SE_Let_1".
      eapply AlgPure_Sexp in H; try eassumption. apply Pure_subst. apply IHe1. apply H.
    SSCase "SE_Let_2".
      (* interesting case (on paper) since e2 is not closed by sigma if sigma is finite
      and only binds the free variables of the original let expression.
      However, here (Coq formalization) we consider total substitutions!
      So it all works flawlessly, exactly like in the previous cases.
      *)
    eapply AlgPure_Sexp in H; try eassumption. apply Pure_subst. apply IHe2. apply H.
  SCase "E_In". inversion H2; subst.
    SSCase "SE_In_1".
      eapply AlgPure_Sexp in H; try eassumption. apply Pure_subst. apply IHe. apply H.
    SSCase "SE_In_2". rename m into T.
      (* here is where the fun starts! we need to do simultaneous induction on T *)
      eapply AlgPure_Sexp in H; try eassumption.
      apply Pure_subst. eapply AlgPure_implies_Pure_for_Types. eassumption. apply H.
  SCase "E_Entity". (* like E_UnOp, just n times *)
    admit. (* Exp_ind is too weak in this case, this can be fixed with some work *)
  SCase "E_Dot". (* like E_UnOp *)
    inversion H2; subst. eapply AlgPure_Sexp in H; try eassumption. apply Pure_subst. apply IHe. apply H.
  SCase "E_Add". (* like E_UnOp *)
    inversion H2; subst; eapply AlgPure_Sexp in H; try eassumption; apply Pure_subst; eauto.
  SCase "E_Acc". (* like E_Let *)
    inversion H2; subst; eapply AlgPure_Sexp in H; try eassumption; apply Pure_subst; eauto.
  SCase "E_App". (* like E_UnOp *)
    inversion H2; subst; eapply AlgPure_Sexp in H; try eassumption; apply Pure_subst; eauto.
Case "AlgPure_implies_Pure_for_Types". clear AlgPure_implies_Pure_for_Types.
  (* this first part is exactly like above, no induction used *)
  (MType_cases (induction T) SCase); intros; apply P; intros; apply CP;
    (split; [apply subst_exp_many_closes |]);
    (split; [apply L10; apply L9; assumption |]);
    (split; [apply L11_plus_L13; apply L9; assumption |]);
    (split; [apply L_calls_labeled_pure; apply L9; assumption |]); intros.
  SCase "T_Any". inversion H. (* Any doesn't have subexpressions *)
  SCase "T_Integer". inversion H. (* Integer doesn't have subexpressions *)
  SCase "T_Text". inversion H. (* Text doesn't have subexpressions *)
  SCase "T_Logical". inversion H. (* Logical doesn't have subexpressions *)
  SCase "T_Coll".
    inversion H; subst. specialize (IHT e H4 H0).
    inversion IHT; subst. specialize (H2 sigma). inversion H2.
    destruct H3 as [H30 [H31 [H32 [H33 H34]]]]. apply H34; assumption. 
  SCase "T_Entity". 
    inversion H; subst. specialize (IHT e H4 H0).
    inversion IHT; subst. specialize (H2 sigma). inversion H2.
    destruct H3 as [H30 [H31 [H32 [H33 H34]]]]. apply H34; assumption. 
  SCase "T_Refine".
    inversion H; subst.
    SSCase "ST_Refine_1".
      specialize (IHT e0 H4 H0).
      inversion IHT; subst. specialize (H2 sigma). inversion H2.
      destruct H3 as [H30 [H31 [H32 [H33 H34]]]]. apply H34; assumption. 
    SSCase "ST_Refine_2". (* the only really interesting case in this lemma *)
      specialize (AlgPure_implies_Pure e H0).
      inversion AlgPure_implies_Pure; subst. specialize (H2 sigma).
      inversion H2; subst. destruct H3 as [H30 [H31 [H32 [H33 H34]]]]. apply H34. assumption.
Admitted. (* can't Qed because there is still one admit in there, E_Entity case *)


(* Bad simplification idea (before somebody else suggests it)
-- this would really suprise me if it worked *)

Lemma AlgPure_implies_Pure_Sexp : forall e e', AlgPure e -> Sexp_exp e' e -> Pure e'
with AlgPure_implies_Pure_Sexp_for_Types : forall T e, Sexp_type e T -> AlgPure e -> Pure e.
Admitted. (* <- I highly suspect this will break in this proof -- induction hypothesis too weak *)

Theorem AlgPure_implies_Pure_simple : forall e, AlgPure e -> Pure e.
Proof.
  intros; apply P; intros; apply CP;
    (split; [apply subst_exp_many_closes |]);
    (split; [apply L10; apply L9; assumption |]);
    (split; [apply L11_plus_L13; apply L9; assumption |]);
    (split; [apply L_calls_labeled_pure; apply L9; assumption |]);
    intros e' H'; apply Sexp_exp_subst_exp in H'; destruct H' as [e'' [H1 H2]]; subst.
  apply Pure_subst. eapply AlgPure_implies_Pure_Sexp; eassumption.
Qed.



(* Andy's rule-based definition of alg. purity *)

Inductive TA_Pure : MType -> Prop :=
  | TAP_Any : TA_Pure T_Any
  | TAP_Integer : TA_Pure T_Integer
  | TAP_Text : TA_Pure T_Text
  | TAP_Logical : TA_Pure T_Logical
  | TAP_Coll : forall T, TA_Pure T -> TA_Pure (T_Coll T)
  | TAP_Entity : forall l T, TA_Pure T -> TA_Pure (T_Entity l T)
  | TAP_Refine : forall x T e, TA_Pure T -> A_Pure e -> TA_Pure (T_Refine x T e)
with A_Pure : Exp -> Prop :=
  | AP_Var : forall x, A_Pure (E_Var x)
  | AP_Value : forall v, A_Pure (E_Value v)
  | AP_UnOp : forall op e, A_Pure e -> A_Pure (E_UnOp op e)
  | AP_BinOp : forall op e1 e2, A_Pure e1 -> A_Pure e2 -> A_Pure (E_BinOp op e1 e2)
  | AP_If : forall e1 e2 e3, A_Pure e1 -> A_Pure e2 -> A_Pure e3 -> A_Pure (E_If e1 e2 e3)
  | AP_Let : forall x e1 e2, A_Pure e1 -> A_Pure e2 -> A_Pure (E_Let x e1 e2)
  | AP_In : forall e T, A_Pure e -> TA_Pure T -> A_Pure (E_In e T)
  | AP_Entity : forall les, (forall l e, In (l,e) les -> A_Pure e) -> A_Pure (E_Entity les)
  | AP_Dot : forall e l, A_Pure e -> A_Pure (E_Dot e l)
  | AP_Add : forall e1 e2, A_Pure e1 -> A_Pure e2 -> A_Pure (E_Add e1 e2)
  | AP_Acc : forall x e1 y e2 e3, A_Pure e1 -> A_Pure e2 -> A_Pure e3 ->
               (forall_bool (fun env =>
               forall_bool (fun v1 => forall_bool (fun v2 => forall_bool (fun v3 =>
                 syn_beq_res
                   (LBind (R e3 (env_update (env_update env x v2) y v3)) (fun z =>
                          (R e3 (env_update (env_update env x v1) y z))))
                   (LBind (R e3 (env_update (env_update env x v1) y v3)) (fun z =>
                          (R e3 (env_update (env_update env x v2) y z))))
               )))) = true ) ->
                 A_Pure (E_Acc x e1 y e2 e3)
  | AP_App : forall f e, A_Pure e -> labeled_pure f = true ->
                 A_Pure (E_App f e).

Tactic Notation "A_Pure_cases" tactic(first) tactic(c) :=
  first;
  [ c "AP_Var"
  | c "AP_Value"
  | c "AP_UnOp"
  | c "AP_BinOp"
  | c "AP_If"
  | c "AP_Let"
  | c "AP_In"
  | c "AP_Entity"
  | c "AP_Dot"
  | c "AP_Add"
  | c "AP_Acc"
  | c "AP_App"
  ].

Tactic Notation "TA_Pure_cases" tactic(first) tactic(c) :=
  first;
  [ c "TAP_Any"
  | c "TAP_Integer"
  | c "TAP_Text"
  | c "TAP_Logical"
  | c "TAP_Coll"
  | c "TAP_Entity"
  | c "TAP_Refine"
  ].


Axiom labeled_pure_f_has_A_Pure_body : forall f arg body,
  labeled_pure f = true ->
  (arg,body) = functions f ->
  A_Pure body.


Lemma AL9 : forall e sigma, A_Pure e -> A_Pure (subst_exp_many e sigma).
Admitted.

Lemma AL10 : forall e, A_Pure e -> terminating e.
Admitted.

Lemma AL11_plus_L13 : forall e, A_Pure e -> unique_result e.
Admitted.

(* this needs induction while before it was trivial *)
Lemma AL_calls_labeled_pure : forall e,
  A_Pure e ->
  calls_labeled_pure e.
Admitted.

(* this needs induction while before it was trivial *)
Lemma A_Pure_Sexp : forall e,
  A_Pure e ->
  (forall e' : Exp, Sexp_exp e' e -> A_Pure e').
Admitted.

Theorem APure_implies_Pure : forall e, A_Pure e -> Pure e
with TAPure_implies_Pure_for_Types : forall T e, TA_Pure T -> Sexp_type e T -> Pure e.
Proof.
Case "APure_implies_Pure". clear APure_implies_Pure.
  intros.
(* XXX Doing these things before the induction is bad (breaks the induction hypothesis)
-- however it's no longer so easy to do them uniformaly on all cases after induction
(as it is done in AlgPure_implies_Pure) *)
  apply P; intro sigma; apply CP;
    (split; [apply subst_exp_many_closes |]);
    (split; [apply AL10; apply AL9; assumption |]);
    (split; [apply AL11_plus_L13; apply AL9; assumption |]);
    (split; [apply AL_calls_labeled_pure; apply AL9; assumption |]).
  generalize dependent sigma. (A_Pure_cases (induction H) SCase); intros sigma e' H';
    apply Sexp_exp_subst_exp in H'; destruct H' as [e'' [G1 G2]]; subst.
  SCase "AP_Var". inversion G2. (* variables don't have subexpressions *)
  SCase "AP_Value". inversion G2. (* values don't have subexpressions *)
  SCase "AP_UnOp".
    inversion G2; subst. apply Pure_subst. eapply IHA_Pure. (* XXX stuck *)
Admitted. (* It's not trivial to adapt the previous proof to work with Andy's suggested definition *)



(* Trying  to prove that alg. purity is preserved by small-step reduction *)

Lemma AlgPure_value : forall v,
  AlgPure (E_Value v).
Proof.
  intro v. constructor.
  unfold calls_labeled_pure. intros. inversion H.
  intros. inversion H.
  intros. inversion H. (* values don't have subexpressions *)
Qed.

(* variant of Lemma 9 *)
Lemma AlgPure_subst : forall x v e,
  AlgPure e ->
  AlgPure (subst_exp e v x).
Admitted.

Lemma Sexp_exp_and_seq : forall vs e T,
  Sexp_exp e (and_seq (map (fun v' : Value => E_In (E_Value v') T) vs)) ->
  (e = E_True) \/
  (exists v1, exists vs1, vs = v1 :: vs1 /\
     (e = (E_In (E_Value v1) T) \/
     (e = (and_seq (map (fun v' : Value => E_In (E_Value v') T) vs1))))).
Proof.
  destruct vs as [| v1 vs1]; intros; simpl in *. inversion H.
  repeat eexists.
    destruct vs1 as [| v2 vs2]. simpl in *. inversion H; subst. left. reflexivity.
    right. repeat eexists. left. reflexivity.
    simpl in *.
    inversion H; subst. right. repeat eexists. right. reflexivity.
    inversion H; subst. right. repeat eexists. left. reflexivity.
Qed.

Lemma reduction_preserves_Alg_Pure : forall e e',
  Step e e' ->
  AlgPure e ->
  AlgPure e'.
Proof.
  intros e e' H. (step_cases (induction H) Case); intro H'; try (apply AlgPure_value); inversion H'; subst.
  Case "step_cond_true". apply H1. constructor.
  Case "step_cond_false". eapply H1. constructor.
  Case "step_let_val". apply AlgPure_subst. eapply H1. constructor.
  Case "step_appl".
    apply AlgPure_subst.
    unfold calls_labeled_pure in H0.
    assert(labeled_pure f = true). eapply H0. reflexivity.
    eapply labeled_pure_f_has_Alg_Pure_body; eassumption.
  Case "step_accum_lets". admit. (* XXX complicated? *)
  Case "step_test_entity_1".
    constructor.
    unfold calls_labeled_pure. intros. inversion H3.
    intros. inversion H3.
    intros. inversion H3; subst. apply AlgPure_value. apply H2. constructor. constructor. assumption.
  Case "step_test_coll_3".
    constructor.
      unfold calls_labeled_pure. intros.
        destruct (vb_to_vs (out_C v)); unfold and_seq in H3; simpl in H3; inversion H3.
      intros. destruct (vb_to_vs (out_C v)); unfold and_seq in H3; simpl in H3; inversion H3.
      intros. apply Sexp_exp_and_seq in H3. destruct H3.
        subst. apply AlgPure_value.
        destruct H3 as [v1 [vs1 [H31 [H32 | H32]]]].
          subst. admit. (* XXX the easy (base) case AlgPure (E_In (E_Value v1) T) *)
          subst. admit. (* XXX the hard case -- probably need some nested induction on vb_to_vs (out_C v) *)
  Case "step_test_refine".
    constructor.
    unfold calls_labeled_pure. intros. inversion H2.
    intros. inversion H2.
    intros. inversion H2; subst.
      constructor.
        unfold calls_labeled_pure. intros. inversion H3.
        intros. inversion H3.
        intros. inversion H3; subst.
        apply AlgPure_value.
        inversion H3; subst. apply AlgPure_value. apply H1.
          constructor. constructor. assumption.
      apply AlgPure_subst. apply H1. constructor. apply ST_Refine_2.
Admitted. (* XXX still has a couple of admits
XXX also the context rules not treated yet *)


(* Trying same proof with Andy's definition -- it's simpler! *)


Lemma A_Pure_value : forall v,
  A_Pure (E_Value v).
Proof. intros. constructor. Qed.

(* variant of Lemma 9 *)
Lemma A_Pure_subst : forall x v e,
  A_Pure e ->
  A_Pure (subst_exp e v x).
Admitted.


Lemma reduction_preserves_A_Pure : forall e e',
  Step e e' ->
  A_Pure e ->
  A_Pure e'.
Proof.
  intros e e' H. (step_cases (induction H) Case); intro H';
    try (apply A_Pure_value); inversion H'; subst; try (constructor; eauto; fail).
  Case "step_cond_true". apply H3.
  Case "step_cond_false". apply H4.
  Case "step_let_val". apply A_Pure_subst. eapply H3.
  Case "step_appl".
    apply A_Pure_subst. eapply labeled_pure_f_has_A_Pure_body; eassumption.
  Case "step_accum_lets".
    constructor. assumption. clear H0.
    induction vs as [| v' vs']; simpl. constructor.
    constructor. apply A_Pure_subst. assumption. apply IHvs'.
  Case "step_test_entity_1".
    inversion H3; subst.
    constructor; try assumption. constructor.
  Case "step_test_coll_3".
    inversion H3; subst. clear H' H2 H3.
    induction (vb_to_vs (out_C v)) as [| v0 vs]; simpl. apply A_Pure_value.
    constructor. apply IHvs. constructor. apply A_Pure_value. assumption.
  Case "step_test_refine".
    inversion H2; subst.
    constructor. constructor. apply A_Pure_value. assumption.
    apply A_Pure_subst. assumption.
  Case "step_entity_ctxt".
    constructor. intros. apply in_app_or in H1. destruct H1.
    eapply H2. apply in_or_app. left. eassumption.
    simpl in H1. destruct H1.
    inversion H1; subst. apply IHStep. eapply H2. apply in_or_app. right. simpl. left. reflexivity.
    eapply H2. apply in_or_app. right. simpl. right. eassumption.
Qed.


(* Trying to relate Andy's definition with the current one *)

Lemma AlgPure_implies_A_Pure : forall e, AlgPure e -> A_Pure e
with AlgPure_implies_TA_Pure : forall T, (forall e, Sexp_type e T -> AlgPure e) -> TA_Pure T.
Proof.
Case "AlgPure_implies_A_Pure". clear AlgPure_implies_A_Pure.
  intros e H. (Exp_cases (induction e) SCase); inversion H; subst; constructor.
  SCase "E_UnOp". apply IHe. apply H2. constructor.
  SCase "E_BinOp".
    apply IHe1. apply H2. constructor.
    apply IHe2. apply H2. constructor.
  SCase "E_If".
    apply IHe1. apply H2. constructor.
    apply IHe2. apply H2. constructor.
    apply IHe3. apply H2. constructor.
  SCase "E_Let".
    apply IHe1. apply H2. constructor.
    apply IHe2. apply H2. constructor.
  SCase "E_In".
    apply IHe. apply H2. constructor.
    eapply AlgPure_implies_TA_Pure. intros e' H'. apply H2. constructor. apply H'.
  SCase "E_Entity". admit. (* XXX induction principle *)
  SCase "E_Dot".
    apply IHe. apply H2. constructor.
  SCase "E_Add".
    apply IHe1. apply H2. constructor.
    apply IHe2. apply H2. constructor.
  SCase "E_Acc".
    apply IHe1. apply H2. constructor.
    apply IHe2. apply H2. constructor.
    apply IHe3. apply H2. constructor.
    eapply H1. reflexivity.
  SCase "E_App".
    apply IHe. apply H2. constructor.
    eapply H0. reflexivity.
Case "AlgPure_implies_TA_Pure". clear AlgPure_implies_TA_Pure.
  intros T H. (MType_cases (induction T) SCase); constructor.
  SCase "T_Coll".
    apply IHT. intros e' H'. apply H. constructor. apply H'.
  SCase "T_Entity".
    apply IHT. intros e' H'. apply H. constructor. apply H'.
  SCase "T_Refine".
    apply IHT. intros e' H'. apply H. constructor. apply H'.
    apply AlgPure_implies_A_Pure. apply H. apply ST_Refine_2.
Admitted. (* there is still this stupid E_Entity case *)

Lemma A_Pure_implies_AlgPure : forall e, A_Pure e -> AlgPure e
with TAPure_implies_AlgPure_for_Types : forall T e, TA_Pure T -> Sexp_type e T -> AlgPure e.
Proof.
Case "A_Pure_implies_AlgPure". clear A_Pure_implies_AlgPure.
  intros e H. (A_Pure_cases (induction H) SCase);
  (apply AP; [ unfold calls_labeled_pure; intros; try (inversion H; fail); try (inversion H0; fail); try (inversion H1; fail); try (inversion H2; fail); try (inversion H3; fail)
            | intros; try (inversion H; fail); try (inversion H0; fail); try (inversion H1; fail); try (inversion H2; fail); try (inversion H3; fail)
            | intros; try (inversion H; fail); try (inversion H0; fail); try (inversion H1; fail); try (inversion H2; fail); try (inversion H3; fail) ]).
  SCase "AP_UnOp".
    inversion H0; subst; assumption.
  SCase "AP_BinOp".
    inversion H1; subst; assumption.
  SCase "AP_If".
    inversion H2; subst; assumption.
  SCase "AP_Let".
    inversion H1; subst; assumption.
  SCase "AP_In".
    inversion H1; subst. assumption.
    eapply TAPure_implies_AlgPure_for_Types; eassumption.
  SCase "AP_Entity".
    inversion H1; subst. eapply H0; eassumption.
  SCase "AP_Dot".
    inversion H0; subst. assumption.
  SCase "AP_Add".
    inversion H1; subst; assumption.
  SCase "AP_Acc".
    inversion H3; subst. apply H2.
    inversion H3; subst; assumption.
  SCase "AP_App".
    inversion H1; subst. apply H0.
    inversion H1; subst; assumption.
Case "TAPure_implies_AlgPure_for_Types". clear TAPure_implies_AlgPure_for_Types.
  intros T e He1 He2. (TA_Pure_cases (induction He1) SCase); inversion He2; subst; auto.
Qed.


(* Giving progress a shot *)

Require Import DeclarativeTypeSystem.

Definition is_value_dec e : {exists v, e = E_Value v} + {~exists v, e = E_Value v}.
  intro. destruct e; try (right; intro Hc; destruct Hc; inversion H; fail).
  left. eexists. reflexivity.
Qed.

(*
Definition is_value e :=
  match e with
  | E_Value v => true
  | _ => false
  end.

Lemma Is_value : forall e,
  is_value e = true <-> (exists v, e = E_Value v).
Proof.
  intros. induction e; split; simpl; intros;
    try (destruct H; inversion H; fail);
    try (inversion H; fail); auto.
  eexists. reflexivity.
Qed.
*)


Theorem logical_soundness_1a : forall E e T,
  E |- e ::: T ->
  forall sigma, FE E sigma -> Proper(R e sigma) = true.
Admitted.

Theorem logical_soundness_1b : forall E e T,
  E |- e ::: T ->
  forall sigma, FE E sigma -> F T (out_V (R e sigma)) sigma = true.
Admitted.

Lemma canonical_form_logical : forall v,
  nil |- E_Value v ::: T_Logical ->
  In_Logical v = true.
Proof.
  intros. apply logical_soundness_1b with (sigma := env_empty) in H. simpl in H. apply H.
    unfold FE. intros. inversion H0.
Qed.

Lemma canonical_form_integer : forall v,
  nil |- E_Value v ::: T_Integer ->
  In_Integer v = true.
Proof.
  intros. apply logical_soundness_1b with (sigma := env_empty) in H. simpl in H. apply H.
    unfold FE. intros. inversion H0.
Qed.

Lemma canonical_form_text : forall v,
  nil |- E_Value v ::: T_Text ->
  In_Text v = true.
Proof.
  intros. apply logical_soundness_1b with (sigma := env_empty) in H. simpl in H. apply H.
    unfold FE. intros. inversion H0.
Qed.

Lemma canonical_form_entity : forall v l T,
  nil |- E_Value v ::: T_Entity l T ->
  is_E v && v_has_field l v = true.
Proof.
  intros. apply logical_soundness_1b with (sigma:=env_empty) in H. simpl in H.
    rewrite andb_true_iff in H. destruct H. apply H.
    unfold FE. intros. inversion H0.
Qed.

(* Implied Judgments Lemma from the paper *)
Lemma implied_judgements_3_part : forall E e T,
  E |- e ::: T ->
  wf_type E T.
Admitted.

Require Import Axioms. (* only for axiom_eq *)

Lemma canonical_form_entity_additional : forall v l T,
  nil |- E_Value v ::: T_Entity l T ->
  (nil |- E_Value (v_dot l v) ::: T).
Proof.
  intros.
  destruct (fresh nil) as [x Hx].
  eapply ExpSub with (x := x). apply ExpConst. constructor.
  constructor. unfold T_Sing_x. simpl. constructor. reflexivity.
    constructor; simpl. repeat constructor. assumption.
    constructor. repeat constructor. assumption.
      apply implied_judgements_3_part in H. inversion H. assumption.
  intros. simpl in *.
  rewrite env_update_same in H1.
  assert(v0 = (v_dot l v)). apply syn_beq_val_equals in H1. rewrite axiom_eq in H1.
    remember (syn_beq_val v0 (v_dot l v)) as b. destruct b.
      symmetry in Heqb. apply syn_beq_val_equals in Heqb. subst. reflexivity. inversion H1.
  subst.
  apply logical_soundness_1b with (sigma:=sigma) in H. simpl in H.
  do 2 (rewrite andb_true_iff in H; destruct H).
  assumption.
  assumption.
  assumption.
Qed.

Lemma canonical_form_coll : forall v T,
  nil |- E_Value v ::: T_Coll T ->
  is_C v = true.
Proof. (* same proof for entities (copy paste) *)
  intros. apply logical_soundness_1b with (sigma:=env_empty) in H. simpl in H.
    rewrite andb_true_iff in H. destruct H. apply H.
    unfold FE. intros. inversion H0.
Qed.


Lemma canonical_form_coll_additional : forall v T,
  nil |- E_Value v ::: T_Coll T ->
  (forall v0, v_mem v0 v = true -> nil |- E_Value v0 ::: T).
Proof. 
  intros.
  destruct (fresh nil) as [x Hx].
  eapply ExpSub with (x := x). apply ExpConst. constructor.
  constructor. unfold T_Sing_x. simpl. constructor. reflexivity.
    constructor; simpl. repeat constructor. assumption.
    constructor. repeat constructor. assumption.
      apply implied_judgements_3_part in H. inversion H. assumption.
  intros. simpl in *.
  rewrite env_update_same in H2.
  assert(v0 = v1). apply syn_beq_val_equals in H2. rewrite axiom_eq in H2.
    remember (syn_beq_val v1 v0) as b. destruct b.
      symmetry in Heqb. apply syn_beq_val_equals in Heqb. subst. reflexivity. inversion H2.
  subst.
  apply logical_soundness_1b with (sigma:=sigma) in H. simpl in H.
  rewrite andb_true_iff in H; destruct H.
  rewrite <- forall_bool_true_iff in H3. specialize (H3 v1).
  rewrite implb_true_iff in H3. apply H3. assumption.
  assumption.
  assumption.
Qed.


Lemma test_val_progress : forall v T,
  exists e', Step (E_In (E_Value v) T) e'.
Proof.
  intros; destruct T; try (eexists; constructor; fail).
  Case "T = T_Coll T".
    remember (is_C v) as b. symmetry in Heqb. destruct b.
      eexists; apply step_test_coll_3; eassumption.
      eexists; apply step_test_coll_1; eassumption.
  Case "T = T_Entity s T".
    remember (is_E v && v_has_field s v) as b. symmetry in Heqb. destruct b.
      eexists; apply step_test_entity_1; assumption.
      eexists; apply step_test_entity_2; assumption.
Qed.

Theorem progress : forall e T,
  nil |- e ::: T ->
  (~exists v, e = E_Value v) ->
  exists e', Step e e'.
Proof.
  intros e T H Hv. remember (nil:EnvT) as E.
  (has_type_cases (induction H) Case); intros; subst.
  Case "ExpSub". apply IHhas_type; trivial.
  Case "ExpConst". apply False_ind. apply Hv. eexists. reflexivity.
  Case "ExpOperatorUn". admit. (* XXX *)
  Case "ExpOperatorBi". admit. (* XXX *)
  Case "ExpCond".
    destruct (is_value_dec e1) as [He1 | He1].
    SCase "e1 is value".
      destruct He1 as [v He1]. subst. apply canonical_form_logical in H0.
      apply logical_only_two in H0; destruct H0; subst; eexists; constructor.
    SCase "e1 not a value".
      edestruct IHhas_type1 as [e' IH1]; trivial.
      eexists. apply step_cond_ctxt. apply IH1.
  Case "ExpVar". inversion H0.
  Case "ExpLet".
    destruct (is_value_dec e1) as [He1 | He1].
    SCase "e1 is value".
      destruct He1 as [v He1]. subst.
      eexists; constructor.
    SCase "e1 not a value".
      edestruct IHhas_type1 as [e' IH1]; trivial.
      eexists. apply step_let_ctxt. apply IH1.
  Case "ExpTest".
    destruct (is_value_dec e) as [He | He].
    SCase "e is value".
      destruct He as [v He]. subst. apply test_val_progress.
    SCase "e not a value".
      edestruct IHhas_type as [e' IH1]; trivial.
      eexists. apply step_test_ctxt. apply IH1.
  Case "ExpEntity".
    induction les as [| [l e] les'].
    SCase "les=nil". exists (E_Value v_eempty). constructor.
      compute. reflexivity.
      intros. inversion H2.
      intros. compute in H2. inversion H2.
    SCase "les = (l,e) :: les'". simpl in *.
      destruct (is_value_dec e) as [He | He]; admit. (* XXX *)
(* previous proof attempt -- formalization gets in the way of a simple proof
      SSCase "e is value".
        destruct He as [v He]. subst.
        admit. (* XXX *)
      SSCase "e not a value".
        unfold dom in H. rewrite fst_split_cons in H.
        assert(In l (fst (split lts))) as Hl. inversion H. simpl. left. reflexivity.
        apply in_fst_split in Hl. destruct Hl as [T Hl].
        edestruct H1 as [e' IH1]; trivial. left. reflexivity. apply Hl. assumption.
        eexists. assert((l, e) :: les' = nil ++ (l, e) :: les') as HH. reflexivity. rewrite HH.
        apply step_entity_ctxt. intros. inversion H2. apply IH1.
*)
  Case "ExpDot".
    destruct (is_value_dec e) as [He | He].
    SCase "e is value".
      destruct He as [v He]. subst.
      apply canonical_form_entity in H.
      eexists; constructor; assumption.
    SCase "e not a value".
      edestruct IHhas_type as [e' IH1]; trivial.
      eexists. apply step_dot_ctxt. apply IH1.
  Case "ExpAdd".
    destruct (is_value_dec e1) as [He1 | He1].
    SCase "e1 is value".
      destruct He1 as [v1 He1]. subst.
      destruct (is_value_dec e2) as [He2 | He2].
      SSCase "e2 is value".
        destruct He2 as [v2 He2]. subst.
        apply canonical_form_coll in H0.
        eexists; constructor; assumption.
      SSCase "e2 not a value".
        edestruct IHhas_type2 as [e2' IH2]; trivial.
        eexists. apply step_add_ctxt_2. apply IH2.
    SCase "e1 not a value".
      edestruct IHhas_type1 as [e1' IH1]; trivial.
      eexists. apply step_add_ctxt_1. apply IH1.
  Case "ExpAcc".
    destruct (is_value_dec e1) as [He1 | He1].
    SCase "e1 is value".
      destruct He1 as [v1 He1]. subst.
      apply canonical_form_coll in H.
      eexists; constructor. assumption. apply Permutation_refl.
    SCase "e1 not a value".
      edestruct IHhas_type1 as [e1' IH1]; trivial.
      eexists. apply step_acc_ctxt. apply IH1.
Qed. (* XXX ExpEntity case messy, ExpOperatorUn and ExpOperatorBi boring, otherwise all done *)

