(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)


Require Export Basic.
Require Export Values.
Require Export InsertionSorting.
Require Import TheoryList.
Require Import Coq.Bool.DecBool.


(** ** Enforcing distinctness of entity keys *)

(** We could (try to) enforce distinctness of entity keys using refinements (set comprehensions).
Refinements (set comprehensions) are defined in the  
#<a href="http://coq.inria.fr/stdlib/Coq.Init.Specif.html">Specif</a># module,
which seems to be loaded by default. 
[[
Inductive Value' : Type :=
  | G' : General -> Value'
  | E' : {svm : list (string * Value') | NoDup (fst (split svm)) } -> Value'
  | C' : list Value' -> Value'
]] *)
(** Problem:
[[
Error: Non strictly positive occurrence of "_UNBOUND_REL_3" in
 "{svm : list (string * Value') | NoDup (fst (split svm))} -> Value'"
]] *)

(** Note that refinements are not magic in any way, [{x : T | P x}] is just a (dependently typed)
pair of a term x of type T and of a proof that x satisfies P.
In particular this means that in order to obtain the element of type T out of the pair
of type [{x : T | P x}] one needs to explicitly 
#<a href="http://coq.inria.fr/stdlib/Coq.Init.Specif.html&#35;Subset_projections">project it out</a>#
using [proj1_sig]. So we could define [Value] like this
[[
Inductive Value'' : Type :=
  | G'' : General -> Value''
  | E'' : forall svm : list (string * Value''),  NoDup (fst (split svm)) -> Value''
  | C'' : list Value'' -> Value''
]] *)
(** Problem:
[[
Error: Non strictly positive occurrence of "_UNBOUND_REL_2" in
 "forall svm : list (string * Value''), NoDup (fst (split svm)) -> Value''"
]] *)

(** The NoDup predicate is defined in the
#<a href="http://coq.inria.fr/stdlib/Coq.Lists.List.html&#35;ReDun">List</a>#
module.
If we need a special equality relation on keys
(don't think so, syntactic equality on strings should be fine) then have a look at
#<a href="http://coq.inria.fr/stdlib/Coq.Lists.SetoidList.html&#35;NoDupA">SetoidList</a>#
that defines a [NoDupA] predicate. 
Also note that many of the functions in
#<a href="http://coq.inria.fr/stdlib/Coq.FSets.FMapWeakList.html&#35;Raw.NoDupA">FMapWeakList</a>#
(which has a [list (key * elt)] as implementation)
are defined to preserve the [NoDupA] predicate (maybe under certain conditions).
*)

(** The main advantage of baking distinctness into the definition of [Value] would be
that then all operations on values could assume this invariant for their inputs
and  would be forced to preserve it for the outputs.

One alternative to this would be to enforce distinctness after defining RawValue. A first attempt would be:
[[
Inductive GoodValue : RawValue -> Prop :=
  | good_G : forall g, GoodValue g
  | good_E : forall svm : list (string * RawValue),  NoDup (fst (split svm)) -> AllSnd GoodValue svm -> GoodValue (E svm)
  | good_C : forall vb : list RawValue, All GoodValue vb -> GoodValue (C vb)
]]
*)
(** Unfortunately this does not work since the occurrences of [GoodValue] in [AllSnd GoodValue svm] and 
[All GoodValue vb] are not strictly positive. *)

(** Defining All inductively makes this work. *)

Inductive IndAll (X : Type) (P : X -> Prop) : list X -> Prop :=
  | All_nil : IndAll X P nil
  | All_cons : forall x xs, P x -> IndAll X P xs -> IndAll  X P (x :: xs).
Implicit Arguments IndAll [X].

Lemma All_equiv_IndAll : forall (X : Type) (P : X -> Prop) (xs : list X),
  All P xs <-> IndAll P xs.
Proof.
  intros X P xs. split; intro H.
  Case "->".
    induction xs as [| x xs'].
    SCase "xs = nil". apply All_nil.
    SCase "xs = x :: xs'".  simpl in H. destruct H as [H1 H2].
      apply All_cons. apply H1. apply IHxs'. apply H2.
  Case "<-".
    induction xs as [| x xs']; simpl.
    SCase "xs = nil". trivial.
    SCase "xs = x :: xs'".
      inversion H. split.
        apply H2.
        apply IHxs'. assumption.
Qed.

Lemma All_In : forall (X : Type) (P : X -> Prop) x xs,
  IndAll P xs -> In x xs -> P x.
Proof.
  intros X P x xs H2 H1. induction xs. inversion H1.
    inversion H2. simpl in H1. inversion H1. subst. assumption.
      apply IHxs; assumption.
Qed.


Inductive GoodValue : RawValue -> Prop :=
  | good_G : forall g, GoodValue (G g)
  | good_E : forall svm : list (string * RawValue),  NoDup (fst (split svm)) -> IndAll GoodValue (snd (split svm)) -> GoodValue (E svm)
  | good_C : forall vb : list RawValue, IndAll GoodValue vb -> GoodValue (C vb).

(** We can now define the operations on [GValue]s instead of [RawValue]s, where *)
Definition GValue := {v : RawValue | GoodValue v}.

(** ** Specifying the equality relation on values *)

(** A first attempt at defining the value equality relation
was to try to use the
#<a href="http://coq.inria.fr/library/Coq.Lists.List.html&#35;Permutation">
Permutation</a># predicate on lists
(it also provides two induction principles [Permutation_ind] and [Permutation_ind_bis]
that could be useful in proofs): *)

Inductive EqValue_first_try : RawValue -> RawValue -> Prop :=
  | eq_G_first_try : forall g1 g2, g1 = g2 -> EqValue_first_try (G g1) (G g2)
  | eq_E_first_try : forall svm1 svm2,
                              NoDup (fst (split svm1))
                              -> NoDup (fst (split svm2))
                              -> Permutation svm1 svm2
                -> EqValue_first_try (E svm1) (E svm2)
  | eq_C_first_try : forall vb1 vb2, Permutation vb1 vb2 -> EqValue_first_try (C vb1) (C vb2).

(** This first attempt at defining [EqValue] is simple but _wrong_.
Instead of using Permutation which uses [=] internally, 
the EqValue predicate should be applied recursively when checking the
equality of entities and collections. *)

(** Defining my own [PermutationA] -
the bad thing about this is that I lose all the nice results already proved about [Permutation].
Is there anything like PermutationA in the Coq library which we can use directly? *)
Section PermutationA.
  Variable A : Type.
  Variable eqA : A -> A -> Prop.

  Inductive PermutationA : list A -> list A -> Prop :=
      | perma_nil: PermutationA nil nil
      | perma_skip: forall (x x':A) (l l':list A), eqA x x' -> PermutationA l l'
                    -> PermutationA (cons x l) (cons x' l')
      | perma_swap: forall (x x' y y':A) (l:list A), eqA x x' -> eqA y y'
                    -> PermutationA (cons y (cons x l)) (cons x' (cons y' l))
      | perma_trans: forall (l l' l'':list A), PermutationA l l' -> PermutationA l' l'' -> PermutationA l l''.
End PermutationA.
Implicit Arguments PermutationA [A].

Definition LiftBinRelToPairs (X Y: Type) (R1 : X -> X -> Prop) (R2 : Y -> Y -> Prop) :=
  fun p1 p2 : X * Y => R1 (fst p1) (fst p2) /\ R2 (snd p2) (snd p2).

(** We can now give a second definition for [EqValue]. *)

Inductive EqValue_second_try : RawValue -> RawValue -> Prop :=
  | eq_G_second_try : forall g1 g2, g1 = g2 -> EqValue_second_try (G g1) (G g2)
  | eq_E_second_try : forall svm1 svm2,
                NoDup (fst (split svm1))
                -> NoDup (fst (split svm2))
                -> PermutationA (LiftBinRelToPairs string RawValue (fun s1 s2 => s1 = s2) EqValue_second_try) svm1 svm2
                -> EqValue_second_try (E svm1) (E svm2)
  | eq_C_second_try : forall vb1 vb2,
                PermutationA EqValue_second_try vb1 vb2 -> EqValue_second_try (C vb1) (C vb2).

(** This definition fulfills the Coq strict positivity restriction for inductive definitions. *)

(** Note that [eq_E] has the explicit condition that svm1 and svm2 have no duplicate keys
(and here we can add this condition without breaking anything).
Without this condition are forced to take duplicate keys into account in the definition.
This doesn't make things much more complicated. *)

Lemma orb_false : forall b1 b2, b1 || b2 = false -> b1 = false /\ b2 = false.
  Proof. intros. split; destruct b1; destruct b2; auto. Qed.

Section remove_dup.
  Hypotheses X Y : Type.
  Hypothesis eqX_dec : forall x y : X, {x = y}+{x <> y}.

  Fixpoint remove_dup_helper (avoid : list X) (xys: list (X * Y)) :=
    match xys with
    | nil => nil
    | (x,y) :: xys' =>
      if contained eqX_dec x avoid then
        remove_dup_helper avoid xys'
      else
        (x,y) :: (remove_dup_helper (x :: avoid) xys')
    end.

  Definition remove_dup := remove_dup_helper nil.

  Lemma remove_dup_helper_sound : forall xys avoid,
     NoDup (fst (split (remove_dup_helper avoid xys))) /\
     forall xa, In xa avoid -> ~(In xa (fst (split (remove_dup_helper avoid xys)))).
  Proof.
    induction xys as [| [x y] xys']; simpl.
    Case "xys = nil". split. apply NoDup_nil. intros xa H Hc. inversion Hc.
    Case "xys = (x,y) :: xys'". intro avoid. 
      destruct (contained eqX_dec x).
      (*SCase "contained eqX_dec x = "true" (duplicate)". -- Coq bug *) apply IHxys'.
      SCase "b = false (fresh)".
        split;
        specialize IHxys' with (avoid:=x::avoid);
        destruct IHxys' as [IH1 IH2].
        SSCase "1st conjunct".
          rewrite split_cons. simpl.
          apply NoDup_cons.
          SSSCase "~In". apply IH2. simpl. left. reflexivity.
          SSSCase "NoDup". apply IH1.
        SSCase "2nd conjunct".
          intros xa H1 H2.
          simpl in IH2. rewrite split_cons in H2. simpl in H2.
          destruct H2.
          SSSCase "x = xa". rewrite H in *. apply n. apply H1.
          SSSCase "In xa ...". eapply IH2. right. apply H1. apply H.
  Qed.
  
  Theorem remove_dup_sound : forall xys, NoDup (fst (split (remove_dup xys))).
  Proof. intro xys. specialize remove_dup_helper_sound with (xys:=xys)(avoid:=nil).
         intro H. destruct H as [H1 H2]. apply H1.
  Qed.

  Lemma remove_dup_helper_complete1' : forall xys avoid,
    (forall x y, In (x,y) xys -> ~ In x avoid ->
      exists y', (assoc eqX_dec x xys = Some y') /\ In (x,y') (remove_dup_helper avoid xys)).
  Proof.
    induction xys as [| [x' y'] xys']; intros avoid x y H1 H2. simpl in H1. inversion H1.
    simpl in *. destruct H1.
    Case "x'=x /\ y'=x".
      inversion H. subst.
      eapply not_in_contained_right in H2. destruct H2 as [pf H2]. erewrite H2. try assumption. simpl. exists y. split.
      SCase "assoc".
        fold value. remember (eqX_dec x x). destruct s. simpl. reflexivity.
        apply False_ind. apply n. reflexivity.
      SCase "= \/ In". left. reflexivity.
    Case "In (x,y) xys'".
      destruct (contained eqX_dec x' avoid).
      SCase "contained eqX_dec x' avoid=true".
        destruct (eqX_dec x x') as [H1 | H1].
        SSCase "x = x'". exists y'. split.
          reflexivity. subst. apply H2 in i. inversion i.
        SSCase "x <> x'".        
          apply (IHxys' avoid) in H. destruct H as [y0 [H' H'']].
          exists y0. split. simpl. apply H'.
          apply H''. assumption.
      SCase "b=false". simpl.
        remember (eqX_dec x x') as H1. clear HeqH1. destruct H1 as [H1 | H1].
        SSCase "x = x'". exists y'. split.
          reflexivity. left. subst. reflexivity.
        SSCase "x <> x'".
          specialize IHxys' with (avoid:=x'::avoid) (x:=x) (y:=y).
          apply IHxys' in H. destruct H as [y0 [H' H'']].
          exists y0. split. simpl. assumption.
          right. apply H''.
          intro Hc. simpl in Hc. destruct Hc. symmetry in H0. apply H1 in H0. inversion H0.
          apply H2. assumption.
  Qed.
  
  Lemma remove_dup_helper_complete1 : forall xys avoid,
    (forall x y, In (x,y) xys -> ~ In x avoid -> exists y', In (x,y') (remove_dup_helper avoid xys)).
  Proof. intros. eapply remove_dup_helper_complete1' in H0. destruct H0 as [y' [H1 H2]].
    exists y'. eassumption. eassumption. Qed.

  Lemma remove_dup_helper_complete2 : forall xys avoid,
    (forall x y, In (x,y) (remove_dup_helper avoid xys) -> In (x,y) xys).
  Proof.
    induction xys as [| [x' y'] xys']; intros avoid x y H. simpl in H. inversion H.
    simpl in *.
    destruct (contained eqX_dec x' avoid).
    Case "contained=left". right. eapply IHxys'. eassumption.
    Case "contained=right". simpl in H. destruct H. left. assumption. right.
      eapply IHxys'. eassumption.
  Qed.

  Theorem remove_dup_complete1' : forall xys,
    (forall x y, In (x,y) xys ->
      exists y', (assoc eqX_dec x xys = Some y') /\ In (x,y') (remove_dup xys)).
  Proof.
    intro xys. unfold remove_dup. intros x y H.
    eapply remove_dup_helper_complete1'. eassumption. intro Hc. inversion Hc.
  Qed.

  Corollary remove_dup_complete1 : forall xys,
    (forall x y, In (x,y) xys -> exists y', In (x,y') (remove_dup xys)).
  Proof.
    intro xys. unfold remove_dup. intros x y H.
    eapply remove_dup_helper_complete1. eassumption. intro Hc. inversion Hc.
  Qed.

  Theorem remove_dup_complete2 : forall xys,
    (forall x y, In (x,y) (remove_dup xys) -> In (x,y) xys).
  Proof.
    intro xys. unfold remove_dup. intros x y H.
    apply remove_dup_helper_complete2 with (avoid := nil). assumption.
  Qed.

  Fixpoint remove (x : X) (xys : list (X * Y)) :=
  match xys with
  | nil => nil
  | (x',y') :: xys' => if eqX_dec x x' then remove x xys' else (x',y') :: remove x xys'
  end.

End remove_dup.
Implicit Arguments remove_dup [X Y].

Theorem eq_str_dec : forall (s1 s2 : string), {s1 = s2} + {s1 <> s2}.
Proof. repeat (decide equality). Qed.

(*--- EqRawValueBegin *)
Inductive EqRawValue : RawValue -> RawValue -> Prop :=
  | eq_rval_G : forall g1 g2, g1 = g2 -> EqRawValue (G g1) (G g2)
  | eq_rval_E : forall svm1 svm2,
                PermutationA (LiftBinRelToPairs string RawValue (fun s1 s2 => s1 = s2) EqRawValue)
                             (remove_dup eq_str_dec svm1) (remove_dup eq_str_dec svm2)
                -> EqRawValue (E svm1) (E svm2)
  | eq_rval_C : forall vb1 vb2, PermutationA EqRawValue vb1 vb2 -> EqRawValue (C vb1) (C vb2).
(*--- EqRawValueEnd *)

(** XXX TODO: prove that [EqValue] is an equivalence relation *)

Require Import Relations.

Theorem equivalence_EqRawValue : equivalence RawValue EqRawValue.
Admitted.

(** So using [PermutationA] and [NoDup] or maybe [remove_dup] we can quite easily define the equality
relation on values.
In any case, we still need to define an equality _function_ on values [beq_val]
which can be used in all other functions in our intended model.
The relation [EqValue] is only the specification of the concrete implementation
[beq_val]. *)



(** Syntactic equality function on values *)

Fixpoint syn_beq_rval (v1 : RawValue) (v2 : RawValue) : bool :=
  match v1, v2 with
  | G g1, G g2 => beq_gen g1 g2
  | E svm1, E svm2 =>
    (fix syn_beq_svm (svm1 svm2: list (string*RawValue)) : bool :=
      match svm1, svm2 with
      | nil, nil => true
      | (s1,v1) :: svm1', (s2,v2) :: svm2' =>
        beq_str s1 s2 && syn_beq_rval v1 v2 && syn_beq_svm svm1' svm2'
      | _, _ => false
      end
    ) svm1 svm2
  | C vb1, C vb2 =>
    (fix syn_beq_vb (vb1 vb2: list RawValue) : bool :=
      match vb1, vb2 with
      | nil, nil => true
      | v1 :: vb1', v2 :: vb2' => syn_beq_rval v1 v2 && syn_beq_vb vb1' vb2'
      | _, _ => false
      end
    ) vb1 vb2
  | _, _ => false
  end.

Lemma syn_beq_rval_refl : forall v, syn_beq_rval v v = true.
Proof.
  Admitted.

Theorem syn_beq_rval_true_implies_EqValue : forall v1 v2,
  syn_beq_rval v1 v2 = true -> EqRawValue v1 v2.
Admitted.

Lemma syn_beq_rval_equals : forall v1 v2, syn_beq_rval v1 v2 = true ->
  v1 = v2.
Proof.
  intros. unfold syn_beq_rval in H. destruct v1. destruct v2.
  Case "Generals".
    destruct g; destruct g0; inversion H. remember (BinInt.Zcompare z z0) as r. destruct r; try inversion H1.
    symmetry in Heqr. apply Zcompare.Zcompare_Eq_eq in Heqr. rewrite Heqr. reflexivity.
Admitted.

(** Linear ordering on [RawValue] *)

Require Export ZArith.

(* Lexicographic order on strings *)

Fixpoint cmp_bool (b1 b2 : bool) : comparison :=
  match b1, b2 with
  | false, false => Eq
  | false, true => Lt
  | true, false => Gt
  | true, true => Eq
  end.

Require Import Ascii.

Fixpoint cmp_ascii (c1 c2 : ascii) : comparison :=
  match c1, c2 with
  | Ascii n0 n1 n2 n3 n4 n5 n6 n7, Ascii m0 m1 m2 m3 m4 m5 m6 m7 =>
    match cmp_bool n0 m0 with
      | Lt => Lt
      | Gt => Gt
      | Eq =>
        match cmp_bool n1 m1 with
        | Lt => Lt
        | Gt => Gt
        | Eq =>
          match cmp_bool n2 m2 with
          | Lt => Lt
          | Gt => Gt
          | Eq =>
            match cmp_bool n3 m3 with
            | Lt => Lt
            | Gt => Gt
            | Eq =>
              match cmp_bool n4 m4 with
              | Lt => Lt
              | Gt => Gt
              | Eq =>
                match cmp_bool n5 m5 with
                | Lt => Lt
                | Gt => Gt
                | Eq =>
                  match cmp_bool n6 m6 with
                  | Lt => Lt
                  | Gt => Gt
                  | Eq => cmp_bool n7 m7
                  end
                end
              end
            end
          end
        end
    end
  end.

Fixpoint cmp_str (s1 s2 : string) : comparison :=
  match s1, s2 with
  | EmptyString, EmptyString => Eq
  | EmptyString, _ => Lt
  | _, EmptyString => Gt
  | String c1 s1', String c2 s2' =>
      match cmp_ascii c1 c2 with
      | Lt => Lt
      | Gt => Gt
      | Eq => cmp_str s1' s2'
      end
  end.

(** Order is: [G_Integer _ < G_Text _ < G_Logical false < G_Logical true < G_Null] *)
Fixpoint cmp_gen (g1 g2 : General) : comparison :=
  match g1, g2 with
  | G_Integer i1, G_Integer i2 => Zcompare i1 i2
  | G_Integer _, _ => Lt
  | _, G_Integer _ => Gt
  | G_Text s1, G_Text s2 => cmp_str s1 s2
  | G_Text _, _ => Lt
  | _, G_Text _ => Gt
  | G_Logical b1, G_Logical b2 => cmp_bool b1 b2
  | G_Logical _, _ => Lt
  | _, G_Logical _ => Gt
  | G_Null, G_Null => Eq
  end.

(** Order is: [G _< E _ < C _] and on entities and collections lexicographic *)
Fixpoint cmp_rval (v1 v2 : RawValue) : comparison :=
  match v1, v2 with
  | G g1, G g2 => cmp_gen g1 g2
  | G _, _ => Lt
  | _, G _ => Gt
  | E svm1, E svm2 =>
    (fix cmp_svm (svm1 svm2: list (string*RawValue)) : comparison :=
      match svm1, svm2 with
      | nil, nil => Eq
      | nil, _ => Lt
      | _, nil => Gt
      | (s1,v1) :: svm1', (s2,v2) :: svm2' =>
        match cmp_str s1 s2 with
        | Lt => Lt
        | Gt => Gt
        | Eq => 
          match cmp_rval v1 v2 with
          | Lt => Lt
          | Gt => Gt
          | Eq => cmp_svm svm1' svm2'
        end
      end
     end
    ) svm1 svm2
  | E _, _ => Lt
  | _, E _ => Gt
  | C vb1, C vb2 =>
    (fix cmp_vb (vb1 vb2: list RawValue) : comparison :=
      match vb1, vb2 with
      | nil, nil => Eq
      | nil, _ => Lt
      | _, nil => Gt
      | v1 :: vb1', v2 :: vb2' => 
        match cmp_rval v1 v2 with
        | Lt => Lt
        | Gt => Gt
        | Eq => cmp_vb vb1' vb2'
        end
      end
    ) vb1 vb2
  end.

(** We prove that [cmp_val] (order) is a generalization of [syn_beq_rval]
(syntactic equality). *)

Lemma cmp_ascii_eq : forall c1 c2, cmp_ascii c1 c2 = Eq -> beq_ascii c1 c2 = true.
Proof.
  intros c1 c2 H.
  destruct c1; destruct c2; simpl in *.
  Admitted.
(* this kills Coq 
  destruct b; destruct b0; destruct b1; destruct b2; 
  destruct b3; destruct b4; destruct b5; destruct b6;
  destruct b7; destruct b8; destruct b9; destruct b10; 
  destruct b11; destruct b12; destruct b13; destruct b14;
  simpl in *.
*)

Lemma eq_cmp_ascii : forall c1 c2, beq_ascii c1 c2 = true -> cmp_ascii c1 c2 = Eq.
Admitted.

Lemma if_false_eq_true : forall (b1 b2 : bool),
  b2 = true -> b1 = true -> (if b1 then b2 else false) = true.
Proof. intros b1 b2. destruct b1; intros H1 H2. assumption. assumption. Qed.

Lemma if_false_eq_true_rev : forall (b1 b2 : bool),
  (if b1 then b2 else false) = true -> b2 = true /\ b1 = true.
Proof. intros b1 b2. destruct b1; intros H; split;
  try reflexivity; try assumption; try inversion H.
Qed.

Lemma match_Eq_Eq : forall c1 c2,
  c1 = Eq -> c2 = Eq ->
           match c1 with
            | Eq => c2
            | Lt => Lt
            | Gt => Gt
            end = Eq.
Proof. intros c1 c2. destruct c1; destruct c2; intros H1 H2; 
  try reflexivity; try inversion H1; try inversion H2. Qed.

Lemma match_Eq_Eq_rev : forall c1 c2,
  match c1 with
  | Eq => c2
  | Lt => Lt
  | Gt => Gt
  end = Eq -> c1 = Eq /\ c2 = Eq.
Proof. intros c1 c2. destruct c1; destruct c2; intro H; split;
  try reflexivity; try inversion H1; try inversion H2; try inversion H;
  try assumption.
Qed.

Lemma cmp_str_eq : forall s1 s2, cmp_str s1 s2 = Eq -> beq_str s1 s2 = true.
Proof.
  induction s1; intros s2 H; destruct s2; simpl in *; try reflexivity; try (inversion H).
  rewrite (if_false_eq_true (beq_ascii a a0) (beq_str s1 s2)); try reflexivity.
    apply IHs1.
    apply (match_Eq_Eq_rev (cmp_ascii a a0) (cmp_str s1 s2))in H.
    destruct H. assumption.
    apply cmp_ascii_eq.
    apply (match_Eq_Eq_rev (cmp_ascii a a0) (cmp_str s1 s2))in H.
    destruct H. assumption.
Qed.

Lemma eq_cmp_str : forall s1 s2, beq_str s1 s2 = true -> cmp_str s1 s2 = Eq.
Proof.
  induction s1; intros s2 H; destruct s2; simpl in *; try reflexivity; try (inversion H).
  apply match_Eq_Eq.
    apply (if_false_eq_true_rev (beq_ascii a a0)) in H. destruct H.
    apply eq_cmp_ascii. assumption.

    apply IHs1.
    apply (if_false_eq_true_rev (beq_ascii a a0)) in H. destruct H.
    assumption.
Qed.

Lemma match_Eq : forall c,
  c = Eq -> match c with
            | Eq => true
            | Lt => false
            | Gt => false
            end = true.
Proof. intro c. destruct c; intro H. reflexivity. inversion H. inversion H. Qed.

Lemma match_Eq_rev : forall c,
  match c with
  | Eq => true
  | Lt => false
  | Gt => false
  end = true -> c = Eq.
Proof. intro c. destruct c; intro H. reflexivity. inversion H. inversion H. Qed.

Lemma cmp_gen_eq : forall g1 g2, cmp_gen g1 g2 = Eq -> beq_gen g1 g2 = true.
Proof.
  intros g1 g2 H.
  destruct g1; destruct g2; simpl in *; try reflexivity; try inversion H.
    apply match_Eq. apply H.
    apply cmp_str_eq. assumption.
    destruct b; destruct b0; simpl in *; try reflexivity; try inversion H.
Qed.

Lemma eq_cmp_gen : forall g1 g2, beq_gen g1 g2 = true -> cmp_gen g1 g2 = Eq.
Proof.  
  intros g1 g2 H.
  destruct g1; destruct g2; simpl in *; try reflexivity; try inversion H.
    apply match_Eq_rev in H. apply H.
    apply eq_cmp_str. assumption.
    destruct b; destruct b0; simpl in *; try reflexivity; try inversion H.
Qed.

Theorem cmp_rval_eq : forall v1 v2, cmp_rval v1 v2 = Eq -> syn_beq_rval v1 v2 = true.
Proof.
  induction v1; intros v2 H1; destruct v2; simpl in *; try inversion H1.
    apply cmp_gen_eq; assumption.
    Case "E".
      clear H2. rename svm into svm1; rename l into svm2. generalize dependent svm2.
      induction svm1 as [| [s1 v1] svm1']; intros svm2 H1;
      destruct svm2 as [| [s2 v2] svm2']; try reflexivity; try inversion H1.
        SCase "svm1 = (s1,v1) :: svm1' /\ svm2 = (s2,v2) :: svm2'".
        clear H2; simpl in *; destruct H as [H' H''].
        apply match_Eq_Eq_rev in H1; destruct H1 as [H11 H12].
        apply match_Eq_Eq_rev in H12; destruct H12 as [H121 H122].
        apply cmp_str_eq in H11; rewrite H11.
        apply H' in H121; rewrite H121; simpl.
        apply IHsvm1'. apply H''. apply H122.
    Case "C".
      clear H2. rename vb into vb1; rename l into vb2. generalize dependent vb2.
      induction vb1 as [| v1 vb1']; intros vb2 H1;
      destruct vb2 as [| v2 vb2']; try reflexivity; try inversion H1.
        SCase "vb1 = v1 :: vb1' /\ vb2 = v2 :: vb2'".
        clear H2; simpl in *; destruct H as [H' H''].
        apply match_Eq_Eq_rev in H1; destruct H1 as [H11 H12].
        apply H' in H11; rewrite H11; simpl.
        apply IHvb1'. apply H''. apply H12.
Qed.

Theorem eq_cmp_rval : forall v1 v2, syn_beq_rval v1 v2 = true -> cmp_rval v1 v2 = Eq.
Proof.
  induction v1; intros v2 H1; destruct v2; simpl in *; try inversion H1.
    apply eq_cmp_gen; assumption.
    Case "E".
      (* XXX try to abstract some of this duplication into a tactic *)
      clear H2. rename svm into svm1; rename l into svm2. generalize dependent svm2.
      induction svm1 as [| [s1 v1] svm1']; intros svm2 H1;
      destruct svm2 as [| [s2 v2] svm2']; try reflexivity; try inversion H1.
        SCase "svm1 = (s1,v1) :: svm1' /\ svm2 = (s2,v2) :: svm2'".
        clear H2; simpl in *; destruct H as [H' H''].
        rewrite <- andb_assoc in H1; apply and_eq_true in H1; destruct H1 as [H11 H12].
        apply and_eq_true in H12; destruct H12 as [H121 H122].
        apply eq_cmp_str in H11; rewrite H11.
        apply H' in H121; rewrite H121; simpl.
        apply IHsvm1'. apply H''. apply H122.
    Case "C".
      clear H2. rename vb into vb1; rename l into vb2. generalize dependent vb2.
      induction vb1 as [| v1 vb1']; intros vb2 H1;
      destruct vb2 as [| v2 vb2']; try reflexivity; try inversion H1.
        SCase "vb1 = v1 :: vb1' /\ vb2 = v2 :: vb2'".
        clear H2; simpl in *; destruct H as [H' H''].
        apply and_eq_true in H1; destruct H1 as [H11 H12].
        apply H' in H11; rewrite H11; simpl.
        apply IHvb1'. apply H''. apply H12.
Qed.



Definition le_cmp_val (v1 v2 : RawValue) : Prop := 
  match cmp_rval v1 v2 with
  | Le => True
(** XXX WTF -- Coq bug!?
[[ 
  | Eq => True
  | Gt => False
]]
*)
  end.

Lemma le_cmp_val_refl : forall v, le_cmp_val v v.
Admitted.

Lemma le_cmp_val_trans : forall v1 v2 v3,
  le_cmp_val v1 v2 /\ le_cmp_val v2 v3 -> le_cmp_val v1 v3.
Admitted.

Lemma le_cmp_val_asym : forall v1 v2,
  le_cmp_val v1 v2 -> le_cmp_val v2 v1 -> v1 = v2.
Admitted.

(** ** Normalization function *)

Fixpoint remove_dup_e_val (v : RawValue) :=
  match v with
  | G g => G g
  | E svm => 
    let svm' := map (fun sv => match sv with (s,v) => (s, remove_dup_e_val v) end) svm in
    E (remove_dup eq_str_dec svm')
  | C vb => C (map remove_dup_e_val vb)
  end.

(*--- SortedSvmVbBegin *)
Definition le_sv (sv1 sv2 : (string * RawValue)) : Prop :=
  match sv1, sv2 with
  (s1,_), (s2,_) => cmp_str s1 s2 = Lt \/ cmp_str s1 s2 = Eq
  end.
Definition Sorted_svm (svm : list (string * RawValue)) : Prop :=
  Sorted le_sv svm.
Definition le_rval (v1 v2 : RawValue) : Prop :=
  cmp_rval v1 v2 = Lt \/ cmp_rval v1 v2 = Eq.
Definition Sorted_vb (vb : list RawValue) : Prop := Sorted le_rval vb.
(*--- SortedSvmVbEnd *)

Lemma le_rval_trans : transitive RawValue le_rval.
Admitted.

Lemma le_rval_total : forall x y : RawValue, le_rval x y \/ le_rval y x.
Admitted.

Lemma le_sv_trans : transitive (string*RawValue) le_sv.
Admitted.

Lemma le_sv_total : forall x y : (string*RawValue), le_sv x y \/ le_sv y x.
Admitted.

Definition le_rval_dec : forall x y : RawValue, {le_rval x y} + {~ le_rval x y}.
Proof. intros. unfold le_rval. remember (cmp_rval x y) as c. destruct c.
  apply left. right. reflexivity.
  apply left. left. reflexivity.
  apply right. intro H. destruct H; inversion H.
Qed.

Definition le_sv_dec:
  forall sv1 sv2 : string * RawValue, {le_sv sv1 sv2} + {~ le_sv sv1 sv2}.
Proof. intros. destruct sv1 as [s1 v1]. destruct sv2 as [s2 v2].
  unfold le_sv. remember (cmp_str s1 s2) as c. destruct c.
  apply left. right. reflexivity.
  apply left. left. reflexivity.
  apply right. intro H. destruct H; inversion H.
Qed.

Fixpoint sort_c_val (v : RawValue) := 
  match v with
  | G g => G g
  | E svm => E (map (fun sv => match sv with (s,v) => (s, sort_c_val v) end) svm)
  | C vb => let vb' := map sort_c_val vb in C (insertion_sort le_rval le_rval_dec vb')
  end.

Fixpoint sort_e_val (v : RawValue) :=
  match v with
  | G g => G g
  | E svm => 
    let svm' := map (fun sv => match sv with (s,v) => (s, sort_e_val v) end) svm in
    E (insertion_sort le_sv le_sv_dec svm')
  | C vb => C (map remove_dup_e_val vb)
  end.

(*--- NormalizeBegin *)
Definition normalize_val (v : RawValue) := 
  sort_c_val (sort_e_val (remove_dup_e_val v)).
(*--- NormalizeEnd *)

Lemma normalize_preserves_EqValue :
  forall v, EqRawValue v (normalize_val v).
Admitted.

(** ** Equality function on values *)

(*--- RawValueBEqBegin *)
Definition beq_rval (v1 v2 : RawValue) :=
  syn_beq_rval (normalize_val v1) (normalize_val v2).
(*--- RawValueBEqEnd *)
  
(** We prove that [beq_val] satisfies its specification,
ie. the relation [EqValue] *)


Theorem beq_val_correct : forall v1 v2, beq_rval v1 v2 = true <-> EqRawValue v1 v2.
Proof.
  intros v1 v2. unfold beq_rval. split; intro H.
  Case "->".
    apply syn_beq_rval_true_implies_EqValue in H.
    (* use: normalize_preserves_EqValue + EqValue equivalence (2 x transitivity) *)
    admit.
  Case "<-".
    (* need stronger property: on values in normal form EqValue implies syn_beq_rval 
       -- this means I need to define when a value is in normal form *)
    admit.
Qed.



