(*
    Copyright (c) 2010  Catalin Hritcu
    Licensed under the Creative Commons - Attribution 3.0 Unported
    The license is available at: http://creativecommons.org/licenses/by/3.0/
*)


Require Export Scalars.
Require Export List.
Require Export String.

(** * Values *)

(** We tell Coq not (to try) to generate induction principles for RawValue.
The generated induction principles are useless anyway (no IH for E and C):
[[
RawValue_rect
     : forall P : RawValue -> Type,
       (forall g : General, P (G g)) ->
       (forall l : list (string * RawValue), P (E l)) ->
       (forall l : list RawValue, P (C l)) -> forall v : RawValue, P v
]]
*)
Unset Elimination Schemes.

(*--- ValueBegin *)
Inductive RawValue : Type :=
  | G : General -> RawValue
  | E : list (string * RawValue) -> RawValue
  | C : list RawValue -> RawValue.
(*--- ValueEnd *)

Set Elimination Schemes.


(** ** Strong induction principles for values (proved by hand) *)


(* A first failed attempt to define and prove RawValue_ind

Definition In_range_map (A B : Type) (v : B) (svm : list (A * B)) : Prop := In v (snd (split svm)).
Implicit Arguments In_range_map [A B].

(** Simple auxiliary lemma *)
Lemma split_cons : forall X Y (xy : X*Y) (xys : list (X*Y)), 
  split (xy :: xys) = ((fst xy) :: (fst (split xys)), (snd xy) :: (snd (split xys))).
Proof. intros. destruct xy. simpl. destruct (split xys). reflexivity. Qed.

Lemma RawValue_ind :
       forall P : RawValue -> Prop,
       (forall g : General, P (G g)) ->
       (forall svm : list (string * RawValue), (forall v, In_range_map v svm -> P v) ->  P (E svm)) ->
       (forall vb : list RawValue, (forall v, In v vb -> P v) ->  P (C vb))
       -> forall v : RawValue, P v.
Proof.
  intros P Hg He Hc.
  fix 1. (* added this *)
  destruct v as [g | svm | vb].
  Case "v = G g". apply Hg.
  Case "v = E svm".
    induction svm as [| sv svm']; apply He; intros v H.
    SCase "svm=nil". inversion H.
    SCase "svn=sv :: svm'". unfold In_range_map in H.
      rewrite split_cons in H. simpl in H. 
      inversion H as [H1 | H2]. (*!!! this doesn't work for Type *)
      rewrite <- H1.
      apply RawValue_ind. admit.
  Case "v = C svm".
    apply Hc. intros v H.
    induction vb as [| v' vb'].
    SCase "vb=nil". inversion H.
    SCase "vb = v' vb'". unfold In in H. admit.
Qed. (* Doesn't work -- fix *)
Other problem:
Error: Inversion would require case analysis on sort Type which is not
allowed for inductive definition or.
Error: Case analysis on sort Type is not allowed for inductive definition or.
*)


(** **** A simple strong induction principle, which unfortunately only works for Prop *)

(**
[[
RawValue_ind
     : forall P : RawValue -> Prop,
       (forall g : General, P (G g)) ->
       (forall svm : list (string * RawValue), AllSnd P svm -> P (E svm)) ->
       (forall vb : list RawValue, All P vb -> P (C vb)) ->
       forall v : RawValue, P v
]]

The proof technique is borrowed from
#<a href="http://adam.chlipala.net/cpdt/html/InductiveTypes.html">
Adam Chlipala's book (Section 3.7 Nested Inductive Types)</a>#:
*)

Section All.
  Variable X : Type.
  Variable P : X -> Prop.
  Fixpoint All (xs : list X) : Prop :=
    match xs with
      | nil => True
      | cons x xs' => P x /\ All xs'
    end.
End All.
Implicit Arguments All [X].

Section AllSnd.
  Variable X Y : Type.
  Variable P : Y -> Prop.
  Fixpoint AllSnd (xys : list (X *Y)) : Prop :=
    match xys with
      | nil => True
      | (x,y) :: xys' => P y /\ AllSnd xys'
    end.
End AllSnd.
Implicit Arguments AllSnd [X Y].

Section RawValue_ind.
  Variable P : RawValue -> Prop.
  Hypothesis Hg : forall g : General, P (G g).
  Hypothesis He : forall svm : list (string * RawValue), AllSnd P svm ->  P (E svm).
  Hypothesis Hc : forall vb : list RawValue, All P vb ->  P (C vb).

  Fixpoint RawValue_ind (v : RawValue) : P v :=
  match v with
  | G g => Hg g
  | E svm => He svm 
     ((fix list_pairs_Value_ind (svm : list (string * RawValue)) : AllSnd P svm :=
       match svm with
       | nil => I
       | (s,v') :: vb' => conj (RawValue_ind v') (list_pairs_Value_ind vb')
       end
     ) svm)
  | C vb => Hc vb
     ((fix list_Value_ind (vb : list RawValue) : All P vb :=
       match vb with
       | nil => I
       | v' :: vb' => conj (RawValue_ind v') (list_Value_ind vb')
       end
     ) vb)
  end.
End RawValue_ind.


(* A first failed attempt to define RawValue_rect 
-- This looks a lot like fold now
Section AllT.
  Variable X : Type.
  Variable P : X -> Type.
  Variable f_reduce : Type -> Type -> Type.
  Variable f_default : Type.
  Fixpoint AllT (xs : list X) : Type :=
    match xs with
      | nil => f_default
      | cons x xs' => f_reduce (P x) (AllT xs')
    end.
End AllT.
Implicit Arguments AllT [X].

Section AllSndT.
  Variable X Y : Type.
  Variable P : Y -> Type.
  Variable f_reduce : Type -> Type -> Type.
  Variable f_default : Type.
  Fixpoint AllSndT (xys : list (X *Y)) : Type :=
    match xys with
      | nil => f_default
      | (x,y) :: xys' => f_reduce (P y) (AllSndT xys')
    end.
End AllSndT.
Implicit Arguments AllSndT [X Y].

Section RawValue_rect.
  Variable P : RawValue -> Type.
  Variable f_reduce : Type -> Type -> Type.
  Variable f_default : Type.
  Hypothesis Hg : forall g : General, P (G g).
  Hypothesis He : forall svm : list (string * RawValue), AllSndT P f_reduce f_default svm ->  P (E svm).
  Hypothesis Hc : forall vb : list RawValue, AllT P f_reduce f_default vb ->  P (C vb).

  Fixpoint RawValue_rect (v : RawValue) : P v :=
  match v as v0 return P v0 with
  | G g => Hg g
  | E svm => He svm 
     ((fix list_pairs_Value_ind (svm : list (string * RawValue)) : AllSndT P f_reduce f_default svm :=
       match svm as svm0 return AllSndT P f_reduce f_default svm0 with
       | nil => f_default
       | (s,v') :: svm' => f_reduce (RawValue_rect v') (list_pairs_Value_ind svm')
       end
     ) svm)
  | C vb => Hc vb
     ((fix list_Value_ind (vb : list RawValue) : AllT P f_reduce f_default vb :=
       match vb with
       | nil => f_default
       | v' :: vb' => f_reduce (RawValue_rect v') (list_Value_ind vb')
       end
     ) vb)
  end.
End RawValue_rect.

Error: The term "f_default" has type "Type" while it is expected to have type
 "AllSndT P f_reduce f_default nil".
*)

(** **** A more complex strong induction principle that also works for Type *)

(** The more complex induction principle described
#<a href="http://logical.saclay.inria.fr/coq-puma/messages/76e52ec953b33a7e">here</a>#
by Stephane Lescuyer is useful in defining RawValue_rect 
[[
RawValue_rect
     : forall (P : RawValue -> Type) (P' : list RawValue -> Type)
         (P'' : list (string * RawValue) -> Type),
       (forall g : General, P (G g)) ->
       (forall svm : list (string * RawValue), P'' svm -> P (E svm)) ->
       P'' nil ->
       (forall (s : string) (v : RawValue) (svm : list (string * RawValue)),
        P v -> P'' svm -> P'' ((s, v) :: svm)) ->
       (forall vb : list RawValue, P' vb -> P (C vb)) ->
       P' nil ->
       (forall (v : RawValue) (vb : list RawValue), P v -> P' vb -> P' (v :: vb)) ->
       forall v : RawValue, P v
]]
*)

Section RawValue_rect.
  Variable P : RawValue -> Type.
  Variable P' : list RawValue -> Type.
  Variable P'' : list (string * RawValue) -> Type.
  Hypothesis Hg : forall g : General, P (G g).
  Hypothesis He : forall svm : list (string * RawValue), P'' svm -> P (E svm).
  Hypothesis He_nil : P'' nil.
  Hypothesis He_cons : forall s v svm, P v -> P'' svm -> P'' ((s,v) :: svm).
  Hypothesis Hc : forall vb : list RawValue, P' vb -> P (C vb).
  Hypothesis Hc_nil : P' nil.
  Hypothesis Hc_cons : forall v vb, P v -> P' vb -> P' (v :: vb).

  Fixpoint RawValue_rect (v : RawValue) : P v :=
  match v as v0 return P v0 with
  | G g => Hg g
  | E svm => He svm 
     ((fix list_pairs_Value_rect (svm : list (string * RawValue)) : P'' svm :=
       match svm as svm0 return P'' svm0 with
       | nil => He_nil
       | (s,v') :: svm' => He_cons s v' svm' (RawValue_rect v') (list_pairs_Value_rect svm')
       end
     ) svm)
  | C vb => Hc vb
     ((fix list_Value_rect (vb : list RawValue) : P' vb :=
       match vb as vb0 return P' vb0 with
       | nil => Hc_nil
       | v' :: vb' => Hc_cons v' vb' (RawValue_rect v') (list_Value_rect vb')
       end
     ) vb)
  end.
End RawValue_rect.

(** As usual, RawValue_rec is just a specialization of RawValue_rect to Set *)
Definition RawValue_rec (P : RawValue -> Set) (P' : list RawValue -> Set) (P'' : list (string * RawValue) -> Set) :=
  RawValue_rect P P' P''.

(** We can also specialize RawValue_rect to Prop.
In this case the result is probably equivalent to RawValue_ind above.
At least Thorsten Altenkirch mentions 
#<a href="http://logical.saclay.inria.fr/coq-puma/messages/76e52ec953b33a7e">in this thread</a>#
that the two should be equivalent by "Bekic's lemma". *)

Definition RawValue_ind' (P : RawValue -> Prop) (P' : list RawValue -> Prop) (P'' : list (string * RawValue) -> Prop) :=
  RawValue_rect P P' P''.

(** XXX Can I even state this equivalence? -- Q: Why aren't [RawValue_ind P] and [RawValue_ind' P P' P''] in Prop?
[[
Lemma ind_equiv_ind' : forall (P : RawValue -> Prop)
      (P' : list RawValue -> Prop) (P'' : list (string * RawValue) -> Prop),
      RawValue_ind P -> RawValue_ind' P P' P''

The term "RawValue_ind P" has type
"(forall g : General, P (G g)) ->
 (forall svm : list (string * RawValue), AllSnd P svm -> P (E svm)) ->
 (forall vb : list RawValue, All P vb -> P (C vb)) -> forall v : RawValue, P v"
which should be Set, Prop or Type
]]
*)

(** *** The size of a values **) 

(** Below are four different ways of defining value size.
This definition is not used anywhere currently,
but it was a good exercise and might be useful for
proving by hand that certain complex functions on values
are terminating.
 *)

(** Note: [size_val] does not treat duplicate keys in entities in a special way. *)

(** Variant 1: using fold *)
Fixpoint size_val (v : RawValue) : nat :=
  match v with
  | G g => 1
  | E svm => fold_left (fun a sv => match sv with (s,v') => a + (size_val v') end) svm 1
  | C vb => fold_left (fun a v => a + (size_val v)) vb 1
  end.

(** Proving some properties of this version *)
Property size_G : forall g, size_val (G g) = 1.
Proof. intro. reflexivity. Qed.

Property size_E_nil : size_val (E nil) = 1.
Proof. reflexivity. Qed.

Lemma cons_to_app : forall (X : Type) (x : X) xs, x :: xs = app (x :: nil) xs.
Proof. reflexivity. Qed.

Require Import Arith.

Lemma size_E_cons_helper : forall svm n m,
fold_left
  (fun (a : nat) (sv : string * RawValue) =>
   let (_, v') := sv in a + size_val v') svm
  (n + m) =
n +
fold_left
  (fun (a : nat) (sv : string * RawValue) =>
   let (_, v') := sv in a + size_val v') svm m.
Proof.
  induction svm as [| [s' v'] svm']; intros n m.
  Case "svm = nil". reflexivity.
  Case "svm = (s,v) :: svm'". simpl. rewrite <- IHsvm'. rewrite plus_assoc. reflexivity.
Qed.

Lemma S_to_plus : forall n, S n = n + 1.
Proof. intro n. omega. Qed.

Property size_E_cons : forall s v svm, size_val (E ((s,v) :: svm)) = size_val v + size_val (E svm).
Proof. intros s v svm. simpl. rewrite S_to_plus. apply size_E_cons_helper. Qed.

Property size_C_nil : size_val (C nil) = 1.
Proof. reflexivity. Qed.

Lemma size_C_cons_helper : forall vb n m,
fold_left (fun (a : nat) (v0 : RawValue) => a + size_val v0) vb (n + m) =
n +
fold_left (fun (a : nat) (v0 : RawValue) => a + size_val v0) vb m.
Proof.
  induction vb as [| v vb']; intros n m.
  Case "vb = nil". reflexivity.
  Case "vb = v :: vb'". simpl. rewrite <- IHvb'. rewrite plus_assoc. reflexivity.
Qed.

Property size_C_cons : forall v vb, size_val (C (v :: vb)) = size_val v + size_val (C vb).
Proof. intros v vb. simpl. rewrite S_to_plus. apply size_C_cons_helper. Qed.

(** Variant 2: using nested fix -- works, but less elegant *)
Fixpoint size_val' (v : RawValue) : nat :=
  match v with
  | G g => 1
  | E svm =>
    (fix size_svm (svm : list (string * RawValue)) :=
      match svm with
      | nil => 1
      | (s,v') :: svm' => size_val' v' + size_svm svm'
      end
    ) svm       
  | C vb => 
     (fix size_svm (svm : list RawValue) :=
      match svm with
      | nil => 1
      | v' :: vb' => size_val' v' + size_svm vb'
      end
    ) vb  
  end.

(** We prove that size_val and size_val' are equivalent
    (trying out the induction hypothesis) *)

Lemma size_val_same : forall v, size_val v = size_val' v.
Proof.
  induction v as [g | svm | vb].
  Case "v = G g". reflexivity.
  Case "v = E svm".
    induction svm as [| [s' v'] svm'].
    SCase "svm = nil". reflexivity.
    SCase "svm = sv :: svb'". simpl in *.
      destruct H as [H1 H2].
      rewrite S_to_plus.
      rewrite size_E_cons_helper.
      rewrite <- H1.
      rewrite (IHsvm' H2).
      reflexivity.
  Case "v = C vb".
    induction vb as [| v vb'].
    SCase "vb = nil". reflexivity.
    SCase "vb = sv :: svb'". simpl in *.
      destruct H as [H1 H2].
      rewrite S_to_plus.
      rewrite size_C_cons_helper.
      rewrite <- H1.
      rewrite (IHvb' H2).
      reflexivity.
Qed.

(** Variant 3: Trying to give an even simpler definition to [size_val] using map and sum -
not as simple as it could be in the entity case. *)

Fixpoint sum (xs : list nat) : nat :=
  match xs with
    | nil => O
    | cons x xs' => plus x (sum xs')
  end.

Fixpoint map2 (X Y Z : Type) (f : Y -> Z) (xys : list (X * Y)) :=
  match xys with
  | nil => nil
  | ((x,y) :: xys') => (f y) :: (map2 X Y Z f xys')
  end.
Implicit Arguments map2 [X Y Z].

Fixpoint sum2 (X : Type) (xns : list (X*nat)) : nat :=
  match xns with
    | nil => O
    | (x,n) :: xns' => plus n (sum2 X xns')
  end.
Implicit Arguments sum2 [X].

Fixpoint size_val'' (v : RawValue) : nat :=
  match v with
  | G g => 1
(*  | E svm => sum (map size_val'' (snd (split svm))) + 1  -- this doesn't work! *)
(*  | E svm => sum (map2 size_val'' svm) + 1 -- this doesn't work: Recursive call to size_val'' has not enough arguments. *)
  | E svm => sum2 (map (fun sv => match sv with (s, v) => (s,size_val'' v) end) svm) +1
  | C vb => sum (map size_val'' vb) +1 (* this works *)
  end.

(** XXX Can I take the original case for entities 
[E svm => sum (map size_val'' (snd (split svm))) + 1], which Coq can't prove terminating,
and prove it terminating by hand?
[[
(*Require Export Recdef.*)
Function size_val''_simpler (v : RawValue) {measure size_val v} : nat :=
  match v with
  | G g => 1
  | E svm => sum (map size_val''_simpler (snd (split svm))) + 1
  | C vb => sum (map size_val''_simpler vb) +1
  end
]]
Anomaly: type_of: variable size_val''_simpler unbound. Please report.

*)
Definition f (P : RawValue -> Prop) (v : {x : RawValue | P x }) := v.

(* Lemma size_val''_same : forall v, size_val'' v = size_val v.
Proof.
  induction v as [g | svm | vb].
  Case "v = G g". reflexivity.
  Case "v = E svm". unfold size_val''. admit.
  Case "v = C vb". unfold size_val''.
*)

(** Variant 4: We can directly use RawValue_rec to define [size_val].
There is a lot of noise (many useless arguments) 
but the nice thing is that we don't need explicit recursion, 
and therefore we can use arbitrary functions for H... 
without having to worry about Coq's syntactic termination restrictions.
*)

Definition size_val'''' (v : RawValue) : nat :=
  let P := fun v  => nat in
  let P' := fun vb => nat in
  let P'' := fun svm => nat in
  let Hg := fun g => 1 in
  let He := fun svm => fun n => S n in
  let He_nil := 0 in
  let He_cons := fun s v svm => fun n m => n + m in
  let Hc := fun vb => fun n => S n in
  let Hc_nil := 0 in
  let Hc_cons := fun v vb => fun n m => n + m in
  RawValue_rec P P' P'' Hg He He_nil He_cons Hc Hc_nil Hc_cons v.
